/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	ChipAl
**
**FILE NAME:
**	CardAl.c
**
**ABSTRACT:
**	Card Abstraction Layer
**
**DETAILS:
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives/Reference Designs/Tigris/Linux/Kernel Driver/CardAL.c $
**	$Revision: 2 $
**	$Date: 1/22/01 11:38a $
*******************************************************************************
******************************************************************************/
#define _CARDAL_C_	1
#define		MODULE_ID			DBG_ENABLE_CARDAL

#include "linuxmodver.h"

#include "Product.h"
#include "CardMgmt.h"
#include "CardALV.h"
#include "SmSysIf.h"

#ifndef MEMORY_TAG
#define	MEMORY_TAG		'DXNC'
#endif

/******************************************************************************
Private Data Types
******************************************************************************/

typedef struct CARDAL_CONTEXT_S {
  void 				* AdapterContext;	// Global Adapter Context from CardALAdapterInit
  MODULE_LINE_CARD_METHODS_T	LineCard[CDAL_LINE_END];// PC Card with External Line Interface
  MODULE_CARDAL_OPTIONAL_T	CardAlModule[CDAL_OPTIONAL_END - CDAL_LINE_END];
  VOID				* ModuleContext[CDAL_MODULES_END];
} CARDAL_CONTEXT_T;

/******************************************************************************
Module Private Function Prototypes
******************************************************************************/
LOCAL void * get_module_context
(
  CDSL_ADAPTER_T		* pThisAdapter,
  unsigned int			ModuleId
)
{
  return
#ifdef DBG
    (ModuleId >= MAX_DRIVER_MODULES || !pThisAdapter) ? NULL :
#endif
    pThisAdapter->pModuleContext[ModuleId];
}

LOCAL void set_module_context
(
  CDSL_ADAPTER_T		* pThisAdapter,
  unsigned int			ModuleId,
  CARDAL_CONTEXT_T 		* UserContext
)
{
#ifdef DBG
  if(ModuleId < MAX_DRIVER_MODULES)
#endif
    pThisAdapter->pModuleContext[ModuleId] = UserContext;
}

LOCAL INLINE NDIS_STATUS cdalInitModules(
    CARDAL_CONTEXT_T 	* pCardAl
);

LOCAL INLINE NDIS_STATUS cdalEnableModules(
    CARDAL_CONTEXT_T 	* pCardAl
);
/******************************************************************************
Module Public Functions
******************************************************************************/

#if PROJECTS_MEDIUM == PROJECT_MEDIUM_ADSL
/******************************************************************************
*FUNCTION NAME:
*	CardALCfgInit
*
*ABSTRACT:
*	Requirement of the Architecture to read tuning parameters from a
*	Windowsxx Registry for the CardAL module registry parameters.
*
*RETURN:
*	NDIS_STATUS
******************************************************************************/
NDIS_STATUS CardALCfgInit(
    IN	CDSL_ADAPTER_T	 *pThisAdapter,
    IN	PTIG_USER_PARAMS pParams  //    parameter struct from the download app
) {
  NDIS_STATUS			Status=STATUS_SUCCESS;

  // read debug flag from registry cache into adapter var
  pThisAdapter->DebugFlag = pThisAdapter->PersistData.DebugFlag;

  return Status;
}
#endif

#if PROJECTS_MEDIUM == PROJECT_MEDIUM_ADSL
/******************************************************************************
FUNCTION NAME:
	CardALAdapterInit

ABSTRACT:
	CardAl Initialization of context structure. All CardAL child modules
	are initialized in a two stage process. First the Init function is
	called. If all goes well, the Enable function will then be called.

	* Note: All Child functions are optional

RETURN:
	Pointer to CardAl Context

DETAILS:
******************************************************************************/
NDIS_STATUS CardALAdapterInit(
  CDSL_ADAPTER_T	* pThisAdapter) {
  NDIS_STATUS		Status;
  CARDAL_CONTEXT_T 	* pCardAl = NULL;

  //*
  //* Allocate Parent CardAl Context Init Phase 0
  //*
  //*
  Status = ALLOCATE_MEMORY(pCardAl, sizeof(*pCardAl));
  if(Status != STATUS_SUCCESS) {
    NdisWriteErrorLogEntry(pThisAdapter->MiniportAdapterHandle,
			   NDIS_ERROR_CODE_OUT_OF_RESOURCES,
			   4,
			   NdisStatus,
			   __FILEID__,
			   __LINE__,
			   RSS_INIT_ERROR_FAILED_ALLOCATE_CARDAL_ADPTR);
    return(STATUS_FAILURE);
  }
  CLEAR_MEMORY(pCardAl, sizeof(*pCardAl));

  //*
  //* Sanity Check
  //*
#ifdef DBG
  if (sizeof(pCardAl->LineCard) != sizeof(gLineCards) ||
      sizeof(pCardAl->CardAlModule)/sizeof(pCardAl->CardAlModule[0]) !=
      sizeof(gOptCardModules)/sizeof(gOptCardModules[0]) ||
      ((sizeof(gLineCards)/sizeof(gLineCards[0])) +
       (sizeof(gOptCardModules)/sizeof(gOptCardModules[0]))) != CDAL_MODULES_END) {
    //* This is a build problem. The Global arrays of functions must match
    //* the pCardAl context size. The summation of the number of arrays must equal the
    //* number of context pointers.
    ErrPrt("Failed cardAL context size check\n");
    FREE_MEMORY(pCardAl,sizeof(*pCardAl),0);
    return(STATUS_FAILURE);
  }
#endif

  //*
  //* Exchanged pointers with public context structure
  //*
  set_module_context(pThisAdapter, IF_CARDAL_ID, pCardAl);
  pCardAl->AdapterContext = pThisAdapter;

  //*
  //* Load CardAl Module Entry Points
  //*
  COPY_MEMORY(&pCardAl->LineCard, &gLineCards, sizeof(pCardAl->LineCard));

  COPY_MEMORY(&pCardAl->CardAlModule, &gOptCardModules,
	      sizeof(pCardAl->CardAlModule));

  //*
  //* Initialize and enable all modules
  //*
  Status = cdalInitModules(pCardAl);
  if(Status != STATUS_SUCCESS) {
    CardALAdapterHalt(pThisAdapter);
    return(Status);
  }

  Status = cdalEnableModules(pCardAl);
  if(Status != STATUS_SUCCESS) {
    CardALAdapterHalt(pThisAdapter);
    return(Status);
  }

  //*
  //* Successful Initialization Complete
  //*
  return(STATUS_SUCCESS);
}
#endif

#if PROJECTS_MEDIUM == PROJECT_MEDIUM_ADSL
/******************************************************************************
FUNCTION NAME:
	CardAlShutdown

ABSTRACT:
	Release all resources allocated by CardAlInit.  Processing order is reversed
	from Init.

RETURN:
	VOID

DETAILS:
******************************************************************************/
NDIS_STATUS CardALAdapterHalt(
    CDSL_ADAPTER_T	* pThisAdapter
) {
  CARDAL_CONTEXT_T 	* pCardAl;
  DWORD 		Count;

  pCardAl = get_module_context(pThisAdapter, IF_CARDAL_ID);
  if(!pCardAl)
    return(STATUS_FAILURE);

  //*
  //* Release Optional Modules
  //*
  for (Count = CDAL_OPTIONAL_START; Count <  CDAL_OPTIONAL_END; Count++) {
    if(pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Shutdown) {
      (* pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Shutdown)(pCardAl->ModuleContext[Count]);
      if(!pCardAl->ModuleContext[Count]) {
	CardALAdapterHalt(pThisAdapter);
	return(STATUS_FAILURE);
      }
    }
  }

  //*
  //* Release Line Cards
  //*
  for (Count = CDAL_MODULES_START; Count <  CDAL_LINE_END; Count++) {
    if(pCardAl->LineCard[Count].Shutdown) {
      (* pCardAl->LineCard[Count].Shutdown)(pCardAl->ModuleContext[Count]);
      if(!pCardAl->ModuleContext[Count]) {
	CardALAdapterHalt(pThisAdapter);
	return(STATUS_FAILURE);
      }
    }
  }

  //*
  //* Release Primary Context
  //*
  set_module_context(pThisAdapter, IF_CARDAL_ID, NULL);

  FREE_MEMORY(pCardAl,sizeof(*pCardAl),0);

  return(STATUS_SUCCESS);
}
#endif

/******************************************************************************
FUNCTION NAME:
	CDALSetLineState

ABSTRACT:
	Set Line State.  This function is used to enable or disable a line.

RETURN:
	NTSTATUS

DETAILS:
******************************************************************************/
NDIS_STATUS CDALSetLineState(
    CDSL_ADAPTER_T	* pThisAdapter,
    CARD_LINE_STATE_T	NewState
) {
  NDIS_STATUS		Status = STATUS_SUCCESS;
  CARDAL_CONTEXT_T 	* pCardAl;
  DWORD			Count;

  pCardAl = get_module_context(pThisAdapter, IF_CARDAL_ID);
  if(!pCardAl)
    return(STATUS_FAILURE);

  for (Count = CDAL_MODULES_START;
       Count <  CDAL_LINE_END && Status == STATUS_SUCCESS;
       Count++) {
    if(pCardAl->LineCard[Count].SetLineState) {
      Status = (* pCardAl->LineCard[Count].SetLineState)(pCardAl->ModuleContext[Count], NewState);
    }
  }
  return(Status);
}

/******************************************************************************
CARDAL Module Functions
******************************************************************************/

/******************************************************************************
FUNCTION NAME:
	CDALGetModuleHandle

ABSTRACT:
	Return Handle of CardAL Module as related to the external 
	pThisAdapter handle

RETURN:
	Handle of requested module or Null if failure

DETAILS:
******************************************************************************/
VOID * CDALGetModuleHandle(
    CDSL_ADAPTER_T * pThisAdapter,
    CDAL_MODULES_T CDALModule
)
{
  CARDAL_CONTEXT_T 	* pCardAl;

  if(!pThisAdapter ||
     CDALModule >= CDAL_MODULES_END)
    return(NULL);

  pCardAl = get_module_context(pThisAdapter, IF_CARDAL_ID);
  if(!pCardAl)
    return(NULL);

  return(pCardAl->ModuleContext[CDALModule]);
}

/******************************************************************************
Private Functions
******************************************************************************/

/******************************************************************************
FUNCTION NAME:
	cdalInitModules

ABSTRACT:
	Call Init functions for all CardAl Modules

RETURN:
	NDIS_STATUS:

DETAILS:
******************************************************************************/
LOCAL INLINE NDIS_STATUS cdalInitModules(
    CARDAL_CONTEXT_T 	* pCardAl
) {
  DWORD	Count;

  for (Count = CDAL_MODULES_START; Count <  CDAL_LINE_END; Count++) {
    if(pCardAl->LineCard[Count].Init) {
      pCardAl->ModuleContext[Count] = (* pCardAl->LineCard[Count].Init)(pCardAl->AdapterContext,pCardAl);
      if(!pCardAl->ModuleContext[Count])
	return(STATUS_FAILURE);
    }
  }

  for (Count = CDAL_OPTIONAL_START; Count <  CDAL_OPTIONAL_END; Count++) {
    if(pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Init) {
      pCardAl->ModuleContext[Count] = (* pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Init)(pCardAl->AdapterContext,pCardAl);
      if(!pCardAl->ModuleContext[Count])
	return(STATUS_FAILURE);
    }
  }

  return(STATUS_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	cdalEnableModules

ABSTRACT:
	Call Enable functions for all CardAl Modules

RETURN:
	NDIS_STATUS:

DETAILS:
******************************************************************************/
LOCAL INLINE NDIS_STATUS cdalEnableModules(
    CARDAL_CONTEXT_T 	* pCardAl
) {
  DWORD 		Count;
  NDIS_STATUS		Status;

  for (Count = CDAL_MODULES_START; Count <  CDAL_LINE_END; Count++) {
    if(pCardAl->LineCard[Count].Enable) {
      Status = (* pCardAl->LineCard[Count].Enable)(pCardAl->ModuleContext[Count]);
      if(Status != STATUS_SUCCESS)
	return(STATUS_FAILURE);
    }
  }

  for (Count = CDAL_OPTIONAL_START; Count <  CDAL_OPTIONAL_END; Count++) {
    if(pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Enable) {
      Status = (* pCardAl->CardAlModule[Count - CDAL_OPTIONAL_START].Enable)(pCardAl->ModuleContext[Count]);
      if(Status != STATUS_SUCCESS)
	return(STATUS_FAILURE);
    }
  }

  return(STATUS_SUCCESS);
}

/***************************************************************************
FUNCTION NAME:
	CardAlRegStoreHeadEnd

ABSTRACT:
	
	This function is not implemented for Linux.

RETURN:
	none

DETAILS:
******************************************************************************/
NDIS_STATUS CardAlRegStoreHeadEnd(DWORD VendorId) {
  return(STATUS_SUCCESS);
}
