/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	CardAL
**
**FILE NAME:
**	CardALTigrisDp.c
**
**ABSTRACT:
**	Falcon HAL Interface to Pairgain Data Pump code
**
**DETAILS:
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Tigris/Linux_2.4.3_ATM_Driver/KernelModule/CardALTigrisDp.c $
**	$Revision: 11 $
**	$Date: 7/25/01 5:06p $
******************************************************************************/
#include "linuxmodver.h"

#include "Common.h"
#include "CardALTigrisV.h"
#include "../CommonData.h"
#include "CardALTigrisHal.h"
#include "CardALTigrisHalV.h"
#include "CardALTigrisDp.h"
#include "ChipALCdslV.h"
#include "CardALV.h"
#include "CardMgmt.h"
#include "sysinfo.h"

#include <linux/interrupt.h>
#include <asm/system.h>

#define CHIPAL_CONNECT_TIMER_ID  "CHCT"
#define CARDAL_MAX_CONNECT_CRC    50

static /*__inline*/ CDSL_ADAPTER_T *TIGRIS_CONTEXTtoCDSL_ADAPTER_CLN
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN const char		* fnc
)
{
  CDSL_ADAPTER_T		* pThisAdapter;

#if 0/*CLN*/
  printk(KERN_ERR "CnxADSL %s: %s pControllerContextS=%p\n",
	 __FUNCTION__,
	 fnc, pControllerContextS);
#endif

  if (!pControllerContextS)
    pThisAdapter = NULL;
  else {
    VERIFY_TIGRIS(pControllerContextS);

    pThisAdapter = ((TIGRIS_CONTEXT_T *)pControllerContextS)->pThisAdapter;

    VERIFY_ADAPTER(pThisAdapter);
  }

  return pThisAdapter;
}

#define TIGRIS_CONTEXTtoCDSL_ADAPTER(a) TIGRIS_CONTEXTtoCDSL_ADAPTER_CLN(a,__FUNCTION__)

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpGetCurrentTick

ABSTRACT:
	Return the Timer's Tick Count

RETURN:
******************************************************************************/
TICK_TYPE ChipAlTigrisDpGetCurrentTickV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  return CardAlTigrisGetCurrentTick(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS));
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpWaitQedfer

ABSTRACT:
	Waits on qedfer flags to be set. 

RETURN:
******************************************************************************/
void ChipAlTigrisDpWaitQedferV2
(
    IN TIGRIS_CONTEXT_S	* pControllerContextS,
    IN WORD		Flags,
    IN WORD		Resolution,
    IN WORD		Timeout
)
{
  CardAlTigrisWaitQedfer(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			 Flags, Resolution, Timeout);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDPGetCurrentTick

ABSTRACT:
	sets qedfer flags 

RETURN:
******************************************************************************/
void ChipAlTigrisDpSignalQedferV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN WORD Flags
)
{
  CardAlTigrisSignalQedfer(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			   Flags);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpLinkUp

ABSTRACT:
	Indicate Link is up

RETURN:
******************************************************************************/
BOOLEAN ChipAlTigrisDpLinkUp
(
    void   * pControllerContextS /* TIGRIS_CONTEXT_S * */
)
{
  TIGRIS_CONTEXT_T		* pControllerContext = pControllerContextS;
  CDSL_ADAPTER_T		* pThisAdapter;
  CARDAL_ADAPTER_T		* pCardALAdapter;
  PDATAPUMP_CONTEXT		pDatapump;
  DWORD				AdptrChk;

#if 1/*CLN*/
  if (!pControllerContextS)
    printk(KERN_ERR "CnxADSL %s: pControllerContextS=%p\n",
	   __FUNCTION__,
	   pControllerContextS);
#endif

  pThisAdapter = pControllerContext->pThisAdapter;

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(pThisAdapter, &pCardALAdapter);

  pDatapump = pControllerContext->pDatapump;

  VERIFY_TIGRIS(pControllerContext);
  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_CARDAL(pCardALAdapter);
  VERIFY_DATAPUMP(pDatapump);

  if (AdptrChk == RESULT_SUCCESS)
  {
    DWORD CrcCnt = pDatapump->sys_err_cnt.crc-pCardALAdapter->CurrentCRCErrors;

    if ((0 <= CrcCnt) && (CrcCnt < CARDAL_MAX_CONNECT_CRC))
    {
      pThisAdapter->MediaConnectStatus = MediaConnected;

      CmgmtIndicateLinkUp(pThisAdapter);

      pCardALAdapter->AdslLineUp = TRUE;
    }
    else
    {
      // reschedule timer and try later
      pCardALAdapter->CurrentCRCErrors = pDatapump->sys_err_cnt.crc;

      UtilTmrStartTmr(&pThisAdapter->TimerBlock,
		      pCardALAdapter,
		      (DWORD)CHIPAL_CONNECT_TIMER_ID,
		      TIMER_RESOLUTION_2_SEC,
		      &ChipAlTigrisDpLinkUp,
		      pControllerContextS);
    }
  }

  return TRUE; // only one shot timer
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDPIndicateLinkUp

ABSTRACT:
	Indicate Link is up

RETURN:
******************************************************************************/
void ChipAlTigrisDPIndicateLinkUp
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CDSL_ADAPTER_T		* pThisAdapter;
  CARDAL_ADAPTER_T		* pCardALAdapter;
  PDATAPUMP_CONTEXT		pDatapump;
  DWORD				AdptrChk;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS);

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(pThisAdapter, &pCardALAdapter);

  pDatapump = ((TIGRIS_CONTEXT_T *)pControllerContextS)->pDatapump;

  VERIFY_TIGRIS(pControllerContextS);
  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_CARDAL(pCardALAdapter);
  VERIFY_DATAPUMP(pDatapump);

  if (AdptrChk == RESULT_SUCCESS)
  {
    UtilTmrClearTmr(&pThisAdapter->TimerBlock,
		    pCardALAdapter,
		    (DWORD)CHIPAL_CONNECT_TIMER_ID);

    pCardALAdapter->CurrentCRCErrors = pDatapump->sys_err_cnt.crc;

    UtilTmrStartTmr(&pThisAdapter->TimerBlock,
		    pCardALAdapter,
		    (DWORD)CHIPAL_CONNECT_TIMER_ID,
		    TIMER_RESOLUTION_2_SEC,
		    &ChipAlTigrisDpLinkUp,
		    pControllerContextS);
  }
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDPIndicateLinkDown

ABSTRACT:
	Indicate Link is down

RETURN:
******************************************************************************/
void ChipAlTigrisDPIndicateLinkDown
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CDSL_ADAPTER_T		* pThisAdapter;
  CARDAL_ADAPTER_T		* pCardALAdapter;
  DWORD				AdptrChk;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS);
  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(pThisAdapter, &pCardALAdapter);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_CARDAL(pCardALAdapter);
  VERIFY_TIGRIS(pControllerContextS);

  // Clear any outstanding Connect timer
  UtilTmrClearTmr(&pThisAdapter->TimerBlock,
		  pCardALAdapter,
		  (DWORD)CHIPAL_CONNECT_TIMER_ID);

  pThisAdapter->MediaConnectStatus = MediaNotConnected;

  CmgmtIndicateLinkDown(pThisAdapter);

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(pThisAdapter, &pCardALAdapter);
  if (AdptrChk == RESULT_SUCCESS)
    pCardALAdapter->AdslLineUp = FALSE;
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpDisableIRQ1_2

ABSTRACT:
	Disable ADSL interrupts

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpDisableIRQ1_2V2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CDSL_ADAPTER_T 		*pThisAdapter;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS);
  // if I haven't been initialized then just ignore the request
  if (pThisAdapter) {
    CHIPAL_T			* pThisInterface;
    TIGRIS_HAL_CONTEXT_T	* pContext;

    // get the hal context
    pThisInterface = (CHIPAL_T *)pThisAdapter->ChipMod;
    pContext = (TIGRIS_HAL_CONTEXT_T *)pThisInterface->ServiceContext[CAL_SERVICE_FALCON];

    // Keep a count to support nested calls
    pContext->DisableInterruptCount++;

    // Disable interrupts together
    // (This runs at interrupt level (DIRQL) because this involves a read/modify/write operation)
    ChipAlTigrisHalDisableIRQ1_2(pThisAdapter);
  }
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpEnableIRQ1_2

ABSTRACT:
	Enable ADSL interrupts

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpEnableIRQ1_2V2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CDSL_ADAPTER_T		* pThisAdapter;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS);
  // if I haven't been initialized then just ignore the request
  if (pThisAdapter) {
    CHIPAL_T			* pThisInterface;
    TIGRIS_HAL_CONTEXT_T	* pContext;

    // get the hal context
    pThisInterface = (CHIPAL_T *)pThisAdapter->ChipMod;
    pContext = (TIGRIS_HAL_CONTEXT_T *)pThisInterface->ServiceContext[CAL_SERVICE_FALCON];

    // if this is the bottom call of nested calls.
    if (--pContext->DisableInterruptCount == 0)
    {
      // We can twiddle the bits at any level since interrupts are off
      ChipAlTigrisHalEnableIRQ1_2(pThisAdapter);
    }
  }
}

static unsigned long AtomicFlags;/*CLN*/ /* These 2 static vars are safe */
static long AtomicCnt;/*CLN*/

/***************************************************************************
   Macro         : ChipAlTigrisDpAtomicStart
   Parameters    : none
   Returns       : none
   Description   : This macro will disable interrupts until the next atomic
   				   end call.
****************************************************************************/
MODULAR VOID ChipAlTigrisDpAtomicStartV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  unsigned long LocalFlags;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  local_irq_save(LocalFlags);
#else
  save_and_cli(LocalFlags);
#endif

  // if this is the first start then
  // save the interrupt enable state
  // and disable interrupts
  if (AtomicCnt++ == 0)
    AtomicFlags = LocalFlags;
}

/***************************************************************************
   Macro         : ChipAlTigrisDpAtomicEnd
   Parameters    : none
   Returns       : none
   Description   : This macro will re-enable interrupts after an atomic
   				   start call.
****************************************************************************/
MODULAR VOID ChipAlTigrisDpAtomicEndV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
#if 1/*CLN*/
  if (AtomicCnt <= 0)
    printk(KERN_ERR "CnxADSL %s: AtomicCnt <= 0\n",
	   __FUNCTION__);
  else
#endif
  if (--AtomicCnt == 0)
  {
    // if this is the last end then
    // restore the interrupt enable state
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    local_irq_restore(AtomicFlags);
#else
    restore_flags(AtomicFlags);
#endif
  }
}

/******************************************************************************
FUNCTION NAME:
    CardALTigrisSITrap

ABSTRACT:
	Wrapper routine for controller code to notify us that it's ready to 
	process software trap

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpSITrapV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CardALTigrisSITrap(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS));
}

/***************************************************************************
   Macro         : ChipAlTigrisDpIsrStart
   Parameters    : none
   Returns       : none
   Description   : This function is here to satisfy the PairGain code.  For
                   Linux it has nothing to do.
****************************************************************************/
MODULAR VOID ChipAlTigrisDpIsrStartV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
}

/***************************************************************************
   Macro         : ChipAlTigrisDpIsrEnd
   Parameters    : none
   Returns       : none
   Description   : This function is here to satisfy the PairGain code.  For
                   Linux it has nothing to do.
****************************************************************************/
MODULAR VOID ChipAlTigrisDpIsrEndV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
}

/***************************************************************************
   Macro         : ChipAlTigrisDpDelayMsec
   Parameters    : duration - delay amount in milliseconds.
   Returns       : none
   Description   : This macro will call the cardAL function to pause for the
                   requested amount of time.
****************************************************************************/
MODULAR VOID ChipAlTigrisDpDelayMsec 
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    DWORD			duration 
)
{
  CDSL_ADAPTER_T		* pThisAdapter;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS);
  CardALDelayMsec(pThisAdapter, duration);
}

/***************************************************************************
   Macro         : ChipAlTigrisDpGetNonVolatileParms
   Parameters    : pointers to the return values read from the registry   
   Returns       : none
   Description   : This function returns the nonvolatile params read from
                   the system registry
****************************************************************************/
MODULAR VOID ChipAlTigrisDpGetNonVolatileParms
(
    IN PDATAPUMP_CONTEXT	pDatapump,
    IN XCVR_OTHER_CFG_TYPE	* pXcvr
)
{
  CDSL_ADAPTER_T		* pThisAdapter;

  pThisAdapter = TIGRIS_CONTEXTtoCDSL_ADAPTER(pDatapump->pExternalContext);

  // if I haven't been initialized then just ignore the request
  if (pThisAdapter) {
    pXcvr->sys_autohandshake	    = pThisAdapter->PersistData.AutoSenseHandshake;
    pXcvr->sys_autosense_wiresfirst = pThisAdapter->PersistData.AutoSenseWiresFirst;
    pXcvr->sys_autowiring_selection = pThisAdapter->PersistData.AutoWiringSelection;
    pXcvr->sys_autowiring_num_combos= pThisAdapter->PersistData.AutoWiringNumCombos;
    pXcvr->AdslHeadEnd		    = pThisAdapter->PersistData.AdslHeadEnd;
    pXcvr->AdslHeadEndEnvironment   = pThisAdapter->PersistData.AdslHeadEndEnvironment;
  }
}

/***************************************************************************
   Macro         : ChipAlTigrisDpTestAndSet
   Parameters    : flag = the flag we want to set, we will set it if it's
                   not already set.
   Returns       : return TRUE when flag is not set
                   and FALSE if already set
   Description   : This macro will perform atomic testing and setting
                   of the flag.
   Note          : BSO provides a library macro for C16x to perform this
                   operation.
****************************************************************************/
MODULAR BOOLEAN ChipAlTigrisDpTestAndSet
(
    IN volatile LOCK_TYPE		* x
)
{
  BOOLEAN RetVal = !test_and_set_bit(0, x);

  return RetVal;
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSetRstXcvr

ABSTRACT:
	Sets or Resets the transceiver

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpSetRstXcvrV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BOOLEAN			State
)
{
  ChipAlTigrisHalSetRstXcvr(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			    State);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSetGpio

ABSTRACT:
	Sets or Resets the GPIO pin

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpSetGpioV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BYTE			Pin,
    IN BOOLEAN			State
)
{
  ChipAlTigrisSetGpio(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
		      Pin, State);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSetAFEHybridSelect

ABSTRACT:
	Sets or Resets the AFE Hybrid Select lines

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpSetAFEHybridSelectV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN UINT8			Select_Line,
    IN BOOLEAN			State
)
{
  ChipAlTigrisHalSetAFEHybridSelect(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
				    Select_Line, State);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSetPwrDwnXcvr

ABSTRACT:
	Sets or Resets the Falcon Power Down pin

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpSetPwrDwnXcvrV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BOOLEAN			State
)
{
  ChipAlTigrisSetPwrDwnXcvr(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			    State);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpWriteAFE

ABSTRACT:
	Write data to AFE at specified location

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpWriteAFEV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BYTE			Address,
    IN BYTE			Data
)
{
  ChipAlTigrisHalWriteAfe(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			  Address, Data);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpPendingReadAFE

ABSTRACT:
	Send a read data command to AFE at specified location

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpPendingReadAFEV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BYTE			Address
)
{
  ChipAlTigrisHalPendingReadAfe(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
				Address);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpReadAFE

ABSTRACT:
	Read data from AFE at location defined in ChipAlTigrisDpPendingReadAFE()

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR BYTE ChipAlTigrisDpReadAFEV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  return ChipAlTigrisHalReadAfe(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS));
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpIsAFEDone

ABSTRACT:
	Return whether previous write/read has completed

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR BOOLEAN ChipAlTigrisDpIsAFEDoneV2
(
    /* IN MIC_WaitAFEDone */
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  return ChipAlTigrisHalIsAfeDone(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS));
}

/******************************************************************************
FUNCTION NAME:
	ChipAlDpSwitchHookStateEnq

ABSTRACT:

RETURN:
	OffHook = 0
	OnHook = 1

DETAILS:
******************************************************************************/
GLOBAL BOOLEAN ChipAlTigrisDpSwitchHookStateEnqV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  return ChipAlSwitchHookStateEnq(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS));
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpTraceWrite

ABSTRACT:
	Write Value to PCI Trace Location

RETURN:
	NA.

DETAILS:
******************************************************************************/
MODULAR VOID ChipAlTigrisDpTraceWriteV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
    IN BYTE			Offset,
    IN DWORD			Value
)
{
  ChipAlTigrisHalTraceWrite(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			    Offset, Value);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpGetVendorId

ABSTRACT:
	Return Vendor ID

RETURN:
	.

SEE ALSO:
	MIC_GetVendorId
******************************************************************************/
GLOBAL UINT16 ChipAlTigrisDpGetVendorIdV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CARDAL_ADAPTER_T	* pCardALAdapter;
  DWORD			AdptrChk;

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
					      &pCardALAdapter);

  return CdalADSLGetVendorId(pCardALAdapter);
}

#if 0 /* Not used in linux */
/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpGetAutoWiring

ABSTRACT:
	Return Auto Wiring
        Not used under linux?

RETURN:
******************************************************************************/
GLOBAL UINT16 ChipAlTigrisDpGetAutoWiringV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CARDAL_ADAPTER_T	* pCardALAdapter;
  DWORD			AdptrChk;

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
					      &pCardALAdapter);

  return CdalADSLGetAutoWiring(pCardALAdapter);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpGetAutoHandshake

ABSTRACT:
	Return Auto Wiring
        Not used under linux?

RETURN:
******************************************************************************/
GLOBAL UINT16 ChipAlTigrisDpGetAutoHandshakeV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CARDAL_ADAPTER_T	* pCardALAdapter;
  DWORD			AdptrChk;

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
					      &pCardALAdapter);

  return CdalADSLGetAutoHandshake(pCardALAdapter);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpGetAutoSenseWiresFirst

ABSTRACT:
	Return Auto Wiring
        Not used under linux?

RETURN:
******************************************************************************/
GLOBAL UINT16 ChipAlTigrisDpGetAutoSenseWiresFirstV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS
)
{
  CARDAL_ADAPTER_T	* pCardALAdapter;
  DWORD			AdptrChk;

  AdptrChk = cardALChkAdptrStrucsOkNotDnldgDP(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
					      &pCardALAdapter);

  return CdalADSLGetAutoSenseWiresFirst(pCardALAdapter);
}
#endif

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSetWiring

ABSTRACT:
	Set the wiring relays to the desired pair

RETURN:
	.

SEE ALSO:
	auto_PermuteAutoWire
******************************************************************************/
GLOBAL void ChipAlTigrisDpSetWiringV2
(
    IN TIGRIS_CONTEXT_S		* pControllerContextS,
 BOOLEAN			AutomaticDetection,
 SYS_WIRING_SELECTION_TYPE	WiringSelection
)
{
  ChipAlTigrisHalSetWiring(TIGRIS_CONTEXTtoCDSL_ADAPTER(pControllerContextS),
			   AutomaticDetection, WiringSelection);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSER_InitMsgQ

ABSTRACT:

RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpSER_InitMsgQ
(
    IN CDSL_ADAPTER_T	* pThisAdapter
)
{
  TIGRIS_CONTEXT_T	* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  SER_InitMsgQ(pControllerContext->pDatapump);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpTRM_TerminalInit

ABSTRACT:

RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpTRM_TerminalInit
(
    IN CDSL_ADAPTER_T		* pThisAdapter
)
{
  TIGRIS_CONTEXT_T		* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  TRM_TerminalInit(pControllerContext->pDatapump);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSER_DetectTerminalConnected

ABSTRACT:

RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpSER_DetectTerminalConnected
(
    IN CDSL_ADAPTER_T		* pThisAdapter
)
{
  TIGRIS_CONTEXT_T		* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  SER_DetectTerminalConnected(pControllerContext->pDatapump);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpTRM_AutoUpdateDisable

ABSTRACT:

RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpTRM_AutoUpdateDisable
(
    IN CDSL_ADAPTER_T		* pThisAdapter,
    SYS_LOGON_MODE_TYPE		mode,
    BOOLEAN			disable
)
{
  TIGRIS_CONTEXT_T		* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  TRM_AutoUpdateDisable(pControllerContext->pDatapump, mode, disable);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSER_ClearRemoteLogon

ABSTRACT:

RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpSER_ClearRemoteLogon
(
    IN CDSL_ADAPTER_T		* pThisAdapter
)
{
  TIGRIS_CONTEXT_T		* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  SER_ClearRemoteLogon(pControllerContext->pDatapump);
}

/******************************************************************************
FUNCTION NAME:
	ChipAlTigrisDpSER_RcvIntrHandler

ABSTRACT:
	
RETURN:
******************************************************************************/
GLOBAL void ChipAlTigrisDpSER_RcvIntrHandler
(
    IN CDSL_ADAPTER_T		* pThisAdapter,
    IN char			input_char
)
{
  TIGRIS_CONTEXT_T		* pControllerContext;

  pControllerContext = CDALGetModuleHandle(pThisAdapter, CARDAL_TIGRIS_CONTROLLER);

  VERIFY_ADAPTER(pThisAdapter);
  VERIFY_TIGRIS(pControllerContext);
  VERIFY_DATAPUMP(pControllerContext->pDatapump);

  SER_RcvIntrHandler(pControllerContext->pDatapump, input_char);
}
