/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	CardAL (Card Abstraction Layer)
**
**FILE NAME:
**	CardALV.h
**
**ABSTRACT:
**	This file implements the top level routines that interface to
**	the Linux ATM layer.
**
**DETAILS:
**	????
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Yukon-AV/Linux_2.4.3_ATM_Driver/KernelModule/CardMgmt.c $
**	$Revision: 23 $
**	$Date: 7/27/01 8:46a $
*******************************************************************************
******************************************************************************/

#include "linuxmodver.h"

#include "CnxADSL.h"

#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <asm/unistd.h>
#include <linux/signal.h>
#include <asm/semaphore.h>

#include "LnxTools.h"
#include "../CommonData.h"
#include "Version.h"

#include "CnxTTY.h"
#include "ARMAbstract.h"
#include "CardMgmtVcV.h"
#include "CardMgmtLinkV.h"
#include "CardMgmtV.h"
#include "CardMgmtCtl.h"
#include "SmSysIf.h"
#include "SmSysIfv.h"
#include "CardALV.h"

#include "CardALTigris.h"

#if BITS_PER_LONG != 32
#  error FIXME: this driver only works on 32-bit platforms
#endif

#ifndef PCI_VENDOR_ID_CNX
#define PCI_VENDOR_ID_CNX 0x14f1
#endif
#ifndef PCI_DEVICE_ID_CNX_TIGRIS_ARB
#define PCI_DEVICE_ID_CNX_TIGRIS_ARB 0x1610
#endif
#ifndef PCI_DEVICE_ID_CNX_TIGRIS_ADSL
#define PCI_DEVICE_ID_CNX_TIGRIS_ADSL 0x1611
#endif
#ifndef PCI_DEVICE_ID_CNX_YUKON_ARB
#define PCI_DEVICE_ID_CNX_YUKON_ARB 0x1620
#endif
#ifndef PCI_DEVICE_ID_CNX_SOFTK56
#define PCI_DEVICE_ID_CNX_SOFTK56 0x1621
#endif
#ifndef PCI_DEVICE_ID_CNX_YUKON_ADSL
#define PCI_DEVICE_ID_CNX_YUKON_ADSL 0x1622
#endif

#ifndef DEV_LABEL
#define DEV_LABEL CNX_DEV_LABEL
#endif

/******************************************************************************
**	defines for this function
******************************************************************************/
#define ERROR_FATAL(error)	(LEVEL_FATAL   | MODULE_CARDMGMT | INSTANCE_CARDMGMT | error)
#define ERROR_WARNING(error)	(LEVEL_WARNING | MODULE_CARDMGMT | INSTANCE_CARDMGMT | error)
#define ERROR_NORMAL(error)	(LEVEL_NORMAL  | MODULE_CARDMGMT | INSTANCE_CARDMGMT | error)

#define BD_HEADER_SIZE		(sizeof(MINI_BACK_DOOR_T) - sizeof(MINI_BD_STRUC_UNION_T))

#define ONE_SECOND		1000

//////////////////////////////////////////////
// Linux ATM interface function prototypes
//////////////////////////////////////////////
LOCAL void
CmgmtDeviceClose(
    IN struct atm_dev*	pDev
);

LOCAL int
CmgmtOpen(
    IN struct atm_vcc*	pVcc
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    ,
    IN short	 	vpi,
    IN int		vci
#endif
);

LOCAL void
CmgmtClose(
    IN struct atm_vcc*	pVcc);

LOCAL int
CmgmtProcIOCTL(
    IN struct atm_dev*	pDev,
    IN UINT		Command,
    INOUT PVOID		pArgument);

LOCAL int
CmgmtGetSocketOption(
    IN struct atm_vcc*	pVcc,
    IN int		OptionLevel,
    IN int		OptionName,
    OUT PVOID		OptionValue,
    IN int		OptionLength);

LOCAL int
CmgmtSetSocketOption(
    IN struct atm_vcc*	pVcc,
    IN int		OptionLevel,
    IN int		OptionName,
    IN PVOID		OptionValue,
    IN int		OptionLength);

LOCAL int
CmgmtSendBuffer(
    IN struct atm_vcc*	pVcc,
    IN SK_BUFF_T*	pSkb);

#if 0
LOCAL void
CmgmtPhyPut(
    IN struct atm_dev*	pDev,
    IN UCHAR		value,
    IN ULONG		addr);

LOCAL UCHAR
CmgmtPhyGet(
    IN struct atm_dev*	pDev,
    OUT ULONG		addr);

LOCAL int
CmgmtChangeQOS(
    IN struct atm_vcc*	pVcc,
    IN struct atm_qos*	pQOS,
    IN int		flags);
#endif

LOCAL int
CmgmtPhyStart(
    IN struct atm_dev*	pDev
);

LOCAL int
CmgmtPhyIOCTL(
    IN struct atm_dev*	pDev,
    IN unsigned int	Command,
    INOUT PVOID		pArg);

LOCAL void
CmgmtPhyInterrupt(
    IN struct atm_dev*	pDev
);

LOCAL int
CmgmtProcessRead(
    IN struct atm_dev	*pDev,
    IN loff_t		*pProcPageOffset,
    IN char		*ProcPage);

#ifdef CONFIG_PROC_FS
enum tyuiu {DebugQuery,RxSnr,Bins,X};

LOCAL int
ReadProcCnxAdsl(
    IN char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData,
    enum tyuiu	mode);

LOCAL int
ReadProcCnxAdslDebugQuery(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData);

LOCAL int
ReadProcCnxAdslRxSnr(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData);

LOCAL int
ReadProcCnxAdslBins(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData);

LOCAL int
ReadProcCnxAdslX(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData);
#endif

/******************************************************************************
**	External global variables...
******************************************************************************/
DWORD gAllocCnt;
#if MEM_ALLOC_DBG
ALLOC_HEAD_T gAllocList;
#endif

///////////////////////////////////////////////////////////////////////////////
// Local functions
///////////////////////////////////////////////////////////////////////////////
LOCAL NDIS_STATUS
CmgmtAdapterReset(
    IN CDSL_ADAPTER_T	* pThisAdapter);

LOCAL VOID
CmgmtMainTimerInterrupt(
    IN ULONG		 pThisAdapter);

/******************************************************************************
**	This is the master vector table for all module generic functions.
******************************************************************************/

static const MODULE_FUNCTIONS_T DownloadModuleMethods[] = {
  CHIPAL_FNS,	// Chip Abstraction Layer Module
  ARMAL_FNS	// ARM Abstraction module
};

static const MODULE_FUNCTIONS_T ModemModuleMethods[] = {
  CARDAL_FNS,	//Card Abstraction Module
  BUFMGMT_FNS,	//Buffer Management Module
  FRAMEAL_FNS,	//Frame Abstraction Module
  CELLTEST_FNS  //Cell data test Module
};

// Structure of function pointers used by the Linux ATM layer
static const struct atmdev_ops DeviceOperations = {
  dev_close: CmgmtDeviceClose,
  open: CmgmtOpen,
  close: CmgmtClose,
  ioctl: CmgmtProcIOCTL,
  getsockopt: CmgmtGetSocketOption,
  setsockopt: CmgmtSetSocketOption,
  send: CmgmtSendBuffer,
#if 0
  sg_send: NULL,
#endif
#if 0 /* keep the current hack for now */
  send_iovec: NULL,
#endif
#if 0
  sg_oam: NULL,
#endif
#if 0
  phy_put: NULL, //CmgmtPhyPut,
  phy_get: NULL, //CmgmtPhyGet,
#if 0
  feedback: NULL,
#endif
  change_qos: NULL, //CmgmtChangeQOS,
#endif
  proc_read: CmgmtProcessRead,
  owner: THIS_MODULE,
};

static const struct atmphy_ops PhyOperations = {
  start: CmgmtPhyStart,
  ioctl: CmgmtPhyIOCTL,
  interrupt: CmgmtPhyInterrupt
};

/*
  There are 2 AccessRunner PCI cards              Arb      ,ADSL     ,K56/V92
  Tigris: Arbitration chip and ADSL               14f1:1610,14f1:1611
  Yukon:  Arbitration chip and ADSL and K56 modem 14f1:1620,14f1:1622,14f1:1621
*/

static const struct pci_device_id cnxadsl_pci_tbl[] __devinitdata = {
  { PCI_VENDOR_ID_CNX, PCI_DEVICE_ID_CNX_TIGRIS_ADSL, PCI_ANY_ID, PCI_ANY_ID,
    0, 0, 0},
  { PCI_VENDOR_ID_CNX, PCI_DEVICE_ID_CNX_YUKON_ADSL, PCI_ANY_ID, PCI_ANY_ID,
    0, 0, 0},
  { 0, }
};
MODULE_DEVICE_TABLE(pci, cnxadsl_pci_tbl);

static const struct pci_device_id cnxarb_pci_tbl[] __devinitdata = {
  { PCI_VENDOR_ID_CNX,PCI_DEVICE_ID_CNX_TIGRIS_ARB,
    PCI_ANY_ID,PCI_ANY_ID, 0, 0, 0 },
  { PCI_VENDOR_ID_CNX,PCI_DEVICE_ID_CNX_YUKON_ARB,
    PCI_ANY_ID,PCI_ANY_ID, 0, 0, 0 },
  { 0, }
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static int make_procname(
    char *s,
    int number,
    const char *dir
) {
  int rc;

  rc = sprintf(s,"driver/" DEV_LABEL);

  if (number >= 0) {
    rc += sprintf(s+rc,"/%d",
		  number);

    if (dir)
      rc += sprintf(s+rc,"/%s",
		    dir);
  }

  return rc;
}

static struct proc_dir_entry *my_create_proc_read_entry(
    int number,
    const char *dir,
    read_proc_t *read_proc,
    void *data
) {
  char name[50+1];

  make_procname(name, number, dir);

  if (!read_proc)
    return proc_mkdir(name, 0);
  else
    return create_proc_read_entry(name,
				  (S_IFREG | S_IRUGO),
				  NULL,
				  read_proc,
				  data);
}

static void my_remove_proc_entry(
    int number,
    const char *dir
) {
  char name[50+1];

  make_procname(name, number, dir);

  remove_proc_entry(name, NULL);
}
#endif

static void cnxadsl_fail(
    unsigned int level,
    CDSL_ADAPTER_T * pThisAdapter) {
  switch(level) {
  default:
#ifdef CONFIG_PROC_FS
  case 11:
    my_remove_proc_entry(pThisAdapter->pAtmDevice->number, "X");
  case 10:
    my_remove_proc_entry(pThisAdapter->pAtmDevice->number, "Bins");
  case 9:
    my_remove_proc_entry(pThisAdapter->pAtmDevice->number, "RxSnr");
  case 8:
    my_remove_proc_entry(pThisAdapter->pAtmDevice->number, "DebugQuery");
  case 7:
    my_remove_proc_entry(pThisAdapter->pAtmDevice->number, NULL);
  case 6:
    my_remove_proc_entry(-1, NULL);
#endif

  case 5:
#if HAS_TTY
    cnxtty_fini(pThisAdapter);
#endif

  case 4:
    atm_dev_deregister(pThisAdapter->pAtmDevice);

  case 3:
#if 0
    pci_disable_device(pThisAdapter->pci_dev_adsl);
#endif

  case 2:
    if (pThisAdapter->Tune) /* In case ChipALCfgInit/AdapterInit allocated it*/
      FREE_MEMORY(pThisAdapter->Tune,0,0);

  case 1:
    FREE_MEMORY(pThisAdapter,0,0);

  case 0:
    ;
  }
}

LOCAL int __devinit cnxadsl_init_one(
    struct pci_dev *pci_dev_adsl,
    const struct pci_device_id *ent) {
  int			Status;
  CDSL_ADAPTER_T	*pThisAdapter = NULL;
  struct pci_dev	*pci_dev_arm = NULL;
  int			error = -ENOMEM;
  unsigned int          level = 0;

  {
    const struct pci_device_id *p;

    for (p = cnxarb_pci_tbl; p->vendor; p++)
      while ((pci_dev_arm = pci_find_device(p->vendor, p->device, pci_dev_arm)))
	if (PCI_SLOT(pci_dev_adsl->devfn) == PCI_SLOT(pci_dev_arm->devfn))
	  goto found;

    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>%s:%s %s  -- Failed to find PCI %s device!",
		      __FILE__, __LINE__, __FUNCTION__, "arm"));
    error = -ENODEV;
    goto fail;

  found:
    ;
  }

  // allocate an empty device structure
#define MEMORY_TAG 0
  Status = ALLOCATE_MEMORY(pThisAdapter, sizeof(*pThisAdapter));
  if (Status != STATUS_SUCCESS)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:cnxadsl_init_one  Memory alloc fail"));

    goto fail;
  } else
    level++;

  // init the structure
  CLEAR_MEMORY(pThisAdapter, sizeof(*pThisAdapter));

  // initialize the new device structure
  pThisAdapter->pci_dev_arm = pci_dev_arm;
  pThisAdapter->pci_dev_adsl = pci_dev_adsl;
  pThisAdapter->CnxtPhysicalDriverType =
    (pci_dev_arm->device == PCI_DEVICE_ID_CNX_TIGRIS_ARB) ?
    DRIVER_TYPE_TIGRIS : DRIVER_TYPE_YUKON;
  pThisAdapter->TimerResolution = DEFAULT_TIMER_RESOLUTION;

  {
    TIG_USER_PARAMS *p = &pThisAdapter->PersistData;

    p->CnxtVendorId = pThisAdapter->pci_dev_arm->vendor;
    p->CnxtArmDeviceId = pThisAdapter->pci_dev_arm->device;
    p->CnxtAdslDeviceId = pThisAdapter->pci_dev_adsl->device;
    p->CnxtPhysicalDriverType = pThisAdapter->CnxtPhysicalDriverType;
  }

  level++;
  pThisAdapter->AdapterStructureAllocated = TRUE;

#if 0
  if (pci_enable_device(pci_dev_adsl)) {
    error = -ENODEV;
    goto fail;
  } else
#endif
    level++;

  /**********************************************************************
   **	Load the modules methods table...
   **********************************************************************/
  // register as an ATM device
  pThisAdapter->pAtmDevice = atm_dev_register(DEV_LABEL,
					      &DeviceOperations,
					      -1,
					      0);
  if (!pThisAdapter->pAtmDevice)
    goto fail;
  else
    level++;

  pThisAdapter->pAtmDevice->dev_data = pThisAdapter;

  // initialize the driver fields of the atm device structure
  pThisAdapter->pAtmDevice->ci_range.vpi_bits = 10;
  pThisAdapter->pAtmDevice->ci_range.vci_bits = 16;
  pThisAdapter->pAtmDevice->link_rate = ATM_OC3_PCR;

  // insert my phy function pointers
  pThisAdapter->pAtmDevice->phy = &PhyOperations;

#if HAS_TTY
  cnxtty_init(pThisAdapter);
#endif
  level++;

#ifdef CONFIG_PROC_FS
  if (!my_create_proc_read_entry(-1, NULL,
				 NULL, NULL))
    goto fail;
  else
    level++;

  if (!my_create_proc_read_entry(pThisAdapter->pAtmDevice->number, NULL,
				 NULL, NULL))
    goto fail;
  else
    level++;

  if (!my_create_proc_read_entry(pThisAdapter->pAtmDevice->number, "DebugQuery",
				 ReadProcCnxAdslDebugQuery, pThisAdapter))
    goto fail;
  else
    level++;

  if (!my_create_proc_read_entry(pThisAdapter->pAtmDevice->number, "RxSnr",
				 ReadProcCnxAdslRxSnr, pThisAdapter))
    goto fail;
  else
    level++;

  if (!my_create_proc_read_entry(pThisAdapter->pAtmDevice->number, "Bins",
				 ReadProcCnxAdslBins, pThisAdapter))
    goto fail;
  else
    level++;

  if (!my_create_proc_read_entry(pThisAdapter->pAtmDevice->number, "X",
				 ReadProcCnxAdslX, pThisAdapter))
    goto fail;
  else
    level++;
#endif

  pci_set_drvdata(pci_dev_adsl,pThisAdapter);

  return 0;

 fail:
  cnxadsl_fail(level, pThisAdapter);

  return error;
}

LOCAL void __devexit cnxadsl_remove_one(
    struct pci_dev *pci_dev
) {
  const MODULE_FUNCTIONS_T	* pContext;
  int				cnt;
  int				Status;
  CDSL_LINK_T			* pThisLink;
  BOOLEAN			TimerCancelled;
  CDSL_ADAPTER_T		* pThisAdapter;

  pThisAdapter = pci_get_drvdata(pci_dev);

// Note we are going down and then sync with threads
// Note: We don't currently sync but pause which should guarantee they have
// returned back to user space.
  pThisAdapter->shutdown = TRUE;
  SLEEP(ONE_SECOND);

  if (pThisAdapter->AdapterFullyInitialized)
  {
    pThisAdapter->AdapterFullyInitialized = FALSE;

    {
      int which;

      for (which = 0; which < MAX_THREAD; which++)
	stop_kthread(&pThisAdapter->Thread[which]);
    }

    ChipALDisableInterruptHandler(pThisAdapter);

    // close the message handler
    SmSysIfShutdown(pThisAdapter);

    // Shutdown the link
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink && pThisLink->Allocated)
    {
      UtilTmrClearAllTmrs(&pThisAdapter->TimerBlock, pThisLink);
      CMVCShutdownAllVcs(pThisAdapter, pThisLink);
      CMLCallAllModuleLinkShutdowns(pThisLink);
    }

    // Make sure SystemResources were previously allocated
    if (pThisAdapter->AdapterInitComplete)
    {
      //	Cancel 'MainTimer' events associated with this adapter.
      CANCEL_TIMER(&pThisAdapter->MainTimer, &TimerCancelled);
      if (TimerCancelled)
	pThisAdapter->MainTimerState = TMR_CANCELLED;
      else {
	if ((pThisAdapter->MainTimerState == TMR_TIMING) ||
	    (pThisAdapter->MainTimerState == TMR_EXECUTING))
	{
	  pThisAdapter->MainTimerState = TMR_UNLOADING;

	  TimerCancelled = WAIT_EVENT_TIMEOUT(&pThisAdapter->MainTimerStoppedEvent,
				      TIMER_RESOLUTION_5_SEC);
	  if (!TimerCancelled)
	    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			     DBG_FLAG_DONT_CARE,
			     (DEV_LABEL "=>CardMgmt:cnxadsl_remove_one  MainTimer NOT Cancelled (1)!"));
	  else
	    ;//pThisAdapter->MainTimerState = TMR_CANCELLED;
	}
      }
    }
  }

  /**********************************************************************
   **	Call each modules Shutdown function to free memory and 
   **      prepare driver for unloading.
   **********************************************************************/
  if (pThisAdapter->AdapterInitComplete)
  {
    for (cnt = 0; cnt < ARRAY_SIZE(ModemModuleMethods); cnt++)
    {
      pContext = &ModemModuleMethods[cnt];
      if (pContext->Shutdown)
      {
	Status = (pContext->Shutdown)(pThisAdapter);
	if (Status != STATUS_SUCCESS)
	  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			   DBG_FLAG_DONT_CARE,
			   (DEV_LABEL "=>CardMgmt:cnxadsl_remove_one  Status %i returned from the %s, Adapter shutdown function",
			    Status,
			    ModemModuleMethods[cnt].Description));
      }
    }
  }

  if (pThisAdapter->DownloadInitComplete)
  {
    for (cnt = 0; cnt < ARRAY_SIZE(DownloadModuleMethods); cnt++)
    {
      pContext = &DownloadModuleMethods[cnt];

      if (pContext->Shutdown)
      {
	Status = (pContext->Shutdown)(pThisAdapter);
	if (Status != STATUS_SUCCESS)
	  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			   DBG_FLAG_DONT_CARE,
			   (DEV_LABEL "=>CardMgmt:cnxadsl_remove_one  Status %i returned from the %s, AdapterInit function",
			    Status,
			    DownloadModuleMethods[cnt].Description));
      }
    }
  }

  cnxadsl_fail(-1, pThisAdapter);
}

static struct pci_driver cnxadsl_driver = {
	name:		DEV_LABEL,
	id_table:	cnxadsl_pci_tbl,
	probe:		cnxadsl_init_one,
	remove:		__devexit_p(cnxadsl_remove_one),
};

/*++
    cnxadsl_init

Routine Description:
    This is the initial entry point for the Tigris ADSL driver.  It is
    called once when the module is loaded.

Arguments:
    None

Return Value:
    The status of the operation.
--*/

LOCAL int __init cnxadsl_init(
    void
)
{
  int devs;

  //	TRAP();
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:cnxadsl_init (entry point)"));

  gAllocCnt = 0;

#if MEM_ALLOC_DBG
  InitializeListHead(&gAllocList.List);
#endif

  devs = pci_register_driver(&cnxadsl_driver);
#if 0/*CLN*/
  printk(DEV_LABEL " %s: pci_register_driver=>%d\n",
	 __FUNCTION__, devs);
#endif
  if (devs <= 0) {
    pci_unregister_driver(&cnxadsl_driver);
    return -ENODEV;
  }

  return 0;
}

/*++
    cnxadsl_exit

Routine Description:
    This is the exit entry point for the Tigris ADSL driver.  It is
    called once when the module is unloaded. It is responsible for
    terminating any open transfers and cleaning up any resources allocated.

Arguments:
    None

Return Value:
    None
--*/

LOCAL void __exit cnxadsl_exit(
    void
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,/*CLN!!!*/
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:cnxadsl_exit (entry point)"));

  pci_unregister_driver(&cnxadsl_driver);

  // check for mem leaks
#ifdef DBG
  if (gAllocCnt)
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:cnxadsl_exit  Driver had %ld memory leaks",
		      gAllocCnt));
#endif

#if MEM_ALLOC_DBG
  while (!IsListEmpty(&gAllocList.List))
  {
    PALLOC_HEAD_T pList = (PALLOC_HEAD_T)RemoveHeadList(&gAllocList.List);

    if (!pList || pList->Magic != ALLOC_MAGIC)
    {
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>CardMgmt:cnxadsl_exit  Corrupt memory allocation list"));
      break;
    }

    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:cnxadsl_exit  Memory leak ptr 0x%08lx  tag %4s",
		      (DWORD)pList+sizeof(ALLOC_HEAD_T),(char*)&pList->Tag));
  }
#endif
}

module_init(cnxadsl_init);
module_exit(cnxadsl_exit);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif

MODULE_LICENSE("Proprietary");
MODULE_AUTHOR("Conexant Systems Inc");
MODULE_DESCRIPTION("Conexant ADSL ATM driver");

/*++
    CmgmtDeviceClose

Routine Description:
    This function is called when the ATM device is closed.  It
    performs the adapter halt operation.

Arguments:
    pDev - identifies the device to close

Return Value:
    None
--*/

LOCAL void
CmgmtDeviceClose(
    IN struct atm_dev*	pDev
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtDeviceClose"));
}

/*++
    CmgmtOpen

Routine Description:
    This function is called to open a new virtual connection.  It should
    validate the connection parameters and then prepare the connection to 
    send and receive data.

Arguments:
    pVcc - linked list of defined connections
    vpi - virtual path identifier
    vci - virtual connection identifier

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtOpen(
    IN struct atm_vcc*	pVcc
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    ,
    IN short		vpi,
    IN int		vci
#else
#define vpi pVcc->vpi
#define vci pVcc->vci
#endif
)
{
  int Status = STATUS_SUCCESS;
  CDSL_VC_T * pVCStr;
  CDSL_ADAPTER_T* pThisAdapter = pVcc->dev->dev_data;

  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtOpen"));

  // if the device is not connected then fail the request,
  // if we do this then you won't be able to bring up the connection
  // at boot time, but the user will have to do it manually later	
  //	if (pThisAdapter->MediaConnectStatus != MediaConnected)
  //		return -ENOTREADY;
  if (!pThisAdapter->AdapterFullyInitialized)
    return -ENOTREADY;

  // if this VC is partially open then close it to try again
  if (test_bit(ATM_VF_PARTIAL,&pVcc->flags) && pVcc->dev_data)
    CMVCReleaseVc((CDSL_VC_T*)pVcc->dev_data);

  pVcc->dev_data = NULL;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  // find the connection info from the connection list
  Status = atm_find_ci(pVcc,&vpi,&vci);
  if (Status)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtOpen  Error! vpi/vci not found."));
    return Status;
  }
#endif

  if (vci == ATM_VPI_UNSPEC || vpi == ATM_VCI_UNSPEC)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtOpen  Error! invalid vpi/vci"));
    return -EINVAL;
  }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  // save off the connection identifiers
  pVcc->vpi = vpi;
  pVcc->vci = vci;
#endif

  // set the inuse flag for this connection
  set_bit(ATM_VF_ADDR, &pVcc->flags);

  // only AAL0 and AAL5 currently supported
  if (pVcc->qos.aal != ATM_AAL0 && pVcc->qos.aal != ATM_AAL5) {
#if 1/*CLN*/
    printk(DEV_LABEL " %s: Trying to set aal=%i\n",
	   __FUNCTION__, pVcc->qos.aal);
#endif
    return -EINVAL;
  }

  pVCStr = CMVCAllocateVc(pThisAdapter,
			  GET_LINK_FROM_ADAPTER(pThisAdapter),
#if 0 /* not linux */
			  vpi, vci,
#endif
			  pVcc);

  // save the device specific pointer
  pVcc->dev_data = pVCStr;

  // indicate that the vc is ready for use
  set_bit(ATM_VF_READY,&pVcc->flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
  // increment the use count so the module is not unloaded with the vc open
  MOD_INC_USE_COUNT;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#undef vpi
#undef vci
#endif

  return Status;
}

/*++
    CmgmtClose

Routine Description:
    This function is called to close a virtual connection.  It should
    terminate any active transfers and release any resources allocated
    to the connection.

Arguments:
    pVcc - identifies the virtual connection to close

Return Value:
    None
--*/

LOCAL void
CmgmtClose(
    IN struct atm_vcc*	pVcc)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtClose"));

  // if this vc is not open then Linux is confused, so just ignore it.
  if (!pVcc->dev_data)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtClose  Trying to close invalid VC"));
    return;
  }

  // indicate that the VC is not available for use
  clear_bit(ATM_VF_READY,&pVcc->flags);

  // call the vc module to deactivate the vc
  CMVCReleaseVc(pVcc->dev_data);

  // clear the pointer to the device specific data
  pVcc->dev_data = NULL;

  // clear the assignment bit
  clear_bit(ATM_VF_ADDR,&pVcc->flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
  // decrement the use count so the module can be closed
  MOD_DEC_USE_COUNT;
#endif
}

/*++
    CmgmtProcIOCTL

Routine Description:
    This is the command entry point for the Tigris ADSL driver.  It
    is called whenever an application issues an IOCTL request to the
    driver.  This driver uses it to download the link parameters and
    the device firmware.

Arguments:
    pDev - identifies the device for the command
    Command - command code
    pArgument - send or receive data associated with the command

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtProcIOCTL(
    IN struct atm_dev*	pDev,
    IN UINT		Command,
    INOUT PVOID		pArgument
) {
  int			Status = STATUS_SUCCESS;
  int 			cnt;
  CDSL_ADAPTER_T*	pThisAdapter = pDev->dev_data;

#if 0
  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   pThisAdapter->DebugFlag,
		   (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  Command %x %p",
		    Command, pArgument));
#endif

  if (pThisAdapter->shutdown)  // if shutting down blow it off
  {
    // Note this occurs normally as the user space threads have no way of
    // knowing of the shutdown other than sending an IOCTL and it failing
    // They should quit after the first failure.
    Status = ERR_SHUTDOWN;
  } else {
    switch(Command-TIG_0) {
    case TIG_GET_PRINT_BUFF-TIG_0:
    case TIG_GET_DEBUG_DATA-TIG_0:
      Status = ARMALIoctl(pThisAdapter, Command, pArgument);
      if (Status != STATUS_SUCCESS)
	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 DBG_FLAG_DONT_CARE,
			 (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  ARM command failed, command=%d, Status=0x%x",
			  Command,Status));
      break;
    case TIG_LOAD_START-TIG_0:
    case TIG_LOAD_RECORD-TIG_0:
    case TIG_LOAD_DONE-TIG_0:  // if the IO command is a load record type then send it to the ARM abstraction module
      if (!pThisAdapter->AdapterFullyInitialized)
      {
	Status = ARMALIoctl(pThisAdapter, Command, pArgument);
	if (Status != STATUS_SUCCESS)
	    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			     DBG_FLAG_DONT_CARE,
			     (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  ARM Load command failed, command=%d, Status=0x%x",
			      Command,Status));
      }
      break;
    case TIG_START_DSL-TIG_0:
      if (!pThisAdapter->AdapterFullyInitialized)
      {
	// init the message system
	SmSysIfInit(pThisAdapter);

	// tell the chip al module to initialize services
	Status = ChipALInitServices(pThisAdapter);

	for (cnt = 0; cnt < ARRAY_SIZE(ModemModuleMethods); cnt++)
	  if (ModemModuleMethods[cnt].AdapterInit)
	  {
	    Status = ModemModuleMethods[cnt].AdapterInit(pThisAdapter);
	    if (Status != STATUS_SUCCESS)
	    {
	      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			       DBG_FLAG_DONT_CARE,
			       (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  Module init fail %d 0x%x",
				cnt,Status));
	      return Status;
	    }
	  }

	INIT_TIMER(&pThisAdapter->MainTimer,
		   CmgmtMainTimerInterrupt,
		   (DWORD)pThisAdapter);

	pThisAdapter->MainTimerState = TMR_INITIALIZED;

	INIT_SPIN_LOCK(&pThisAdapter->Lock);
	pThisAdapter->AdapterInitComplete = TRUE;

	Status = CmgmtAdapterReset(pThisAdapter);
	if (Status != STATUS_SUCCESS)
	{
	  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			   DBG_FLAG_DONT_CARE,
			   (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  function CmgmtAdapterReset failed Status=0x%x",
			    Status));
	  return STATUS_FAILURE;
	}

	//	start back ground loop
	INIT_EVENT(&pThisAdapter->MainTimerStoppedEvent);
	START_TIMER(&pThisAdapter->MainTimer, pThisAdapter->TimerResolution);
	pThisAdapter->MainTimerState = TMR_TIMING;

	//	Automatically Start (everything necessary) the ADSL
	CDALSetLineState(pThisAdapter,
			 pThisAdapter->PersistData.LineAutoActivation ?
			 CDAL_LINE_STATE_PERSIST_INITIALIZE:CDAL_LINE_STATE_INT);

	//	Start the kernel threads
	{
	  int which;

	  for (which = 0; which < MAX_THREAD; which++) {
	    pThisAdapter->Thread[which].which = which;
	    pThisAdapter->Thread[which].arg = pThisAdapter;

	    start_kthread(cardALDeviceThread, &pThisAdapter->Thread[which]);
	  }
	}

	pThisAdapter->AdapterFullyInitialized = TRUE;

	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 pThisAdapter->DebugFlag,
			 (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  TIG_START_DSL command completed successfully"));
      }
      break;
    case TIG_SET_PARAMS-TIG_0:  // if the record is user configurable parameters then save them off
      if (pThisAdapter->DownloadInitComplete)
	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 DBG_FLAG_DONT_CARE,
			 (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  TIG_SET_PARAMS command, DownloadInitComplete already set"));
      else {
	DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			 DBG_FLAG_DONT_CARE,
			 (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  TIG_SET_PARAMS command"));

	// load the configuration parameters into the adapter structure
	if (copy_from_user(&pThisAdapter->PersistData, pArgument,
			  offsetof(TIG_USER_PARAMS, CnxtVendorId)))
	  return -EFAULT;

#ifdef DBG
	CardALDebugFlag = pThisAdapter->PersistData.DebugFlag;
#endif

	// run the configuration init on each module
	for (cnt = 0; cnt < ARRAY_SIZE(DownloadModuleMethods); cnt++)
	  if (DownloadModuleMethods[cnt].CfgInit)
	    DownloadModuleMethods[cnt].CfgInit(pThisAdapter,
					       &pThisAdapter->PersistData);

	for (cnt = 0; cnt < ARRAY_SIZE(ModemModuleMethods); cnt++)
	  if (ModemModuleMethods[cnt].CfgInit)
	    ModemModuleMethods[cnt].CfgInit(pThisAdapter,
					    &pThisAdapter->PersistData);

	// Initialize the download modules.  The modem modules will be
	// initialized after the download is complete.
	for (cnt = 0; cnt < ARRAY_SIZE(DownloadModuleMethods); cnt++)
	  if (DownloadModuleMethods[cnt].AdapterInit)
	  {
	    Status = DownloadModuleMethods[cnt].AdapterInit(pThisAdapter);
	    if (Status != STATUS_SUCCESS)
	    {
	      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
			       DBG_FLAG_DONT_CARE,
			       (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  Module init fail %d %x",
				cnt,Status));
	      goto ret;
	    }
	  }

	// mark the download modules as initialized
	pThisAdapter->DownloadInitComplete = TRUE;

      ret:
	{
	  DWORD myStatus=Status;

	  if (copy_to_user(&((PTIG_USER_PARAMS)pArgument)->CommandStatus,
			  &myStatus,
			  sizeof(myStatus)))
	    return -EFAULT;
	}
      }
      break;
    case TIG_DEVICE_SPEC-TIG_0:  // is it a device specific request
      {
	TIG_DEVICE_SPEC_T *pCommand = pArgument;

	// loop through all of the modules to let each have a
	// crack at it.  Stop at the first one that handles it

	// check the card managment module
	Status = cmCtlDeviceSpecific(pThisAdapter,
				     (BACK_DOOR_T*)&pCommand->BackDoorBuf,
				     &pCommand->RequiredSize,
				     &pCommand->ReturnSize);
	if (Status != STATUS_UNSUPPORTED)
	  goto onedone;

	// run the ioctl handler on each module
	for (cnt = 0; cnt < ARRAY_SIZE(DownloadModuleMethods); cnt++)
	  if (DownloadModuleMethods[cnt].IoctlHdlr){
	    Status = DownloadModuleMethods[cnt].IoctlHdlr(
					pThisAdapter,
					(BACK_DOOR_T*)&pCommand->BackDoorBuf,
					&pCommand->RequiredSize,
					&pCommand->ReturnSize);
	    if (Status != STATUS_UNSUPPORTED)
	      goto onedone;
	  }

	for (cnt = 0; cnt < ARRAY_SIZE(ModemModuleMethods); cnt++)
	  if (ModemModuleMethods[cnt].IoctlHdlr){
	    Status = ModemModuleMethods[cnt].IoctlHdlr(
					pThisAdapter,
					(BACK_DOOR_T*)&pCommand->BackDoorBuf,
					&pCommand->RequiredSize,
					&pCommand->ReturnSize);
	    if (Status != STATUS_UNSUPPORTED)
	      goto onedone;
	  }

      onedone:
	;
      }
      break;
    case TIG_IS_SHOWTIME-TIG_0:  // see if its a showtime status request
      {
	DWORD myShowtime=pThisAdapter->Showtime;

	if (copy_to_user(pArgument,&myShowtime,sizeof(myShowtime)))
	  return -EFAULT;
      }
      break;
    case TIG_GET_LOG-TIG_0:  // see if its a log record request
      Status = CdalDiagGetTxtLog(pThisAdapter, pArgument, 256);
      break;
    case TIG_CLEAR_STATS-TIG_0:      // see if its a log record request
      FrameALClearStats(GET_LINK_FROM_ADAPTER(pThisAdapter));
      break;
    default:      // otherwise its not a valid command so err it
      Status = -EBADRQC;
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL "=>CardMgmt:CmgmtProcIOCTL  Unknown IOCTL command, command=%d Status=0x%x",
			Command,Status));
      break;
    }
  }

  return Status;
}

/*++
    CmgmtGetSocketOption

Routine Description:
    This function is called to get option settings for a 
    given connection.  We don't currently have any readable
    options.

Arguments:
    pVcc - identifies the virtual connection
    OptionLevel - level of the option to get
    OptionName - name (actually number) of the option to get
    pOptionValue - pointer to value to get the option
    OptionLength - length of the option value field

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtGetSocketOption(
    IN struct atm_vcc*	pVcc,
    IN int		OptionLevel,
    IN int		OptionName,
    OUT PVOID		pOptionValue,
    IN int		OptionLength
)
{
  int Status = STATUS_SUCCESS;

  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtGetSocketOption  pVcc %p level %d  name %d  lth %d",
		   pVcc, OptionLevel, OptionName, OptionLength));

  return Status;
}

/*++
    CmgmtSetSocketOption

Routine Description:
    This function is called to set an option value
    for a specified connection.  There are currently not
    any settable option values.

Arguments:
    pVcc - identifies the virtual connection
    OptionLevel - level of the option to set
    OptionName - name (actually number) of the option to set
    pOptionValue - pointer to value to set the option to
    OptionLength - length of the option value field

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtSetSocketOption(
    IN struct atm_vcc*	pVcc,
    IN int		OptionLevel,
    IN int		OptionName,
    IN PVOID		OptionValue,
    IN int		OptionLength
)
{
  int Status = STATUS_SUCCESS;

  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtSetSocketOption"));

  return Status;
}

/*++
    CmgmtSendBuffer

Routine Description:
    This function is called to transmit a buffer to the 
    ATM device.

Arguments:
    pVcc - identifies the connection to use for the send
    pSkb - transmit buffer identifier

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtSendBuffer(
    IN struct atm_vcc*	pVcc,
    IN SK_BUFF_T*	pSkb
)
{
  int Status;
  CDSL_VC_T * pVCStr;

  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         DBG_FLAG_DONT_CARE,
  //		         (DEV_LABEL "=>CardMgmt:CmgmtSendBuffer"));

  // get the device vc pointer
  pVCStr = (CDSL_VC_T *)pVcc->dev_data;

  // if the vcc is not enabled then fail
  if (!pVCStr->Connected)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtSendBuffer  Attempt to send on an inactive VC"));
    RELEASE_SKB(pVcc, pSkb);
    return -EINVAL;
  }

  // pass it to the Buffer management module to be transmitted.
  // Status pending means it was successfully queued so tell
  // Linux it was successful.
  Status = BufMgmtSend(pVCStr, pSkb);
  if (Status == COMMAND_PENDING)
    Status = STATUS_SUCCESS;

  return Status;
}

#if 0
/*++
    CmgmtPhyPut

Routine Description:
    This function is called by the Linux ATM kernel to 
    write directly to the PHY module.  This driver does
    not currently use this function.

Arguments:
    pDev - selects the modem device
    value - value to put
    addr - where to put the value

Return Value:
    None
--*/

LOCAL void
CmgmtPhyPut(
    IN struct atm_dev*	pDev,
    IN UCHAR		value,
    IN ULONG		addr
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtPhyPut"));
}

/*++
    CmgmtPhyGet

Routine Description:
    This function is called to get information directly
    from the physical device.  It is currently not 
    implemented for this driver.

Arguments:
    pDev - selects the device to get from
    addr - where to get the data

Return Value:
    -ENOSYS - function not implemented
--*/

LOCAL UCHAR
CmgmtPhyGet(
    IN struct atm_dev*	pDev,
    OUT ULONG		addr
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtPhyGet"));

  return -ENOSYS;
}

/*++
    CmgmtChangeQOS

Routine Description:
    This function is called to adjust the QOS parameters for
    an already open connection.

Arguments:
    pVcc - pointer to the connection to adjust 
    pQOS - new quality of service parameters
    flags - flag word

Return Value:
    None
--*/

LOCAL int
CmgmtChangeQOS(
    IN struct atm_vcc*	pVcc,
    IN struct atm_qos*	pQOS,
    IN int		flags
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtChangeQOS  is not implemented"));

  return -ENOSYS;
}
#endif

LOCAL const char *AalType2Str(int AalType) {
#if 0 /* not linux */
  static const char *tytyu[BD_TESTMOD_AAL_TYPE_END] = {"0", "1", "34", "5"};

  return (AalType >= ARRAY_SIZE(tytyu)) ? "?" : tytyu[AalType];
#else
  const char *s = "?";

  switch (AalType) {
  case ATM_AAL0:
    s = "0";
    break;
  case ATM_AAL1:
    s = "1";
    break;
  case ATM_AAL2:
    s = "2";
    break;
  case ATM_AAL34:
    s = "34";
    break;
  case ATM_AAL5:
    s = "5";
    break;
  default:
    break;
  }

  return s;
#endif
}

/*++
    CmgmtProcessRead

Routine Description:
	This function is called when a process reads from the
	modem device created by the Linux ATM module.  Commonly
	it is a cat of the device by the operator.  So it is
	used to display general driver information and statistics.

Arguments:
    pDev - identifies the device  
    pProcPageOffset - 
    ProcPage -

Return Value:
    -ENOSYS - function not implemented
--*/

LOCAL int
CmgmtProcessRead(
    IN struct atm_dev	*pDev,
    IN loff_t		*pPOS,
    OUT char		*ProcPage
)
{
  CDSL_ADAPTER_T	*pThisAdapter = pDev->dev_data;
  int			State = *pPOS;
  int			RetBytes = 0;
  char Buffer[sizeof(TIG_DEVICE_SPEC_T) + sizeof(BACK_DOOR_T)];
  TIG_DEVICE_SPEC_T	*pDevSpecific = (TIG_DEVICE_SPEC_T*)Buffer;
  BACK_DOOR_T		*pBackDoorBuf = (BACK_DOOR_T*)pDevSpecific->BackDoorBuf;
  int			Index;

#if 0
  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   pThisAdapter->DebugFlag,
		   (DEV_LABEL "=>CardMgmt:CmgmtProcessRead"));
#endif

  ///////////////////////////////////////////////////////////////////////
  // General information
  ///////////////////////////////////////////////////////////////////////
  if (!State--)
  {  
    // print out the header information
    RetBytes += sprintf(ProcPage+RetBytes,
			"\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			"Conexant AccessRunner PCI ADSL Modem Adapter Status\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			"---------------------------------------------------\n\n");

    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_CARDMGMT_GET_CONNECT_STATUS;

    // read the line statistics
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    // report the connection status
    if (pThisAdapter->Showtime)
    {
      BD_ADSL_LINE_STATUS_T* pLineStat = &pBackDoorBuf->Params.BdAdslLineStatus;

      // init the statistics structure
      CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
      pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
      pBackDoorBuf->ReqCode = BD_ADSL_GET_LINE_STATUS;

      // read the line statistics
      CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

      RetBytes += sprintf(ProcPage+RetBytes,
			  "ADSL Line Connected\n");
      RetBytes += sprintf(ProcPage+RetBytes,
			  "Line Rates:   Receive %ld kbps     Transmit %ld kbps\n\n",
			  pLineStat->LineSpeedDownOrNear,
			  pLineStat->LineSpeedUpOrFar);
    } else {
      RetBytes += sprintf(ProcPage+RetBytes,
			  "ADSL Line Not Connected\n");
      RetBytes += sprintf(ProcPage+RetBytes,
			  "Line Rates: Not available\n\n");
    }

    // display the modulation
    {
      BD_ADSL_CFG_T		   BdCfg;

      if (CdalDiagGetAdslConfig(pThisAdapter, (char *)&BdCfg, sizeof(BdCfg)) == RESULT_SUCCESS) {
	const char *msg = "";

	switch (BdCfg.AdslDpRateMode)
	{
	case BD_ADSL_DP_RATE_G_DOT_DMT:
	  msg = "G.DMT, Rate Unlimited  /  Full Rate";
	  break;

	case BD_ADSL_DP_RATE_G_DOT_LITE:
	  msg = "G.Lite, Rate Limited (1.5 Mbps)";
	  break;

	case BD_ADSL_DP_RATE_T1_DOT_413:
	  msg = " T1.413  Issue 2";
	  break;

	case BD_ADSL_DP_RATE_NOT_AVAIL:
	default:
	  msg = "Not Available";
	  break;
	}

	RetBytes += sprintf(ProcPage+RetBytes,
			    "ADSL Modulation:");
	RetBytes += sprintf(ProcPage+RetBytes,
			    msg);
	RetBytes += sprintf(ProcPage+RetBytes,
			    "\n\n");
      }
    }

    // display the connection info
    {
      int i = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
      struct sock *s;

      read_lock(&vcc_sklist_lock);

      {
	struct hlist_node *node;
	int ii;

	for (ii = 0; ii < VCC_HTABLE_SIZE; ii++) {
	  struct hlist_head *head = &vcc_hash[ii];

	  sk_for_each(s, node, head) {
	    struct atm_vcc* pVcc = atm_sk(s);

	    if (pVcc->dev == pDev) {
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
	      struct atm_vcc* pVcc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
	      unsigned long flags;

	      spin_lock_irqsave(&pDev->lock, flags);
#endif
	      for (pVcc = pDev->vccs; pVcc; pVcc = pVcc->next) {
#else
		struct sock *s;

		read_lock(&vcc_sklist_lock);

		for (s = vcc_sklist; s; s = s->next) {
		  struct atm_vcc* pVcc = s->protinfo.af_atm;

		  if (pVcc->dev == pDev) {
#endif
#endif
		    RetBytes += sprintf(ProcPage+RetBytes,
					"%s VPI: %d   VCI: %d\n",
					(i == 0) ?
					"ATM Virtual Channel IDs:":
					"                        ",
					pVcc->vpi,pVcc->vci);
		    i++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
		  }
		}
	      }
	    }

	    read_unlock(&vcc_sklist_lock);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
	  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
	  spin_unlock_irqrestore(&pDev->lock, flags);
#endif
#else
	}
      }
      read_unlock(&vcc_sklist_lock);
#endif
#endif

      if (i == 0)
	RetBytes += sprintf(ProcPage+RetBytes,
			    "ATM Virtual Channel IDs: no connections open\n");
      RetBytes += sprintf(ProcPage+RetBytes,
			  "\n");
    }
    return RetBytes;
  }

  ///////////////////////////////////////////////////////////////////////
  // Physical layer statistics
  ///////////////////////////////////////////////////////////////////////
  if (!State--)
  {
    BD_CARDAL_ADSL_TRANSCEIVER_STATUS_T *pStats = &pBackDoorBuf->Params.AdslTransceiverStatus;

    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_CARDAL_GET_ADSL_TRANSCEIVER_STATUS;

    // read the line statistics
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

#define MAN(v,n) (((v) < 0) ? -((-(v))>>n) : (v)>>n)
#define FRAC2(v,n) ((((((v) < 0) ? -(v) : (v)) & ((1 << n) - 1))*100)/(1 << n))

    RetBytes += sprintf(ProcPage+RetBytes,
"Physical Layer Stats:      Receive                   Transmit\n");

    // print the snr values, they are in 8.8 format
    RetBytes += sprintf(ProcPage+RetBytes,
" Noise Margin      %3d.%02d/%3d.%02d: %3d           %3d.%02d/%3d.%02d: %3d\n",
			MAN(pStats->SNR_Margin[DWN_s].Mar_Cur,8),
			FRAC2(pStats->SNR_Margin[DWN_s].Mar_Cur,8),
			MAN(pStats->SNR_Margin[DWN_s].Mar_Min,8),
			FRAC2(pStats->SNR_Margin[DWN_s].Mar_Min,8),
			pStats->SNR_Margin[DWN_s].Mar_Min_Bin,

			MAN(pStats->SNR_Margin[UP_s].Mar_Cur,8),
			FRAC2(pStats->SNR_Margin[UP_s].Mar_Cur,8),
			MAN(pStats->SNR_Margin[UP_s].Mar_Min,8),
			FRAC2(pStats->SNR_Margin[UP_s].Mar_Min,8),
			pStats->SNR_Margin[UP_s].Mar_Min_Bin);

    // print the attenuation values, they are in 7.9 format
    RetBytes += sprintf(ProcPage+RetBytes,
" Attenuation             %3d.%02d                    %3d.%02d\n",
			MAN(pStats->Attn[DWN_s],9),
			FRAC2(pStats->Attn[DWN_s],9),

			MAN(pStats->Attn[UP_s],9),
			FRAC2(pStats->Attn[UP_s],9));

    // print the transmit power, there is no receive power
    RetBytes += sprintf(ProcPage+RetBytes,
" Power                                          %9d\n",
			pStats->Tx_Power);

    if (0) {
      RetBytes += sprintf(ProcPage+RetBytes,
			  "\n");
      RetBytes += sprintf(ProcPage+RetBytes,
			  "Date: %04u%02u%02u (%u)\n",
			  pStats->Date.year,pStats->Date.month,pStats->Date.day,
			  pStats->Date.dayofweek);
      RetBytes += sprintf(ProcPage+RetBytes,
			  "Time: %02u:%02u:%02u\n",
			  pStats->Time.hr,pStats->Time.min,pStats->Time.sec);
      RetBytes += sprintf(ProcPage+RetBytes,
			  "State: Rx Tx BG %u/%u/%u\n",
			  pStats->Receive_State,pStats->Transmit_State,pStats->Process_State);
      RetBytes += sprintf(ProcPage+RetBytes,
			  "Bits per frame: %u %u\n",
			  pStats->Bits_Per_Frame[DWN_s],
			  pStats->Bits_Per_Frame[UP_s]);

      RetBytes += sprintf(ProcPage+RetBytes,
			  "Startup attempts: %u\n",
			  pStats->Startup_Attempts);
    }
    return RetBytes;
  }

  ///////////////////////////////////////////////////////////////////////
  // Frame abstraction layer statistics
  ///////////////////////////////////////////////////////////////////////
  if (!State--)
  {
    BD_FRAMEAL_ATM_STATS_T *pStats = &pBackDoorBuf->Params.BdFrameALAtmStats;

    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_FRAMEAL_ATM_GET_STATS;
    pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = BD_ATM_STATS_INDEX_FOR_LINK;

    // read the per vc statistics
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    RetBytes += sprintf(ProcPage+RetBytes,
			"\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			"ATM Link Stats:      Receive           Transmit\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			" Bytes            %10lu         %10lu\n",
			pStats->NumRxBytes,
			pStats->NumTxBytes);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Cells            %10lu         %10lu\n",
			pStats->NumRxCells,
			pStats->NumTxCells);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Management Cells %10lu         %10lu\n",
			pStats->NumRxMgmtCells,
			pStats->NumTxMgmtCells);
    RetBytes += sprintf(ProcPage+RetBytes,
			" CLP=0 Cells      %10lu         %10lu\n",
			pStats->NumRxClpEqual0Cells,
			pStats->NumTxClpEqual0Cells);
    RetBytes += sprintf(ProcPage+RetBytes,
			" CLP=1 Cells      %10lu         %10lu\n",
			pStats->NumRxClpEqual1Cells,
			pStats->NumTxClpEqual1Cells);
    RetBytes += sprintf(ProcPage+RetBytes,
			" HEC Errors       %10lu\n",
			pStats->NumRxHecErrs);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Alignment Errors %10lu\n",
			pStats->NumRxCellAlignErrs);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Unroutable Cells %10lu\n",
			pStats->NumRxUnroutCellErrs);
    return RetBytes;
  }

  ///////////////////////////////////////////////////////////////////////
  // AAL layer statistics
  ///////////////////////////////////////////////////////////////////////
  if (!State--)
  {
    BD_FRAMEAL_AAL_STATS_T *pStats = &pBackDoorBuf->Params.BdFrameALAalStats;

    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_FRAMEAL_AAL_GET_STATS;
    pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = BD_AAL_STATS_INDEX_FOR_LINK;

    // read the per vc statistics
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    RetBytes += sprintf(ProcPage+RetBytes,
			"\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			"AAL Link Stats:          Receive           Transmit\n");
    RetBytes += sprintf(ProcPage+RetBytes,
			" Good Bytes           %10lu         %10lu\n",
			pStats->NumRxGoodBytes,
			pStats->NumTxGoodBytes);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Good Frames          %10lu         %10lu\n",
			pStats->NumRxGoodFrames,
			pStats->NumTxGoodFrames);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Discarded Bytes      %10lu\n",
			pStats->NumRxDiscardedBytes);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Discarded Frames     %10lu\n",
			pStats->NumRxDiscardedFrames);
    RetBytes += sprintf(ProcPage+RetBytes,
			" CRC Errors           %10lu\n",
			pStats->NumRxCrcErrs);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Invalid Length Errors%10lu\n",
			pStats->NumRxInvalidLenErrs);
    RetBytes += sprintf(ProcPage+RetBytes,
			" Timeout Errors       %10lu\n\n",
			pStats->NumRxTimeoutErrs);
    return RetBytes;
  }

  ///////////////////////////////////////////////////////////////////////
  // display ATM statistics per VC
  ///////////////////////////////////////////////////////////////////////
  for (Index=0; Index<MAX_VC_PER_LINK; Index++)
  {
    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_FRAMEAL_ATM_GET_STATS;
    pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = Index;

    // read the per vc statistics 
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
    {
      if (!State--)
      {
	BD_FRAMEAL_ATM_STATS_T *pStats = &pBackDoorBuf->Params.BdFrameALAtmStats;
	RetBytes += sprintf(ProcPage+RetBytes,
"ATM Channel Stats:       Receive           Transmit VPI: %lu VCI: %lu\n",
			    pStats->Vpi,pStats->Vci);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Bytes                %10lu         %10lu\n",
			    pStats->NumRxBytes,
			    pStats->NumTxBytes);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Cells                %10lu         %10lu\n",
			    pStats->NumRxCells,
			    pStats->NumTxCells);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Mgmt Cells           %10lu         %10lu\n",
			    pStats->NumRxMgmtCells,
			    pStats->NumTxMgmtCells);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " CLP=0 Cells          %10lu         %10lu\n",
			    pStats->NumRxClpEqual0Cells,
			    pStats->NumTxClpEqual0Cells);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " CLP=1 Cells          %10lu         %10lu\n",
			    pStats->NumRxClpEqual1Cells,
			    pStats->NumTxClpEqual1Cells);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " HEC Errors           %10lu\n",
			    pStats->NumRxHecErrs);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Alignment Errors     %10lu\n",
			    pStats->NumRxCellAlignErrs);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Unroutable Cells     %10lu\n\n",
			    pStats->NumRxUnroutCellErrs);
	return RetBytes;
      }
    }
  }

  ///////////////////////////////////////////////////////////////////////
  // display AAL statistics per VC
  ///////////////////////////////////////////////////////////////////////
  for (Index=0; Index<MAX_VC_PER_LINK; Index++)
  {
    // init the statistics structure
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_FRAMEAL_AAL_GET_STATS;
    pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = Index;

    // read the per vc statistics
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
    {
      if (!State--)
      {
	BD_FRAMEAL_AAL_STATS_T *pStats = &pBackDoorBuf->Params.BdFrameALAalStats;

	RetBytes += sprintf(ProcPage+RetBytes,
"AAL%s Channel Stats:      Receive           Transmit VPI: %lu VCI: %lu\n",
			    AalType2Str(pStats->AalType),
			    pStats->Vpi,
			    pStats->Vci);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Good Bytes           %10lu         %10lu\n",
			    pStats->NumRxGoodBytes,
			    pStats->NumTxGoodBytes);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Good Frames          %10lu         %10lu\n",
			    pStats->NumRxGoodFrames,
			    pStats->NumTxGoodFrames);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Discarded Bytes      %10lu\n",
			    pStats->NumRxDiscardedBytes);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Discarded Frames     %10lu\n",
			    pStats->NumRxDiscardedFrames);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " CRC Errors           %10lu\n",
			    pStats->NumRxCrcErrs);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Invalid Length Errors%10lu\n",
			    pStats->NumRxInvalidLenErrs);
	RetBytes += sprintf(ProcPage+RetBytes,
			    " Timeout Errors       %10lu\n\n",
			    pStats->NumRxTimeoutErrs);
	return RetBytes;
      }
    }
  }

  if (!State--)
  {
    BD_MINI_DATA_BLOCK_T *pProdVer = &pBackDoorBuf->Params.BdCardMgmtProdVer;

    // print the driver version
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_CARDMGMT_GET_PRODUCT_VERSION;

    // get the driver version
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    // output it
    RetBytes += sprintf(ProcPage+RetBytes,
			"Driver version - %s\n",
			(UCHAR*)pProdVer);

    // print the mac address
    CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
    pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
    pBackDoorBuf->ReqCode = BD_CARDMGMT_GET_EEPROM_MAC_ADDR;

    // get the mac address
    CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

    RetBytes += sprintf(ProcPage+RetBytes,
			"EE PROM Mac Address %02x.%02x.%02x.%02x.%02x.%02x\n\n",
			pBackDoorBuf->Params.BdCardMgmtMacAddr[0],
			pBackDoorBuf->Params.BdCardMgmtMacAddr[1],
			pBackDoorBuf->Params.BdCardMgmtMacAddr[2],
			pBackDoorBuf->Params.BdCardMgmtMacAddr[3],
			pBackDoorBuf->Params.BdCardMgmtMacAddr[4],
			pBackDoorBuf->Params.BdCardMgmtMacAddr[5]);
    return RetBytes;
  }

  return RetBytes;
}

/*++
    CmgmtPhyStart

Routine Description:
    Function called to start the physical device.  Currently not used.

Arguments:
    pDev - identifies the device.

Return Value:
    The status of the operation.
--*/

LOCAL int
CmgmtPhyStart(
    IN struct atm_dev*	pDev
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtPhyStart"));

  return STATUS_SUCCESS;
}

/*++
    CmgmtPhyIOCTL

Routine Description:
    This function is called to pass commands directly to the 
    phy module.  This is not used by the driver.

Arguments:
    pDev  - identifies the device
    cmd - command to execute
    arg - data associated with the command

Return Value:
    -ENOSYS - function not implemented
--*/

LOCAL int
CmgmtPhyIOCTL(
    IN struct atm_dev*	pDev,
    IN unsigned int	cmd,
    INOUT void*		arg
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtPhyIOCTL"));

  return -ENOSYS;
}

/*++
    CmgmtPhyInterrupt

Routine Description:
    This function is called to generate an interrupt to
    the physical device.  Not used by this driver.

Arguments:
    pDev - identifies the device

Return Value:
    None
--*/

LOCAL void
CmgmtPhyInterrupt(
    IN struct atm_dev*	pDev
)
{
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtPhyInterrupt"));
}

/******************************************************************************
*FUNCTION NAME:
*	CmgmtIndicateLinkUp
*
*ABSTRACT:
*	This function is called by the chip abstraction layer to indicate
*	that the ADSL link is up.
*
*RETURN:
*
*DETAILS:
******************************************************************************/
void CmgmtIndicateLinkUp(
    IN CDSL_ADAPTER_T		* pThisAdapter)
{
  CDSL_LINK_T		* pThisLink;

  //	Indicate showtime
  pThisAdapter->Showtime = TRUE;
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   DBG_FLAG_DONT_CARE,
		   (DEV_LABEL "=>CardMgmt:CmgmtIndicateLinkUp  - It's SHOWTIME!!!"));

  ChipALLinkEnable(pThisAdapter,ADSL_CHANNEL);

  // inform FrameAL to get one cell sent
  pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
  FrameALLinkUp(pThisLink);
}

/******************************************************************************
*FUNCTION NAME:
*	CmgmtIndicateLinkDown
*
*ABSTRACT:
*	This function is called by the chip abstraction layer to indicate
*	that the ADSL link is down.
*
*RETURN:
*
*DETAILS:
******************************************************************************/
void CmgmtIndicateLinkDown(
    IN CDSL_ADAPTER_T		* pThisAdapter)
{
  CDSL_LINK_T		* pThisLink;

  pThisAdapter->Showtime = FALSE;
  DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		   pThisAdapter->DebugFlag,
		   (DEV_LABEL "=>CardMgmt:CmgmtIndicateLinkDown  - Show's over! :("));

  ChipALLinkDisable(pThisAdapter,ADSL_CHANNEL);

  pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
  FrameALLinkDown(pThisLink);
}

/******************************************************************************
*FUNCTION NAME:
*	CmgmtAdapterReset
*
*ABSTRACT:
*	Reset the adapter and driver.
*
*RETURN:
*	status of initialization and setup
*
*DETAILS:
******************************************************************************/
LOCAL NDIS_STATUS CmgmtAdapterReset(
    IN CDSL_ADAPTER_T	* pThisAdapter)
{
  CDSL_LINK_T				* pThisLink;
  NDIS_STATUS				Status;

  //Reset link data structures for all links
  Status = CMLInitAllLinks(pThisAdapter);
  if (Status != STATUS_SUCCESS)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtAdapterReset  ERROR;  CMLInitAllLinks Failure!"));
    return Status;
  }

  //	Allocate the links (lines, line devices) for this adapter.
  pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
  if (pThisLink)
    pThisLink = CMLAllocateLink(pThisAdapter);

  //Reset VC data structures for all VCs
  pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
  Status = CMVCInitAllVcs(pThisAdapter, pThisLink);
  if (Status != STATUS_SUCCESS)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtAdapterReset  ERROR;  CMVCInitAllVcs Failure!"));
    return Status;
  }

  Status = CMLCallAllModuleLinkInits(pThisAdapter, pThisLink);
  if (Status != STATUS_SUCCESS)
  {
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtAdapterReset  ERROR;  CMLCallAllModuleLinkInits Failure!"));
    return Status;
  }

  //
  //Reset Timer Info
  //
  UtilTmrResetTmr(&pThisAdapter->TimerBlock, MAX_TIMERS);

  return Status;
}

/******************************************************************************
*FUNCTION NAME:
*	CmgmtMainTimerInterrupt
*
*ABSTRACT:
*	Function called back from Linux (registering the Miniport driver
*	initialized a timer with this call-back function).
*	This function is called by Linux periodically (whenever the timer 
*	period registered/initalized expires).
*
*RETURN:
*
*DETAILS:
*	This timer function calls UtilTmrTmrBackground which handles multiple 
*	timers for the miniport driver.
*	The timer is restarted before returning to NDIS.
******************************************************************************/
LOCAL VOID CmgmtMainTimerInterrupt(
    IN 	ULONG	 pAdapt
)
{
  CDSL_ADAPTER_T	* pThisAdapter = (CDSL_ADAPTER_T *)pAdapt;
  LOCK_FLAG(LockFlag);

  if (pThisAdapter->AdapterStructureAllocated)
  {
    if (pThisAdapter->MainTimerState == TMR_TIMING)
    {
      pThisAdapter->MainTimerState = TMR_EXECUTING;

      ACQUIRE_LOCK(&pThisAdapter->Lock, LockFlag);

      UtilTmrTmrBackground(&pThisAdapter->TimerBlock, pThisAdapter->TimerResolution);

      RELEASE_LOCK(&pThisAdapter->Lock, LockFlag);

      //	Normally this line would be used
      //	pThisAdapter->MainTimerState = TMR_COMPLETED;
      //	Instead, reset the timer to fire
      START_TIMER(&pThisAdapter->MainTimer, pThisAdapter->TimerResolution);
      pThisAdapter->MainTimerState = TMR_TIMING;
    } else // Timer is being shutdown or stopped
      SET_EVENT(&pThisAdapter->MainTimerStoppedEvent);
  } else
    DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		     DBG_FLAG_DONT_CARE,
		     (DEV_LABEL "=>CardMgmt:CmgmtMainTimerInterrupt  Adapter Structure NOT Allocated!"));
}

/******************************************************************************
FUNCTION NAME:
	cmCtlDeviceSpecific

ABSTRACT:
	This function coordinates all of device specific
	data requeest for actions that are performed for 
	Card Management module.

DETAILS:
	EVERY MODULE WHICH SUPPORTS *ANY* BACKDOOR ACTION
	(EVEN ONE) *MUST* IMPLEMENT THIS FUNCTION
	IN AT LEAST A SIMILAR MANNER OR FASHION!!!!!!!!

	BACKDOOR CODE/SUBCODE VALUES *WILL BE DESIGNED*
	TO **GUARANTEE** THAT EACH CODE/SUBCODE VALUE
	WILL BE DIRECTED TO *ONLY ONE* MODULE!!

	NOTE:	The BackDoor Status should ONLY BE SET TO FAILURE
		at the beginning of BackDoor parsing by the first
		OID parser (CardMgmt)!!!
		ALL OTHER successor OID parsing of BackDoor should
		only set the BackDoor Status when it is intended
		to be set to a specific Status (i.e., BAD_SIGNATURE,
		FAILURE, or SUCCESS)!!  The BackDoor Status should
		not be set at the beginning of BackDoor parsing
		intending to change it if needed!!!
		This is to keep other modules (BackDoor parsing)
		from overwriting the BackDoor status once a module
		has parsed the BackDoor structure and acted upon
		the data!!!
		If the BackDoor Status is not set to CODE_NOT_SUPPORTED,
		then parsing can be aborted/stopped since some other
		BackDoor parser (module) has parsed the BackDoor
		Code request and indicated an error condition or
		acted upon the request.
******************************************************************************/
NDIS_STATUS cmCtlDeviceSpecific(
    IN CDSL_ADAPTER_T		* pThisAdapter,
    IN BACK_DOOR_T		* pBackDoorBuf,
    OUT PULONG			pNumBytesNeeded,
    OUT PULONG			pNumBytesWritten)
{
  NDIS_STATUS		Status = STATUS_SUCCESS;
  CDSL_LINK_T		* pThisLink;
  CDSL_LINK_CARD_MGMT_T	* pThisCMContext;
  DWORD			ParamSize = 0;
  DWORD 		SysIfUpdated;
  DWORD			SysIfStatus;

  if (pBackDoorBuf->TotalSize >= sizeof(MINI_BACK_DOOR_T))
    SysIfUpdated = pBackDoorBuf->TotalSize - BD_HEADER_SIZE;
  else
    SysIfUpdated = 0;

  //	Assume BACKDOOR ReqCode will be processed,
  //	set the BACKDOOR Status to Code Not Supported.
  //	NOTE:	The BackDoor Status should ONLY BE SET TO
  //		CODE_NOT_SUPPORTED at the beginning of BackDoor
  //		parsing by the first OID parser (CardMgmt)!!!
  //		ALL OTHER successor OID parsing of BackDoor should
  //		only set the BackDoor Status when it is intended
  //		to be set to a specific Status (i.e., BAD_SIGNATURE,
  //		FAILURE, or SUCCESS)!!  The BackDoor Status should
  //		not be set at the beginning of BackDoor parsing
  //		intending to change it if needed!!!
  pBackDoorBuf->ResultCode = ERROR_WARNING(RESULT_REQ_NOT_SUPPORTED);

  if (pBackDoorBuf->TotalSize < SIZEOF_DWORD_ALIGN(BACK_DOOR_T) &&
      pBackDoorBuf->TotalSize != SIZEOF_DWORD_ALIGN(MINI_BACK_DOOR_T))
  {
    pBackDoorBuf->ResultCode = ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
    pBackDoorBuf->NeededSize = SIZEOF_DWORD_ALIGN(BACK_DOOR_T);

    *pNumBytesNeeded = SIZEOF_DWORD_ALIGN(BACK_DOOR_T);

    DBG_CDSL_DISPLAY(DBG_LVL_WARNING,
		     pThisAdapter->DebugFlag,
		     (DEV_LABEL "=>CardMgmt:cmCtlDeviceSpecific  BACK_DOOR:   Entry, Size Failure;  ReqCode=0x%08lX;  NeededSize=0x%08lX.",
		      pBackDoorBuf->ReqCode,
		      pBackDoorBuf->NeededSize));
    return ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
  }

  switch (pBackDoorBuf->ReqCode)
  {
#if 0
  case BD_CDSL_GET_PRODUCT_INFO:
    COPY_MEMORY(pBackDoorBuf->Params.BdCdslProdInfo.RemoteManufacturer,
		pThisAdapter->CdslProdInfo.RemoteManufacturer,
		MAX_MANUFACTURER_LENGTH);
    pBackDoorBuf->Params.BdCdslProdInfo.CdslHwRevision = pThisAdapter->CdslProdInfo.CdslHwRevision;
    pBackDoorBuf->Params.BdCdslProdInfo.RemoteVendorId = pThisAdapter->CdslProdInfo.RemoteVendorId;

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = sizeof(pBackDoorBuf->Params.BdCdslProdInfo);
    break;

  case BD_CDSL_GET_OPERATIONAL_PARAMS:
    pBackDoorBuf->Params.BdCdslOperational.RateMode	= pThisAdapter->CdslOperational.RateMode;
    pBackDoorBuf->Params.BdCdslOperational.RxRate	= pThisAdapter->CdslOperational.RxRate;
    pBackDoorBuf->Params.BdCdslOperational.RxSymRate	= pThisAdapter->CdslOperational.RxSymRate;
    pBackDoorBuf->Params.BdCdslOperational.TxRate	= pThisAdapter->CdslOperational.TxRate;
    pBackDoorBuf->Params.BdCdslOperational.TxSymRate	= pThisAdapter->CdslOperational.TxSymRate;
    pBackDoorBuf->Params.BdCdslOperational.TxPwrMode	= pThisAdapter->CdslOperational.TxPwrMode;
    pBackDoorBuf->Params.BdCdslOperational.TxPwrVal	= pThisAdapter->CdslOperational.TxPwrVal;
    pBackDoorBuf->Params.BdCdslOperational.RxLevel	= pThisAdapter->CdslOperational.RxLevel;	//	Get/Read ONLY!
    pBackDoorBuf->Params.BdCdslOperational.SnMargin	= pThisAdapter->CdslOperational.SnMargin;	//	Get/Read ONLY!

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = sizeof(pBackDoorBuf->Params.BdCdslOperational);
    break;

  case BD_CDSL_SET_OPERATIONAL_PARAMS:
    pThisAdapter->CdslOperational.RateMode	= pBackDoorBuf->Params.BdCdslOperational.RateMode;
    pThisAdapter->CdslOperational.RxRate	= pBackDoorBuf->Params.BdCdslOperational.RxRate;
    pThisAdapter->CdslOperational.RxSymRate	= pBackDoorBuf->Params.BdCdslOperational.RxSymRate;
    pThisAdapter->CdslOperational.TxRate	= pBackDoorBuf->Params.BdCdslOperational.TxRate;
    pThisAdapter->CdslOperational.TxSymRate	= pBackDoorBuf->Params.BdCdslOperational.TxSymRate;
    pThisAdapter->CdslOperational.TxPwrMode	= pBackDoorBuf->Params.BdCdslOperational.TxPwrMode;
    pThisAdapter->CdslOperational.TxPwrVal	= pBackDoorBuf->Params.BdCdslOperational.TxPwrVal;
    //	pThisAdapter->CdslOperational.RxLevel	= pBackDoorBuf->Params.BdCdslOperational.RxLevel;	//	Get/Read ONLY!
    //	pThisAdapter->CdslOperational.SnMargin	= pBackDoorBuf->Params.BdCdslOperational.SnMargin;	//	Get/Read ONLY!

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = 0;
    break;
#endif

  case BD_CARDMGMT_GET_CONNECT_STATUS:
    pBackDoorBuf->Params.BdCardMgmtConnStatus = (DWORD) BACKDOOR_STATUS_IDLE;
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink && pThisLink->CardMgmt)
    {
      pThisCMContext = (CDSL_LINK_CARD_MGMT_T *)pThisLink->CardMgmt;
      pThisCMContext->State = LINK_STATE_CONNECTED;
      if (pThisCMContext->State == LINK_STATE_CONNECTED)
	pBackDoorBuf->Params.BdCardMgmtConnStatus = (DWORD) BACKDOOR_STATUS_DATAMODE;
      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = sizeof(pBackDoorBuf->Params.BdCardMgmtConnStatus);
    break;

  case BD_CARDMGMT_GET_DEBUG_FLAG:
    pBackDoorBuf->Params.BdCardMgmtDebugFlag = pThisAdapter->DebugFlag;

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = sizeof(pBackDoorBuf->Params.BdCardMgmtDebugFlag);
    break;

  case BD_CARDMGMT_SET_DEBUG_FLAG:
    pThisAdapter->DebugFlag = pBackDoorBuf->Params.BdCardMgmtDebugFlag;

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = 0;
    break;

  case BD_CARDMGMT_GET_PRODUCT_INFO:
    pBackDoorBuf->Params.BdCardMgmtProdInfo.DrvrSwVer = pThisAdapter->ProductVersion;
    pBackDoorBuf->Params.BdCardMgmtProdInfo.DrvrTapiVer = TAPI_API_VERSION;
    pBackDoorBuf->Params.BdCardMgmtProdInfo.DrvrNdisVer = DRIVER_VERSION_OF_NDIS;

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;

    ParamSize = sizeof(pBackDoorBuf->Params.BdCardMgmtProdInfo);
    break;

  case BD_CARDMGMT_OPEN_BD_INSTANCE:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink)
    {
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_FAILURE;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_UNAVAILABLE;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_EMPTY;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_SUCCESS;
      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = 0;
    break;

  case BD_CARDMGMT_CLOSE_BD_INSTANCE:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink)
    {
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_FAILURE;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_UNAVAILABLE;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_EMPTY;
      pBackDoorBuf->Params.BdCardMgmtInstanceStatus = BD_STATUS_INSTANCE_SUCCESS;

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = 0;
    break;

  case BD_CARDMGMT_GET_PRODUCT_VERSION:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink)
    {
      ParamSize = strlen(REVISION_LEVEL) + 1;
      if (ParamSize > sizeof(BD_MINI_DATA_BLOCK_T))
      {
	ParamSize = sizeof(BD_MINI_DATA_BLOCK_T) - 1;
	pBackDoorBuf->Params.BdCardMgmtProdVer[ParamSize] = 0;

	pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_DATA_FAILURE);
      } else
	pBackDoorBuf->ResultCode = RESULT_SUCCESS;

      COPY_MEMORY(&pBackDoorBuf->Params.BdCardMgmtProdVer,
		  REVISION_LEVEL,
		  ParamSize);
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    break;

  case BD_CARDMGMT_GET_CURRENT_MAC_ADDR:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink)
    {
      //*****************************************************
      //** These addresses ETH_ADDRESS_LENGTH &
      //**  WAN_ADDRESS_LENGTH are both 6 bytes.
      //*****************************************************
      COPY_MEMORY(&pBackDoorBuf->Params.BdCardMgmtMacAddr,
		  pThisAdapter->PermanentEthAddress,
		  ETH_ADDRESS_LENGTH);

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = ETH_ADDRESS_LENGTH;
    break;

  case BD_CARDMGMT_GET_EEPROM_MAC_ADDR:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink) {
      pBackDoorBuf->ResultCode = CardALChipGetMacAddr(pThisAdapter,
						      pBackDoorBuf->Params.BdCardMgmtMacAddr,
						      ETH_ADDRESS_LENGTH);

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = ETH_ADDRESS_LENGTH;
    break;

  case BD_CARDMGMT_SET_EEPROM_MAC_ADDR:
    pThisLink = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (pThisLink) {
      pBackDoorBuf->ResultCode = CardALChipSetMacAddr(pThisAdapter,
						      pBackDoorBuf->Params.BdCardMgmtMacAddr,
						      ETH_ADDRESS_LENGTH);

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else
      pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);

    ParamSize = 0;
    break;

  default:
    Status = STATUS_UNSUPPORTED;
    break;
  }

  if (Status != STATUS_UNSUPPORTED) {
    pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten = ParamSize +
                        sizeof(pBackDoorBuf->ResultCode) +
                        sizeof(pBackDoorBuf->NeededSize);
  }

  // * Notify System Interace of new API Command
  SysIfStatus = SmSysIfBroadcastMessage(pThisAdapter,
					MESS_API_COMMAND,		// Unique to each channel
					pBackDoorBuf->ReqCode,		// Embedded Message key
					(CHAR *) &pBackDoorBuf->Params,	// Message Buffer - Send only Params Union
					&SysIfUpdated);			// Length only reflects Params Union

  if (SysIfStatus != STATUS_INVALID_PARAMETER)
  {
    // * Invalid parameters only mean the event is not recognized or was not
    // * handled.
    // * Otherwise, we attempted to handle and must update status.
    *pNumBytesNeeded += SysIfUpdated  + BD_HEADER_SIZE;
    *pNumBytesWritten += SysIfUpdated + BD_HEADER_SIZE;
    if (SysIfStatus == -EBADLENGTH)
    {
      pBackDoorBuf->ResultCode = ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
      pBackDoorBuf->NeededSize = SysIfUpdated;
    } else
      pBackDoorBuf->ResultCode = SysIfStatus;
  }
	
  return Status;
}

#ifdef CONFIG_PROC_FS
static const char *GHS_CAP[] = {
  "V8",
  "V8bis",
  "SilentPeriod",
  "GPLOAM"};

static const char *GHS_CAP_922[] = {
  "G9221AnnexA",
  "G9221AnnexB",
  "G9221AnnexC",
  "G9222AnnexAB",
  "G9222AnnexC"
};

static
void xGHS_CAP(
    OUT char *ProcPage,
    INOUT int *pRetBytes,
    const char *s,
    const BD_ADSL_GHS_CAP_ENTRY_T *p
)
{
  int RetBytes = *pRetBytes;

  RetBytes += sprintf(ProcPage+RetBytes,
		      "%s: %1d %1d %1d\n",
		      s,
		      p->LocalEndpoint,
		      p->RemoteEndpoint,
		      p->Negiotiated);

  *pRetBytes = RetBytes;
}

LOCAL int ReadProcCnxAdsl(
    IN char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData,
    enum tyuiu	mode
)
{
  CDSL_ADAPTER_T	*pThisAdapter = ProcData;
  int			RetBytes = 0;

  if (mode==DebugQuery) {
    RetBytes += sprintf(ProcPage+RetBytes,
			"VendorId = %s\n",
			pThisAdapter->VendorId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"NeedReset = %d\n",
			pThisAdapter->NeedReset);
    RetBytes += sprintf(ProcPage+RetBytes,
			"Loading = %d\n",
			pThisAdapter->Loading);
    RetBytes += sprintf(ProcPage+RetBytes,
			"AutoDownload = %d\n",
			pThisAdapter->AutoDownload);
    RetBytes += sprintf(ProcPage+RetBytes,
			"MediaConnectStatus = %ld\n",
			pThisAdapter->MediaConnectStatus);
    RetBytes += sprintf(ProcPage+RetBytes,
			"Showtime = %d\n",
			pThisAdapter->Showtime);
    RetBytes += sprintf(ProcPage+RetBytes,
			"InterruptsInitiailized = %d\n",
			pThisAdapter->InterruptsInitiailized);
    RetBytes += sprintf(ProcPage+RetBytes,
			"InstanceNumber = %d\n",
			pThisAdapter->InstanceNumber);
    RetBytes += sprintf(ProcPage+RetBytes,
			"AdapterInitComplete = %d\n",
			pThisAdapter->AdapterInitComplete);
    RetBytes += sprintf(ProcPage+RetBytes,
			"DownloadInitComplete = %d\n",
			pThisAdapter->DownloadInitComplete);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.CommandStatus = %ld\n",
			pThisAdapter->PersistData.CommandStatus);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.DebugFlag = 0x%08lX\n",
			pThisAdapter->PersistData.DebugFlag);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoSenseHandshake = %ld\n",
			pThisAdapter->PersistData.AutoSenseHandshake);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoSenseWiresFirst = %ld\n",
			pThisAdapter->PersistData.AutoSenseWiresFirst);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoWiringSelection = %ld\n",
			pThisAdapter->PersistData.AutoWiringSelection);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoWiringRelayDelay = %ld\n",
			pThisAdapter->PersistData.AutoWiringRelayDelay);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoWiringRelayEnrg = %ld\n",
			pThisAdapter->PersistData.AutoWiringRelayEnrg);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.AutoWiringNumCombos = %ld\n",
			pThisAdapter->PersistData.AutoWiringNumCombos);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.ARMHwId = 0x%08lX\n",
			pThisAdapter->PersistData.ARMHwId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.ADSLHwId = 0x%08lX\n",
			pThisAdapter->PersistData.ADSLHwId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.ARMFunctionIdentifier = %ld\n",
			pThisAdapter->PersistData.ARMFunctionIdentifier);
    RetBytes += sprintf(ProcPage+RetBytes,  
			"PersistData.ADSLFunctionIdentifier = %ld\n",
			pThisAdapter->PersistData.ADSLFunctionIdentifier);
#if 0
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.HwDebugFlag = 0x%08lX\n",
			pThisAdapter->PersistData.HwDebugFlag);
#endif
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.PhysicalDriverType = %ld\n",
			pThisAdapter->PersistData.PhysicalDriverType);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.GpioADSLFunctionMask = 0x%08lX\n",
			pThisAdapter->PersistData.GpioADSLFunctionMask);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.GpioARMFunctionMask = 0x%08lX\n",
			pThisAdapter->PersistData.GpioARMFunctionMask);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.GpioDirectionMask = 0x%08lX\n",
			pThisAdapter->PersistData.GpioDirectionMask);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.GpioEdgeMask = 0x%08lX\n",
			pThisAdapter->PersistData.GpioEdgeMask);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.CnxtVendorId = 0x%08lX\n",
			pThisAdapter->PersistData.CnxtVendorId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.CnxtArmDeviceId = 0x%08lX\n",
			pThisAdapter->PersistData.CnxtArmDeviceId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.CnxtAdslDeviceId = 0x%08lX\n",
			pThisAdapter->PersistData.CnxtAdslDeviceId);
    RetBytes += sprintf(ProcPage+RetBytes,
			"PersistData.CnxtPhysicalDriverType = %d\n",
			pThisAdapter->PersistData.CnxtPhysicalDriverType);
  } else {
    char Buffer[sizeof(TIG_DEVICE_SPEC_T) + sizeof(BACK_DOOR_T)];
    TIG_DEVICE_SPEC_T	*pDevSpecific = (TIG_DEVICE_SPEC_T*)Buffer;
    BACK_DOOR_T		*pBackDoorBuf = (BACK_DOOR_T*)pDevSpecific->BackDoorBuf;
    struct atm_dev	*pDev = pThisAdapter->pAtmDevice;
    int			Index;

    switch(mode) {
    default:
    case DebugQuery:
      break;
    case RxSnr:
      for (Index=0; Index<=255; Index+=BD_ADSL_SNR_NUMBER)
      {
	BD_ADSL_SNR_T *pBdADSLSnrTable = &pBackDoorBuf->Params.BdADSLSnrTable;

	// init the SNR statistics structure
	CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
	pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
	pBackDoorBuf->ReqCode = BD_ADSL_GET_SNR_TABLE;

	pBdADSLSnrTable->StartingBin = Index;
	pBdADSLSnrTable->EndingBin = Index+BD_ADSL_SNR_NUMBER;

	// read the SNR statistics
	CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

	if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
	{
	  int TableIndex;

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "%3u:",
			      pBdADSLSnrTable->StartingBin);

	  for (TableIndex=0;TableIndex<BD_ADSL_SNR_NUMBER;TableIndex++)
	    RetBytes += sprintf(ProcPage+RetBytes,
				" %04x",
				pBdADSLSnrTable->Entry[TableIndex].BinSNR&0xffff);

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	}
      }
      break;
    case Bins:
      for (Index=0; Index<=255; Index+=BD_ADSL_BIT_ALLOCATION_NUMBER)
      {
	BD_ADSL_BIT_ALLOCATION_T *pBdADSLBitAllocationTable;

	pBdADSLBitAllocationTable = &pBackDoorBuf->Params.BdADSLBitAllocationTable;

	// init the info structure
	CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
	pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
	pBackDoorBuf->ReqCode = BD_ADSL_GET_BIT_ALLOCATION_TABLE;

	pBdADSLBitAllocationTable->StartingBin = Index;
	pBdADSLBitAllocationTable->EndingBin = Index+BD_ADSL_BIT_ALLOCATION_NUMBER;

	// read the info
	CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

	if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
	{
	  int TableIndex;

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "%3u:",
			      pBdADSLBitAllocationTable->StartingBin);

	  for (TableIndex=0;TableIndex<BD_ADSL_BIT_ALLOCATION_NUMBER;TableIndex++)
	    RetBytes += sprintf(ProcPage+RetBytes,
				" %c %01x %02x",
				pBdADSLBitAllocationTable->Entry[TableIndex].BinStatus,
				pBdADSLBitAllocationTable->Entry[TableIndex].BitsAssigned,
				pBdADSLBitAllocationTable->Entry[TableIndex].BitCapacity);

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	}
      }
      break;
    case X:
      {
	BD_ADSL_ALARMS_STATUS_T *pBdADSLAlaramsStatus = &pBackDoorBuf->Params.BdADSLAlaramsStatus;

	// init the info structure
	CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
	pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
	pBackDoorBuf->ReqCode = BD_ADSL_GET_ALARMS;

	// read the info
	CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

	if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
	{
	  static const char *Alarm[] = ALARM_STRING;
#define WHICHALM(x) ((x) < ARRAY_SIZE(Alarm)) ? Alarm[(x)] : "?"

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "Upstream Alarm info\n");

	  for (Index=0; Index<BD_ADSL_ALARMS_NUMBER; Index++)
	    if (pBdADSLAlaramsStatus->UpStreamAlarms.Alarms[Index])
	      RetBytes += sprintf(ProcPage+RetBytes,
				  "%1d:%s\n",
				  Index,
				  WHICHALM(pBdADSLAlaramsStatus->UpStreamAlarms.Alarms[Index]-1));

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "Downstream Alarm info\n");

	  for (Index=0; Index<BD_ADSL_ALARMS_NUMBER; Index++)
	    if (pBdADSLAlaramsStatus->DownStreamAlarms.Alarms[Index])
	      RetBytes += sprintf(ProcPage+RetBytes,
				  "%d:%s\n",
				  Index,
				  WHICHALM(pBdADSLAlaramsStatus->DownStreamAlarms.Alarms[Index]-1));
#undef WHICHALM
	}
      }

      {
	BD_ADSL_GHS_CAP_T *pBdADSLG_HSCapabilities = &pBackDoorBuf->Params.BdADSLG_HSCapabilities;

	// init the info structure
	CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
	pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
	pBackDoorBuf->ReqCode = BD_ADSL_GET_GHS_CAP;

	// read the info
	CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

	if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
	{
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "HSCapabilities\n");

	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP[0],&pBdADSLG_HSCapabilities->V8);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP[1],&pBdADSLG_HSCapabilities->V8bis);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP[2],&pBdADSLG_HSCapabilities->SilentPeriod);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP[3],&pBdADSLG_HSCapabilities->GPLOAM);

	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP_922[0],&pBdADSLG_HSCapabilities->G9221AnnexA);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP_922[1],&pBdADSLG_HSCapabilities->G9221AnnexB);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP_922[2],&pBdADSLG_HSCapabilities->G9221AnnexC);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP_922[3],&pBdADSLG_HSCapabilities->G9222AnnexAB);
	  xGHS_CAP(ProcPage,&RetBytes, GHS_CAP_922[4],&pBdADSLG_HSCapabilities->G9222AnnexC);

	  RetBytes += sprintf(ProcPage+RetBytes,
			      "RemoteCountry: %d\n",
			      pBdADSLG_HSCapabilities->RemoteCountry);
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "RemoteVendor: %02x %02x %02x %02x\n",
			      pBdADSLG_HSCapabilities->RemoteVendorID[0],
			      pBdADSLG_HSCapabilities->RemoteVendorID[1],
			      pBdADSLG_HSCapabilities->RemoteVendorID[2],
			      pBdADSLG_HSCapabilities->RemoteVendorID[3]);
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "RemoteSpecInfo: %02x %02x\n",
			      pBdADSLG_HSCapabilities->RemoteSpecInfo[0],
			      pBdADSLG_HSCapabilities->RemoteSpecInfo[1]);
	}
      }

      {
	BD_ADSL_LOCAL_GHS_CAP_T *pBdADSLocalG_HSCapabilities = &pBackDoorBuf->Params.BdADSLocalG_HSCapabilities;

	// init the info structure
	CLEAR_MEMORY(pBackDoorBuf, sizeof(*pBackDoorBuf));
	pBackDoorBuf->TotalSize = sizeof(*pBackDoorBuf);
	pBackDoorBuf->ReqCode = BD_ADSL_GET_GHS_LOCAL_CAP;

	// read the info
	CmgmtProcIOCTL(pDev, TIG_DEVICE_SPEC, pDevSpecific);

	if (pBackDoorBuf->ResultCode == RESULT_SUCCESS)
	{
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "LocalG_HSCapabilities\n");

	  for (Index=0;Index<4;Index++)
	    if (pBdADSLocalG_HSCapabilities->BdADSLLocalG922Cap&(1<<Index))
	      RetBytes += sprintf(ProcPage+RetBytes,
				  "%s ",
				  GHS_CAP[Index]);
	  for (Index=0;Index<5;Index++)
	    if (pBdADSLocalG_HSCapabilities->BdADSLLocalG922AnnexCap&(1<<Index))
	      RetBytes += sprintf(ProcPage+RetBytes,
				  "%s ",
				  GHS_CAP_922[Index]);
	  RetBytes += sprintf(ProcPage+RetBytes,
			      "\n");
	}
      }
      break;
    }
  }

  *bEndOfFile = TRUE;

  return RetBytes;
}

LOCAL int ReadProcCnxAdslDebugQuery(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData)
{
  return ReadProcCnxAdsl(ProcPage,
			 ProcPageStart,
			 ProcPageOffset,
			 ProcPageLen,
			 bEndOfFile,
			 ProcData,
			 DebugQuery);
}

LOCAL int ReadProcCnxAdslRxSnr(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData)
{
  return ReadProcCnxAdsl(ProcPage,
			 ProcPageStart,
			 ProcPageOffset,
			 ProcPageLen,
			 bEndOfFile,
			 ProcData,
			 RxSnr);
}

LOCAL int ReadProcCnxAdslBins(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData)
{
  return ReadProcCnxAdsl(ProcPage,
			 ProcPageStart,
			 ProcPageOffset,
			 ProcPageLen,
			 bEndOfFile,
			 ProcData,
			 Bins);
}

LOCAL int ReadProcCnxAdslX(
    char	* ProcPage,
    char	** ProcPageStart,
    off_t	ProcPageOffset,
    int		ProcPageLen,
    int		* bEndOfFile,
    void	* ProcData)
{
  return ReadProcCnxAdsl(ProcPage,
			 ProcPageStart,
			 ProcPageOffset,
			 ProcPageLen,
			 bEndOfFile,
			 ProcData,
			 X);
}
#endif

#if 1
/******************************************************************************
FUNCTION NAME:
	CmgmtGetApiList

ABSTRACT:
	Return List of commands for user API. The BdSupportedList must be an
	INORDER list.

RETURN:
	NDIS_STATUS_SUCCESS

DETAILS:
	The storage for BdSupportedList cannot be released until after
	Driver Adapter Inits are complete.  In this case, it is a static
	so it is not necessary to release.
******************************************************************************/
DWORD CmgmtGetApiList(
    VOID			* ExternHandle,
    const DWORD			** MessageList,/*CLN ??*/
    DWORD			* MaxMessages)
{
  static const DWORD ApiSupportedList[] = {
    BD_ADSL_GET_PERFORMANCE,
    BD_ADSL_INIT_DATA_PUMP,
    BD_ADSL_GET_CONN_CONFIG,
    BD_ADSL_SET_CONN_CONFIG,
    BD_ADSL_GET_MODEM_DATA,
    BD_ADSL_GET_DATA_PATH_SELECTION,
    BD_ADSL_SET_DATA_PATH_SELECTION,
    BD_ADSL_GET_LINE_STATUS,
    BD_ADSL_GET_LINE_STATE,
    BD_ADSL_GET_DP_VERSIONS,
    BD_ADSL_GET_SPECIFIC_CMD_RESP,
    BD_ADSL_GET_BUS_CONTROLLER,			/* 0x00060011 */
    BD_ADSL_SET_TRACE,
    BD_ADSL_DEBUG,

    BD_CARDAL_GET_REGISTER,			/* 0x00070001 */
    BD_CARDAL_SET_REGISTER,
    BD_USER_ACTIVATE_LINE,
    BD_USER_DEACTIVATE_LINE,
    BD_CARDAL_GET_TEXT_LOG,
    BD_ADSL_GET_CONFIG,
    BD_ADSL_SET_CONFIG,

    BD_PM_GET_POWER_STATE,			/* 0x00090001 */
    BD_PM_SET_POWER_STATE,

    BD_CARDAL_GET_ADSL_TRANSCEIVER_STATUS,	/* 0x000A0005 */

    BD_ADSL_GET_ALARMS,
    BD_ADSL_GET_SNR_TABLE,
    BD_ADSL_GET_CONTROLLER_LOG_CLT,
    BD_ADSL_SET_CONTROLLER_LOG_CLT,
    BD_ADSL_GET_BIT_ALLOCATION_TABLE,
    BD_ADSL_GET_GHS_CAP,
    BD_ADSL_GET_GHS_LOCAL_CAP,

    BD_ADSL_SET_GHS_LOCAL_CAP,
    BD_ADSL_SET_AUTO_SENSE
  };

  *MaxMessages = ARRAY_SIZE(ApiSupportedList);
  *MessageList = ApiSupportedList;

  return(STATUS_SUCCESS);
}		
#endif
