/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	ChipAL
**
**FILE NAME:
**	ChipAlCdsl.c
**
**ABSTRACT:
**	Chip Abstraction Layer Public Interface File
**
**DETAILS:
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Yukon-AV/Linux_2.4_ATM_Driver/KernelModule/ChipALCdsl.c $
**	$Revision: 8 $
**	$Date: 5/23/01 8:23a $
*******************************************************************************
******************************************************************************/

/******************************************************************************
Includes
******************************************************************************/
#define _CHIPALCDSL_C_
#define MODULE_ID	DBG_ENABLE_CHIPAL

#include "linuxmodver.h"

#include "CardMgmt.h"
#include "FrameAL.h"
#include "../CommonData.h"
#include "LnxTools.h"

#include "ChipALCdslV.h"

#include "SmSysIf.h" 

#ifdef _CHIPALCDSL_C_
/* These are module globals */
static USHORT	MyAdapterCount = 0;
static CHIPAL_T	ChipALHead = { NULL };
static CHIPAL_T	* pChipALHead = NULL;
DWORD		ChipALDebugFlag;/*CLN*/
#endif

/******************************************************************************
ChipAl Module Definitions
******************************************************************************/

#define ERROR_FATAL(error)	(LEVEL_FATAL   | MODULE_CHIPAL | INSTANCE_BASIC2 | error)
#define ERROR_WARNING(error)	(LEVEL_WARNING | MODULE_CHIPAL | INSTANCE_BASIC2 | error)
#define ERROR_NORMAL(error)	(LEVEL_NORMAL  | MODULE_CHIPAL | INSTANCE_BASIC2 | error)

#ifndef MEMORY_TAG
#define MEMORY_TAG		'MCLC'
#endif

#define BITS_PER_BYTE		8
#define CHANNEL_SIZE_BY_SPEED	0
#define MAX_SYNC_WAIT		20

#define DEFAULT_RX_SEGMENT_NO	96
#if (PROJECTS_NDIS_IF == PROJECT_NDIS_IF_WAN)
#define DEFAULT_RX_SEGMENT_SIZE 2048
#else
#define DEFAULT_RX_SEGMENT_SIZE 512
#endif 
#define DEFAULT_RX_CHANNEL_SIZE	(DEFAULT_RX_SEGMENT_NO *  DEFAULT_RX_SEGMENT_SIZE)

#define DEFAULT_TX_SEGMENT_NO	8
#define DEFAULT_TX_SEGMENT_SIZE 2048
#define DEFAULT_TX_CHANNEL_SIZE	(DEFAULT_TX_SEGMENT_NO *  DEFAULT_TX_SEGMENT_SIZE)

#define CHIPAL_ERROR_FATAL(error)	(CHIPAL_FATAL   | CHIPAL_MODULE | CHIPAL_BASIC2 | error)
#define CHIPAL_ERROR_WARNING(error)	(CHIPAL_WARNING | CHIPAL_MODULE | CHIPAL_BASIC2 | error)

#ifdef DBG
#define DBG_DPC_TRACE		0
#else
#define DBG_DPC_TRACE		0
#endif

// soft timer period is 10 ms
#define SOFT_TIMER_PERIOD	10

#define ONE_SEC 1000

/******************************************************************************
Private Data
******************************************************************************/

#if 0
static const unsigned char ReverseByteTable[256] = {
0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,
0x10,0x90,0x50,0xd0,0x30,0xb0,0x70,0xf0,
0x08,0x88,0x48,0xc8,0x28,0xa8,0x68,0xe8,
0x18,0x98,0x58,0xd8,0x38,0xb8,0x78,0xf8,
0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,
0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4,
0x0c,0x8c,0x4c,0xcc,0x2c,0xac,0x6c,0xec,
0x1c,0x9c,0x5c,0xdc,0x3c,0xbc,0x7c,0xfc,
0x02,0x82,0x42,0xc2,0x22,0xa2,0x62,0xe2,
0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2,
0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,
0x1a,0x9a,0x5a,0xda,0x3a,0xba,0x7a,0xfa,
0x06,0x86,0x46,0xc6,0x26,0xa6,0x66,0xe6,
0x16,0x96,0x56,0xd6,0x36,0xb6,0x76,0xf6,
0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,
0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe,
0x01,0x81,0x41,0xc1,0x21,0xa1,0x61,0xe1,
0x11,0x91,0x51,0xd1,0x31,0xb1,0x71,0xf1,
0x09,0x89,0x49,0xc9,0x29,0xa9,0x69,0xe9,
0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9,
0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,
0x15,0x95,0x55,0xd5,0x35,0xb5,0x75,0xf5,
0x0d,0x8d,0x4d,0xcd,0x2d,0xad,0x6d,0xed,
0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,0x7d,0xfd,
0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,
0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3,
0x0b,0x8b,0x4b,0xcb,0x2b,0xab,0x6b,0xeb,
0x1b,0x9b,0x5b,0xdb,0x3b,0xbb,0x7b,0xfb,
0x07,0x87,0x47,0xc7,0x27,0xa7,0x67,0xe7,
0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7,
0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,
0x1f,0x9f,0x5f,0xdf,0x3f,0xbf,0x7f,0xff};
#endif

#ifdef DBG
typedef enum _TRACE_MESSAGES {
  UDEFINED,
  ISR_ENTRY = 1,
  ISR_EXIT,
  SYNC_ENTRY,
  SYNC_EXIT,
  TX_ENTRY,
  TX_EXIT,
  RX_ENTRY,
  RX_EXIT,
  DPC_ENTRY,
  DPC_EXIT,
  /**/
  MAX_TRACE
} TRACE_TYPE;
#define TRACE_QUEUE_LEN 500
static WORD IntCount = 0;
static UCHAR TraceQueue[TRACE_QUEUE_LEN];
static WORD TraceQueueIdx = 0;
#define TRACE_MESSAGE(n) do {TraceQueue[TraceQueueIdx] = n; \
	TraceQueueIdx = ++TraceQueueIdx % TRACE_QUEUE_LEN;} while (0)
#else
#define TRACE_MESSAGE(n)
#endif

/******************************************************************************
Private Prototypes
******************************************************************************/

LOCAL INLINE CHIPAL_T * chipALAllocateInterface(
    void);

LOCAL INLINE void chipALReleaseInterface(
    CHIPAL_T * AdapterToRelease);

LOCAL INLINE void chipALGetChanConfig(
    HW_CH_ENTRY_T 	* pChanConfig,
    CHIPAL_T 		* pThisInterface);

LOCAL INLINE BOOLEAN chipALClearInterrupt(
    CHIPAL_T	* pThisInterface,
    enum WhichFunction	FunctSel,
    DWORD	IntStatus);

LOCAL INLINE void chipALGetStats(
    CDSL_ADAPTER_T * pThisAdapter,
    BACK_DOOR_T	   * pBackDoorBuf);

LOCAL INLINE void chipALClearStats(
    CDSL_ADAPTER_T * pThisAdapter,
    BACK_DOOR_T	   * pBackDoorBuf);

LOCAL HW_CH_ENTRY_T * chipALFindLink(
    CHIPAL_T 		* pThisInterface,
    CDSL_LINK_T		* pCdslLink,
    CHAN_DESCRIPT_T	Description,
    DWORD		Mask);

#if CAL_SOFT_TIMER
VOID chipAlSoftTimer(
    IN ULONG	pInterf);
#endif

#ifdef DBG	// Development Test Routines
// * The ChipALTestxxx Prototypes are for development purposes only
DWORD chipALTestReadBuff(
    char * Destination,
    DWORD BuffSize,
    DWORD * SizeWritten);

DWORD chipALTestWriteBuff(
    CDSL_ADAPTER_T * ThisAdapter,
    char * Destination,
    DWORD Size);

DWORD chipALTestReadConfig(
    CDSL_ADAPTER_T * ThisAdapter,
    BD_CHIPAL_CONFIG_T * pConfigStatus);
#endif // DBG

LOCAL CHIPAL_FAILCODE_T  lChipALRead(
    VOID		* pAdapter,
    HARDWARE_DEVICE_T	HwDevice,
    DWORD		LogicalDevice,
    DWORD		OffsetAddress,
    DWORD		RegisterLength,
    PVOID		pValue);

LOCAL CHIPAL_FAILCODE_T  lChipALWrite(
    VOID		* pAdapter,
    HARDWARE_DEVICE_T	HwDevice,
    DWORD		LogicalDevice,
    DWORD		OffsetAddress,
    DWORD		RegisterLength,
    DWORD		Value,
    DWORD		Mask);

LOCAL CHIPAL_FAILCODE_T lChipALAddEventHandler(
    VOID			* pAdapter,
    IN DWORD			FunctSelect,	// selects either the arm or dsl function
    IN CHIPAL_EVENT_T		Event,		// Enumerated Event to Set
    IN HW_EVENT_TYPE_T		IrqLevel,	// When Event should be called - DIRQl or Dispatch IRQL
    IN CHIPAL_EVENT_HANDLER_T	UserEventHandler,// Function to Call
    IN void			* UserContext	 // Function Context - Unused by ChipAL
);

LOCAL CHIPAL_FAILCODE_T lChipALRemoveEventHandler(
    IN VOID			* pAdapter,
    IN DWORD			FunctSelect,	// selects the arm or dsl function
    IN CHIPAL_EVENT_T		Event,		// Event to Remove
    IN HW_EVENT_TYPE_T		IrqLevel,	// When Event was being called
    IN CHIPAL_EVENT_HANDLER_T	UserEventHandler// Handler to remove
);

LOCAL void lChipALSetEvent(
    VOID			* pAdapter,
    DWORD			FunctSelect,
    CHIPAL_EVENT_T		Event);

LOCAL CHIPAL_FAILCODE_T	lChipAlGetConfig(
    IN VOID			* pAdapter,
    I_O	VOID			* pConfigArg);

LOCAL CHIPAL_FAILCODE_T	lChipAlSetConfig(
    IN VOID			* pAdapter,
    I_O	VOID			* pConfigArg);

LOCAL NTSTATUS lChipALLinkShutdown(
    PVOID pLink,
    DWORD ComChannel);

LOCAL NTSTATUS lChipALLinkDisable(
    PVOID	pAdapter,		// Adapter Instance
    DWORD	ComChannel		// Communications Channel
);

LOCAL NTSTATUS lChipALStartTx(
    PVOID		pLink,
    DWORD 		ComChannel
);

LOCAL NTSTATUS lChipALLinkEnable(
    PVOID		pAdapter,	// Adapter Instance
    DWORD		ComChannel);

/******************************************************************************
Public Functions
******************************************************************************/

/******************************************************************************
FUNCTION NAME:
	ChipALCfgInit

ABSTRACT:
	Requirement of the Architecture to read tuning parameters from a
	Windowxxx Registry

RETURN:
	NDIS_STATUS

DETAILS:
******************************************************************************/
GLOBAL NDIS_STATUS ChipALCfgInit(
    IN	CDSL_ADAPTER_T	*pThisAdapter,	// Driver supplied adapter context
    // (&ThisAdapter)  Required for events that are not associated with a link
    IN	PTIG_USER_PARAMS pParams        // parameter struct from the download app
)
{
  CHIPAL_TUNING_T	*pTune;

  ChipALDebugFlag = pThisAdapter->PersistData.DebugFlag;

  if (!pThisAdapter->Tune) {
    NDIS_STATUS Status;

    Status = ALLOCATE_MEMORY(pThisAdapter->Tune, sizeof(*(pThisAdapter->Tune)));
    if (Status != STATUS_SUCCESS)
      return(STATUS_FAILURE);
  }
  pTune = pThisAdapter->Tune;

  pTune->PhysicalBase    = NULL_MEMORY;
  pTune->IRQ             = NULL_IRQ;

  pTune->RxMaxLatency    = pParams->RxMaxLatency;
  pTune->TxMaxLatency    = pParams->TxMaxLatency;
  pTune->RxMinLatency    = pParams->RxMinLatency;
  pTune->TxMinLatency    = pParams->TxMinLatency;
  pTune->RxInterruptRate = pParams->RxInterruptRate;
  pTune->TxInterruptRate = pParams->TxInterruptRate;
  pTune->RxSpeed         = pParams->RxSpeed;
  pTune->TxSpeed         = pParams->TxSpeed;
  pTune->TxFifoSize      = pParams->TxFifoSize;
  pTune->RxFifoSize      = pParams->RxFifoSize;

  pTune->RxChannelSize   = pParams->RxChannelSize;
  pTune->RxSegments      = pParams->RxSegments;
  pTune->TxChannelSize   = pParams->TxChannelSize;
  pTune->TxSegments      = pParams->TxSegments;

  pTune->TxFifoSize = FIFO_SIZE_TO_MASK(HOST_TO_DEVICE,pTune->TxFifoSize);

  // ensure there was a registry value else replace with default
  if (!pParams->RxChannelSizeCfg)
    pTune->RxChannelSize = DEFAULT_RX_CHANNEL_SIZE;
  if (!pTune->RxSegments || !pParams->RxSegmentsCfg)
    pTune->RxSegments = DEFAULT_RX_SEGMENT_NO;
  if (!pParams->TxChannelSizeCfg)
    pTune->TxChannelSize = DEFAULT_TX_CHANNEL_SIZE;
  if (!pTune->TxSegments || !pParams->TxSegmentsCfg)
    pTune->TxSegments = DEFAULT_TX_SEGMENT_NO;

  // Calculate Default Channel Size
#if CHANNEL_SIZE_BY_SPEED
  {
    DWORD			BytesPerSecond;
    DWORD			BytesPerInterrupt;

    // * ChannelSize and Segments are rounded to DWORDS.

    // * Calculate Receive Channel
    BytesPerSecond = pTune->RxSpeed / BITS_PER_BYTE;
    BytesPerInterrupt = DWORD_ALIGN(BytesPerSecond/(1000/pTune->RxInterruptRate));
    pTune->RxChannelSize = (WORD) DWORD_ALIGN(BytesPerSecond/(1000/pTune->RxMaxLatency));
    pTune->RxSegments = (WORD) (pTune->RxChannelSize / BytesPerInterrupt);

    // * Calculate Transmit Channel
    BytesPerSecond = pTune->TxSpeed / BITS_PER_BYTE;
    BytesPerInterrupt = DWORD_ALIGN(BytesPerSecond/(1000/pTune->TxInterruptRate));
    pTune->TxChannelSize = (WORD) DWORD_ALIGN(BytesPerSecond/(1000/pTune->TxMaxLatency));
    pTune->TxSegments = (WORD) (pTune->TxChannelSize / BytesPerInterrupt);
  }
#else
  // pTune->RxChannelSize = DMA_DEFAULT_CHANNEL_SIZE;
  // pTune->RxSegments = DMA_DEFAULT_SEGMENTS;
  // pTune->TxChannelSize = DMA_DEFAULT_CHANNEL_SIZE / 8;
  // pTune->TxSegments = 2;
#endif

  pTune->TotalSharedMemory = pTune->RxChannelSize + pTune->TxChannelSize;

  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   ("Tx Channel Defaults Size: %d, Segments %d\n",pTune->TxChannelSize,pTune->TxSegments));
  DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		   ChipALDebugFlag,
		   ("Rx Channel Defaults Size: %d, Segments %d\n",pTune->RxChannelSize,pTune->RxSegments));

  return(STATUS_SUCCESS);
}

/******************************************************************************
Name:
	ChipALAdapterInit

Function:
	Create new Instance of an adapter

Description:

******************************************************************************/
NDIS_STATUS ChipALAdapterInit(
    IN CDSL_ADAPTER_T	* CdslAdapter
)
{
  NDIS_STATUS Status = STATUS_FAILURE;
  CHIPAL_T * pThisInterface;
  WORD Index;
  enum WhichFunction FunctSel;
  WORD EventIndex;
  CHIPAL_INF_T * pChipALRemoteInf;

  CHIPAL_ASSERT_TRACE("ChipALAdapterInit() ", CdslAdapter);

  Status = ALLOCATE_MEMORY(pChipALRemoteInf, sizeof(*pChipALRemoteInf));
  if (Status != STATUS_SUCCESS)
    return(STATUS_FAILURE);

  CLEAR_MEMORY(pChipALRemoteInf, sizeof(*pChipALRemoteInf));

  CdslAdapter->pChipALRemoteInf = pChipALRemoteInf;

  // Use the local ChipAL Interface.
  pChipALRemoteInf->CALIntf.pCALRead  = lChipALRead;
  pChipALRemoteInf->CALIntf.pCALWrite = lChipALWrite;
  //pChipALRemoteInf->CALIntf.pCalEnableInterrupt = lChipALEnableInterruptHandler;
  //pChipALRemoteInf->CALIntf.pCalDisableInterrupt = lChipALDisableInterruptHandler;
  pChipALRemoteInf->CALIntf.pCalAddEventHandler    = lChipALAddEventHandler;
  pChipALRemoteInf->CALIntf.pCalRemoveEventHandler = lChipALRemoveEventHandler;
  pChipALRemoteInf->CALIntf.pCalSetEvent 	   = lChipALSetEvent;

  pChipALRemoteInf->CALIntf.iface.Context = CdslAdapter;
  pChipALRemoteInf->ConfigInterface.iface.Context = CdslAdapter;
  pChipALRemoteInf->ConfigInterface.pCalGetConfig = lChipAlGetConfig;
  pChipALRemoteInf->ConfigInterface.pCalSetConfig = lChipAlSetConfig;

  // The context for the link interface a link handle
  // the link handle is not available until LinkInit,
  // and LinkInit will save the link handle in the structure
  pChipALRemoteInf->LinkInterface.pCalLinkShutDown = lChipALLinkShutdown;
  pChipALRemoteInf->LinkInterface.pCalLinkEnable   = lChipALLinkEnable;
  pChipALRemoteInf->LinkInterface.pCalLinkDisable  = lChipALLinkDisable;
  pChipALRemoteInf->LinkInterface.pCalStartTx	   = lChipALStartTx;

  pThisInterface = chipALAllocateInterface();
  if (!pThisInterface)
    return(STATUS_FAILURE);

  // * Exchange SAP's
  pThisInterface->pNdisAdapter = CdslAdapter;
  pThisInterface->pNdisAdapter->ChipMod = pThisInterface;

  // * Initialize Event Handler List
  for (FunctSel = DSL_FUNCTION; FunctSel < MAX_FUNCTION; FunctSel++)
    for (Index = 0; Index < HW_EVENT_TYPE_END; Index++)
      for (EventIndex = 0; EventIndex < HW_EVENT_END; EventIndex++)
	InitializeListHead(&pThisInterface->EventHandler[FunctSel][Index][EventIndex]);

  // * Load Tuning Parameters
  if (!pThisInterface->pNdisAdapter->Tune) {
    NDIS_STATUS Status;

    Status = ALLOCATE_MEMORY(pThisInterface->pNdisAdapter->Tune,
			     sizeof(CHIPAL_TUNING_T));
    if (Status != STATUS_SUCCESS)
      return(STATUS_FAILURE);

    CLEAR_MEMORY(pThisInterface->pNdisAdapter->Tune, sizeof(CHIPAL_TUNING_T));
  }

  pThisInterface->Tune = pThisInterface->pNdisAdapter->Tune;

  // * Save External Event Handler
  pThisInterface->MgmtEventHandler = CardALChipUnknownIntrptEventHdlr;

  // * Load Bus Control Service
  pThisInterface->BusControl = &BUS_CONTROLLER_SERVICE;

  // * Allocate Bus Resources
  pThisInterface->BusResources = (* pThisInterface->BusControl->Init) (
					   pThisInterface->pNdisAdapter,
	                                   pThisInterface->Tune,
	                                   &MyAdapterCount,
					   CdslAdapter->pci_dev_arm,
					   CdslAdapter->pci_dev_adsl,
					   CdslAdapter->CnxtPhysicalDriverType);

  if (!pThisInterface->BusResources)
  {
    pThisInterface->pNdisAdapter->ChipMod = NULL;
    chipALReleaseInterface(pThisInterface);

    return(STATUS_FAILURE);
  }

#if 0
  // * Load Reverse Bit Order Table
  for (Index = 0;Index < BYTE_SIZE;++Index)
    ReverseByteTable[Index] = REVERSE_BYTE(Index);
#endif

  // * Load RegIo Handlers

  // Load Register I/O Controller Table
  COPY_MEMORY(&pThisInterface->RegIoControl[0], &IoDevice[0], sizeof(IoDevice));

  INIT_SPIN_LOCK(&pThisInterface->IsrLock);

  (* pThisInterface->BusControl->Enable)(pThisInterface->BusResources, ARM_FUNCTION, 0);
  (* pThisInterface->BusControl->Enable)(pThisInterface->BusResources, DSL_FUNCTION, 0);

  // * For each controller in the IoDevice table, call the RegIoContInit
  // * routine, if supplied.
  for (Index = 0; Index < MAX_HARDWARE_DEVICES; Index++)
  {
    INIT_SPIN_LOCK(&pThisInterface->RegIoLock[Index]);

    if (pThisInterface->RegIoControl[Index].RegIoContInit)
    {
      pThisInterface->RegIoContext[Index] =
	(* pThisInterface->RegIoControl[Index].RegIoContInit) (
				pThisInterface,
			        pThisInterface->RegIoControl[Index].FunctionSelect);

      // * Verify Controller Loaded Without Errors
      if (!pThisInterface->RegIoContext[Index])
      {
	CHIPAL_ASSERT_TRACE("Register IO Controller Load Failure", TRUE);
	ChipALAdapterHalt(CdslAdapter);
	return(STATUS_FAILURE);
      }
    }
  }

  // * Load Optional Services
  COPY_MEMORY(&pThisInterface->Service[0], &OptionalServices[0],
	      sizeof(OptionalServices));

#if CAL_SOFT_TIMER
  pThisInterface->CalTimerState = TMR_TIMING;

  INIT_TIMER(&pThisInterface->CalTimer, chipAlSoftTimer, (ULONG)pThisInterface);

  INIT_EVENT(&pThisInterface->CalTimerStoppedEvent);

  START_TIMER(&pThisInterface->CalTimer, SOFT_TIMER_PERIOD);
#endif

  // Return Adapter Count to Card Managment Adapter Structure
  CdslAdapter->InstanceNumber = (UCHAR) MyAdapterCount;

  return(STATUS_SUCCESS);
}

/******************************************************************************
Name:
	ChipALInitServices

Function:
	Initializes the adapter services

Description:
******************************************************************************/
NDIS_STATUS ChipALInitServices(
    IN CDSL_ADAPTER_T	* CdslAdapter)
{
  int Index;
  CHIPAL_T * pThisInterface = (CHIPAL_T *)CdslAdapter->ChipMod;

  // * For each service in the Service table, call the ServiceInit
  // * routine, if supplied.
  for (Index = 0; Index < MAX_CHIPAL_SERVICES; Index++)
  {
    if (pThisInterface->Service[Index].ServiceInit)
    {
      pThisInterface->ServiceContext[Index] =
	(* pThisInterface->Service[Index].ServiceInit)(pThisInterface);

      // * Verify Controller Loaded Without Errors
      if (!pThisInterface->ServiceContext[Index])
      {
	CHIPAL_ASSERT_TRACE("ChipAl Service Load Failure", TRUE);
	ChipALAdapterHalt(CdslAdapter);
	return(STATUS_FAILURE);
      }
    }
  }

  return STATUS_SUCCESS;
}

/******************************************************************************
Name:
	ChipALAdapterHalt

Function:
	Release all resources allocated by the Adapter

Description:
******************************************************************************/
NDIS_STATUS ChipALAdapterHalt(
    IN CDSL_ADAPTER_T		* CdslAdapter)
{
  CHIPAL_T 		*pThisInterface;
  WORD			Index;
  enum WhichFunction	FunctSel;
  WORD			EventIndex;
  CAL_EVENT_LIST_T	*ReleasedEvent;
  BOOLEAN		TimerCancelled;

  CHIPAL_ASSERT_TRACE("ChipALAdapterHalt() ", CdslAdapter);

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;
  if (!pThisInterface)
    return(STATUS_FAILURE);

#if CAL_SOFT_TIMER
  if (pThisInterface->CalTimerState == TMR_TIMING)
  {
    pThisInterface->CalTimerState = TMR_CANCELLED;

    if (!WAIT_EVENT_TIMEOUT(&pThisInterface->CalTimerStoppedEvent, ONE_SEC))
    {
      DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		       ChipALDebugFlag,
		       ("ChipALAdapterHalt: ERROR  Wait on Timer timed out!\n"));

      // Timer is still running.  Try to Cancel...
      CANCEL_TIMER(&pThisInterface->CalTimer, &TimerCancelled);

      // if still on queue - something is wrong
      if (!TimerCancelled)
	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
			 ChipALDebugFlag,
			 ("pThisAdapter: ERROR;  Could not remove timer!\n"));
    }
  }
#endif

  // * Release all Event Handlers
  for (FunctSel = DSL_FUNCTION; FunctSel < MAX_FUNCTION; FunctSel++)
    for (Index = 0; Index < HW_EVENT_TYPE_END; Index++)
      for (EventIndex = 0; EventIndex < HW_EVENT_END; EventIndex++)
	while (!IsListEmpty(&pThisInterface->EventHandler[FunctSel][Index][EventIndex]))
	{
	  ReleasedEvent = (CAL_EVENT_LIST_T *) RemoveTailList(
		&pThisInterface->EventHandler[FunctSel][Index][EventIndex]);

	  FREE_MEMORY(ReleasedEvent, sizeof(*ReleasedEvent),
		      CHIPAL_CONTEXT_MEM_FLAGS);
	}

  // * Release Optional Services
  for (Index = 0; Index < MAX_CHIPAL_SERVICES; Index++)
    if (pThisInterface->Service[Index].ServiceInit)
      (* pThisInterface->Service[Index].ServiceShutdown)(pThisInterface->ServiceContext[Index]);

  // * Release all RegIo Controllers
  for (Index = 0; Index < MAX_HARDWARE_DEVICES; Index++)
    if (pThisInterface->RegIoContext[Index])
      (* pThisInterface->RegIoControl[Index].RegIoShutdown)(
		pThisInterface->RegIoContext[Index],
		pThisInterface->RegIoControl[Index].FunctionSelect);

  // * Release Bus Resources
  if (pThisInterface->BusResources)
    (* pThisInterface->BusControl->Release)(CdslAdapter, pThisInterface->BusResources);

  // * Release Tuning Allocations
#if PROJECTS_NAME == PROJECT_NAME_CHEETAH_1
  if (pThisInterface->Tune->TxLoopBackFileA.MaximumLength > 0 &&
      pThisInterface->Tune->TxLoopBackFileA.Buffer)
  {
    FREE_MEMORY(pThisInterface->Tune->TxLoopBackFileA.Buffer,
		pThisInterface->Tune->TxLoopBackFileA.MaximumLength,
		CHIPAL_CONTEXT_MEM_FLAGS);

    pThisInterface->Tune->TxLoopBackFileA.MaximumLength = 0;
    pThisInterface->Tune->TxLoopBackFileA.Buffer = NULL;
  }

  if (pThisInterface->Tune->TxLoopBackFileB.MaximumLength > 0 &&
      pThisInterface->Tune->TxLoopBackFileB.Buffer)
  {
    FREE_MEMORY(pThisInterface->Tune->TxLoopBackFileB.Buffer,
		pThisInterface->Tune->TxLoopBackFileB.MaximumLength,
		CHIPAL_CONTEXT_MEM_FLAGS);
    pThisInterface->Tune->TxLoopBackFileB.MaximumLength = 0;
    pThisInterface->Tune->TxLoopBackFileB.Buffer = NULL;
  }

  if (pThisInterface->Tune->FpgaProgramFile.MaximumLength > 0 &&
      pThisInterface->Tune->FpgaProgramFile.Buffer)
  {
    FREE_MEMORY(pThisInterface->Tune->FpgaProgramFile.Buffer,
		pThisInterface->Tune->FpgaProgramFile.MaximumLength,
		CHIPAL_CONTEXT_MEM_FLAGS);

    pThisInterface->Tune->FpgaProgramFile.MaximumLength = 0;
    pThisInterface->Tune->FpgaProgramFile.Buffer = NULL;
  }
#endif

  // * Release Interface
  pThisInterface->pNdisAdapter->ChipMod = NULL;
  chipALReleaseInterface(pThisInterface);

  FREE_MEMORY(CdslAdapter->pChipALRemoteInf,sizeof(CHIPAL_REMOTE_INF), 0);

  MyAdapterCount--;

  return(STATUS_SUCCESS);
}

extern const ARM_CONTROL_REGISTER_T	ArmControlRegister[DEVICE_TYPE_MAX];

/******************************************************************************
Name:
	ChipALSetFirmGo

Function:
	Set ARM FIRMGO bit in Control Reg

Description:
******************************************************************************/
GLOBAL NTSTATUS ChipALSetFirmGo(
    IN CDSL_ADAPTER_T *	pThisAdapter)
{
  DWORD		Ctl_Val;
  NTSTATUS	Status;

  //Ctl_Val = 0x40; //TIGRIS_CSR_CTRL_FIRMGO
  Ctl_Val = ArmControlRegister[pThisAdapter->CnxtPhysicalDriverType].CsrCtrlFirmGo;

  Status = ChipALWrite(pThisAdapter,
		       P46_ARM,
		       ADDRESS_ARM_BAR_0,
		       CSR_ARM_CTRL,
		       sizeof(DWORD),
		       Ctl_Val,
		       Ctl_Val);

  return Status;
}

/******************************************************************************
Name:
	ChipALEnableInterruptHandler

Function:
	Enable Interrupts

Description:
******************************************************************************/
GLOBAL VOID ChipALEnableInterruptHandler(
    IN CDSL_ADAPTER_T *	ThisAdapter)
{
  CHIPAL_T 	* pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipALEnableInterruptHandler() ", ThisAdapter);

  pThisInterface = (CHIPAL_T *) ThisAdapter->ChipMod;

  if (pThisInterface &&
      pThisInterface->BusControl) {
    (* pThisInterface->BusControl->Enable)(pThisInterface->BusResources, ARM_FUNCTION, 0);
    (* pThisInterface->BusControl->Enable)(pThisInterface->BusResources, DSL_FUNCTION, 0);
  }
}

/******************************************************************************
Name:
	ChipALDisableInterruptHandler

Function:
	Disable Interrupts

Description:
******************************************************************************/
GLOBAL VOID ChipALDisableInterruptHandler(
    IN CDSL_ADAPTER_T *	ThisAdapter)
{
  CHIPAL_T 	* pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipALDisableInterruptHandler() ", ThisAdapter);

  pThisInterface = (CHIPAL_T *) ThisAdapter->ChipMod;

  if (pThisInterface &&
      pThisInterface->BusControl) {
    (pThisInterface->BusControl->Disable)(pThisInterface->BusResources, ARM_FUNCTION, DISABLE_INTERRUPT);
    (pThisInterface->BusControl->Disable)(pThisInterface->BusResources, DSL_FUNCTION, DISABLE_INTERRUPT);
  }
}

#if PROJECTS_NAME == PROJECT_NAME_TIGRIS
/******************************************************************************
FUNCTION NAME:
	ChipALLinkInit		(PROJECT_NAME_TIGRIS)

ABSTRACT:
	Open Full Duplex Data Stream

RETURN:
	ChipAL FailCode

DETAILS:
	This is a Project Specific Routine.  Legacy Versions are at the end of
	this file.

******************************************************************************/
GLOBAL DWORD ChipALLinkInit(
    CDSL_ADAPTER_T *	CdslAdapter,	// Adapter Instance
    CDSL_LINK_T *	CdslLink,	// Link Instance
    DWORD		Speed,		// Link Speed in BPS
    DWORD		ComChannel,	// Communications Channel
    BYTE		Mode		// Framed or NonFramed
)
{
  CHIPAL_T 	* pThisInterface;
  HW_CH_ENTRY_T ChanConfig;
  DWORD TxChannel;
  DWORD RxChannel;

  CHIPAL_ASSERT_TRACE("ChipALLinkInit() ", (CdslAdapter && CdslAdapter->ChipMod));

#if DBG
  TraceQueueIdx = 0;
  IntCount = 0;
#endif

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  if (!pThisInterface ||
      ComChannel <  MIN_COM_CHANNEL ||
      ComChannel >= MAX_COM_CHANNEL)
    return(CHIPAL_ERROR_FATAL(CHIPAL_PARAM_OUT_OF_RANGE));

  CdslAdapter->pChipALRemoteInf->LinkInterface.iface.Context = CdslLink;

  // Map FullDuplex Communications Channel to DMA Channel(s)
  switch(ComChannel)
  {
  case ADSL_CHANNEL:
    TxChannel = ADSL_DMA_CHAN_TX;
    RxChannel = ADSL_DMA_CHAN_RX;
    break;
  default:
    return(CHIPAL_ERROR_FATAL(CHIPAL_INVALID_PARAM));
    break;
  }

  // * Load DMA Control Service
  pThisInterface->DmaControl = &DMA_CHAN_CONTROLLER;

  //Begin TxTxTx
  // * Allocate Transmit Channel
  ChanConfig.ProcessData = FrameALGetTxFullBuffer;
  ChanConfig.TxComplete = FrameALPutTxEmptyBuffer;

  ChanConfig.UserLink = CdslLink;

  ChanConfig.SegmentSize = (WORD) DWORD_ALIGN(pThisInterface->Tune->TxChannelSize / pThisInterface->Tune->TxSegments);
  ChanConfig.NumberOfSegments = pThisInterface->Tune->TxSegments;

  ChanConfig.Description.ElUnion.Element.Direction = HOST_TO_DEVICE;
  ChanConfig.Description.ElUnion.Element.Channel = TxChannel;
  ChanConfig.Description.ElUnion.Element.Event = HW_TX_DMA_COMPLETE;

  pThisInterface->DmaChannel[TxChannel] = (* pThisInterface->DmaControl->ChanOpen)(pThisInterface, &ChanConfig);

  if (!pThisInterface->DmaChannel[TxChannel])
    return(CHIPAL_ERROR_FATAL(CHIPAL_RESOURCE_ALLOCATION_FAILURE));
  //End TxTxTx

  //Begin RxRxRx
  // * Allocate Receive Channel
  ChanConfig.ProcessData = FrameALPutRxFullBuffer;
  ChanConfig.TxComplete = NULL;

  ChanConfig.UserLink = CdslLink;

  ChanConfig.SegmentSize = (WORD) DWORD_ALIGN(pThisInterface->Tune->RxChannelSize / pThisInterface->Tune->RxSegments);
  ChanConfig.NumberOfSegments = pThisInterface->Tune->RxSegments;

  ChanConfig.Description.ElUnion.Element.Direction = DEVICE_TO_HOST;
  ChanConfig.Description.ElUnion.Element.Channel = RxChannel;
  ChanConfig.Description.ElUnion.Element.Event = HW_RX_DMA_COMPLETE;

  pThisInterface->DmaChannel[RxChannel] = (* pThisInterface->DmaControl->ChanOpen)(pThisInterface, &ChanConfig);

  if (!pThisInterface->DmaChannel[RxChannel])
    return(CHIPAL_ERROR_FATAL(CHIPAL_RESOURCE_ALLOCATION_FAILURE));
  //End RxRxRx

  return(CHIPAL_SUCCESS);
}
#endif

/******************************************************************************
FUNCTION NAME:
	lChipALLinkShutdown

ABSTRACT:
	Close Full Duplex Datastream.  If a Null link is supplied, ALL chipal
	datastreams will be closed.

RETURN:
	ChipAL FailCode

DETAILS:
******************************************************************************/
LOCAL NTSTATUS lChipALLinkShutdown(
    PVOID pLink,
    DWORD ComChannel
)
{
  CDSL_LINK_T	* pThisLink = pLink;
  CHIPAL_T 	* pThisInterface = NULL;
  SHORT		Index;
  SHORT		LinkReleased = 0;

  // * Find Chipal Adapter Context
  if (!pThisLink)
  {
    // * Link was not supplied.  Use last adapter instance if loaded.
    pThisInterface = pChipALHead;
    while (pThisInterface->Next)
      pThisInterface = pThisInterface->Next;

    if (!pThisInterface)
      return(CHIPAL_ERROR_FATAL(CHIPAL_PARAM_OUT_OF_RANGE));
  } else if (!pThisLink->pThisAdapter ||
	     !(CHIPAL_T *) pThisLink->pThisAdapter->ChipMod)
    return(CHIPAL_ERROR_FATAL(CHIPAL_PARAM_OUT_OF_RANGE));

  // * Get Chipal Adapter Context from Supplied link
  if (!pThisInterface)
    pThisInterface = (CHIPAL_T *) pThisLink->pThisAdapter->ChipMod;

  CHIPAL_ASSERT_TRACE("ChipALLinkShutdown() ", pThisInterface);

  // * Find and release all Links associated with a Hardware Channel
  for (Index = 0; Index < MAX_DATA_CHANNEL; Index++)
  {
    if (pThisInterface->DmaChannel[Index])
    {
      if (pThisLink == pThisInterface->DmaChannel[Index]->UserLink ||
	  !pThisLink)
      {
	// Disable Interrupts/Channel and Release Allocated Memory
	(* pThisInterface->DmaControl->ChanClose)(pThisInterface->DmaChannel[Index]);
	LinkReleased++;
	pThisInterface->DmaChannel[Index] = NULL;
      }
    }
  }

  if (LinkReleased)
    return(CHIPAL_SUCCESS);
  else
    return(CHIPAL_ERROR_FATAL(CHIPAL_INVALID_PARAM));
}

#if CAL_REVISION >= 1
/******************************************************************************
FUNCTION NAME:
	lChipALLinkEnable

ABSTRACT:
	Disable Full Duplex Data Link

RETURN:
	NTSTATUS:

DETAILS:
******************************************************************************/
LOCAL NTSTATUS lChipALLinkEnable(
    PVOID			pLink,
    DWORD			ComChannel
)
{
  CDSL_LINK_T *	CdslLink = pLink;
  CDSL_ADAPTER_T * CdslAdapter = CdslLink->pThisAdapter;
  CHIPAL_T 	* pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipALLinkEnable() ",
		      (CdslAdapter && CdslAdapter->ChipMod));

  CHIPAL_ASSERT_TRACE("ChipALLinkEnable() ",
		      (CdslAdapter && CdslAdapter->ChipMod));

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(ComChannel)
  {
  case ADSL_CHANNEL:
    if (!pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX] ||
	!pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX])
      return(-EINVAL);

    (* pThisInterface->DmaControl->ChanEnable)(pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX]);
    (* pThisInterface->DmaControl->ChanEnable)(pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX]);
    break;

  default:
    return(-EINVAL);
    break;
  }
  return(CHIPAL_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	lChipALLinkDisable

ABSTRACT:
	Enable Full Duplex Data Link

RETURN:
	NTSTATUS:

DETAILS:
******************************************************************************/
LOCAL NTSTATUS lChipALLinkDisable(
    PVOID		pLink,	        // Adapter Instance
    DWORD		ComChannel	// Communications Channel
)
{
  CDSL_LINK_T    * CdslLink = pLink;	// Adapter Instance
  CDSL_ADAPTER_T * CdslAdapter = CdslLink->pThisAdapter;
  CHIPAL_T       * pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipALLinkDIsable() ",
		      (CdslAdapter && CdslAdapter->ChipMod));

#if DBG
  TraceQueueIdx = 0;
  IntCount = 0;
#endif

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(ComChannel)
  {
  case ADSL_CHANNEL:
    if (!pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX] ||
	!pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX])
      return(-EINVAL);

    (* pThisInterface->DmaControl->ChanDisable)(pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX]);
    (* pThisInterface->DmaControl->ChanDisable)(pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX]);
    break;

  default:
    return(-EINVAL);
    break;
  }
  return(CHIPAL_SUCCESS);
}
#endif

/******************************************************************************
FUNCTION NAME:
	lChipALStartTx

ABSTRACT:
	 Restart the ChipAL Transmit process

RETURN:
	ChipAL Fail Code

DETAILS:
	The ChipAL Transmit function "Pulls" data from the protocol stack.
	Normally, this is done based on a Hardware Interrupt.  In the case
	where the stack does not have any data to deliver to the
	transmitter, the transmit process will terminate.  As a result, a
	function is required to restart the transmit cycle.  The cycle will
	again continue until the protocol stack fails to deliver data when
	requested.
******************************************************************************/
LOCAL NTSTATUS lChipALStartTx(
    PVOID pLink,
    DWORD ComChannel
)
{
  CDSL_LINK_T		*pCdslLink = pLink;
  CHIPAL_T 		*pThisInterface;// ChipAL Adapter Instance

  pThisInterface = (CHIPAL_T *) pCdslLink->pThisAdapter->ChipMod;
  if (!pThisInterface)
    return(CHIPAL_ERROR_FATAL(CHIPAL_PARAM_OUT_OF_RANGE));
  else {
    HW_CH_ENTRY_T	*pTxChannelLink;// ChipAL Link Instance
    CHAN_DESCRIPT_T	TxChannelDes;	// ChipAL Link Description
    CHAN_DESCRIPT_T	SearchMask;	// ChipAL Link Description Mask Value

    // Clear all Elements in Mask
    SearchMask.ElUnion.Value = 0;

    // Clear all Elements in Description
    TxChannelDes.ElUnion.Value = 0;

    // Indicate a search for the Transmit Channel
    TxChannelDes.ElUnion.Element.Direction = HOST_TO_DEVICE;

#if PROJECTS_NAME == PROJECT_NAME_CHEETAH_1
    TxChannelDes.ElUnion.Element.Port = PIO_DATA;
    SearchMask.ElUnion.Element.Port = PIO_DATA;
#endif

    // * We only care about the Direction and Port fields
    // * of the CHAN_DESCRIPT_T Structure
    SearchMask.ElUnion.Element.Direction = TRUE;

    pTxChannelLink = chipALFindLink(pThisInterface,
				    pCdslLink,
				    TxChannelDes,
				    SearchMask.ElUnion.Value);
    // * Verify we were able to locate the Transmit Channel
    if (!pTxChannelLink)
      return(CHIPAL_ERROR_FATAL(CHIPAL_PARAM_OUT_OF_RANGE));
    else {
      LIST_ENTRY	*pHeadList;
      CAL_EVENT_LIST_T	*pEventHandler;

      // * If defined, Call the transmit Data handler. This would normally be a
      // * result of an ISR / DPC call.

      // * Locate Head of the Data Events DPC call back list
      pHeadList = &pThisInterface->EventHandler[DSL_FUNCTION][HW_DPC_EVENT][pTxChannelLink->Description.ElUnion.Element.Event];

      for (pEventHandler = (CAL_EVENT_LIST_T *) pHeadList->pNext;
	   (LIST_ENTRY *) pEventHandler != pHeadList;
	   pEventHandler = (CAL_EVENT_LIST_T *) pEventHandler->HandlerList.pNext)
      {
	// * Check for any attached DPC routine(s) responsible for processing
	// * the data event on this link.  If no routines are assigned, an
	// * error is not recorded. This would only happen if the link were not
	// * initialized, which is the responsibility of the caller.
	if (pEventHandler->EventHandler)
	{
#if MAX_RESPONSE_SIZE
	  // * DPC Routines Reads Params. They should be considered Constant Values
	  (*pEventHandler->EventHandler)(pEventHandler->UserContext,
					 &pThisInterface->ResponseParams[pTxChannelLink->Description.ElUnion.Element.Event]);
#else
	  (*pEventHandler->EventHandler)(pEventHandler->UserContext);
#endif
	}
      }
    }
  }
  return STATUS_SUCCESS;
}

/******************************************************************************
Name:
	ChipALIsrHandler

Function:
	Interrupt Service Routine (Device IRQL)

Description:
******************************************************************************/
irqreturn_t ChipALIsrHandler(
    IN int		irq,
    IN PVOID		pAdapter,
    IN struct pt_regs	*regs
)
{
  CDSL_ADAPTER_T * ThisAdapter = pAdapter;
  CHIPAL_T       * pThisInterface;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  irqreturn_t handled = IRQ_NONE;
#endif

  pThisInterface = (CHIPAL_T *) ThisAdapter->ChipMod;

  if (pThisInterface->BusResources) {
    enum WhichFunction FunctSel;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    disable_irq(irq);
#endif

    for (FunctSel = DSL_FUNCTION; FunctSel <= ARM_FUNCTION; FunctSel++) {
      DWORD            InterruptChan;

      // * Read Interrupt Status and Mask registers
      (*pThisInterface->BusControl->GetStatus)(pThisInterface->BusResources,
					       FunctSel,
					       &InterruptChan);

      // * Determine if this interrupt is for our device and if it requires
      // * bottom half processing
      if (InterruptChan) {
	if (chipALClearInterrupt(pThisInterface, FunctSel, InterruptChan))
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	  tasklet_schedule(&ThisAdapter->InterruptQueStr[FunctSel]);
	handled = IRQ_HANDLED;
#else
	{
	  queue_task(&ThisAdapter->InterruptQueStr[FunctSel], &tq_immediate);
	  mark_bh(IMMEDIATE_BH);	// the immediate bottom half
	}
#endif
      }
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    enable_irq(irq);
#endif
  }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  return handled;
#endif
}

/******************************************************************************
Name:
	ChipALIsrDPCHandler

Function:
	Handler Interrupt Event

Description:
******************************************************************************/
LOCAL VOID ChipALIsrDPCHandler(
    IN IsrHandlerArg pAdapter,
    enum WhichFunction FunctSel)
{
  CDSL_ADAPTER_T	*ThisAdapter = (CDSL_ADAPTER_T *)pAdapter;
#if DBG_DPC_TRACE
  DWORD		RegValue;
#endif

  TRACE_MESSAGE(DPC_ENTRY);

  if (ThisAdapter->ChipMod) {
    CHIPAL_T		* pThisInterface = (CHIPAL_T *) ThisAdapter->ChipMod;
    BUS_RESOURCE_T	* pBusResource = pThisInterface->BusResources;

    // * Make sure BusResources are defined and Interrupts are Initialized.
    if (pBusResource &&
	pBusResource->InterruptsInitiailized) {
      DWORD		InterruptReg[HW_MAX_EVENT_REGISTERS];
      BOOL		InterruptFound;

      do
      {
	WORD	Index;
	LOCK_FLAG(LockFlag);

	// * Enter Critical Section.

	// Protect against other threads/MultiProcessor
	ACQUIRE_LOCK(&pThisInterface->IsrLock,LockFlag);

	InterruptFound = FALSE;

	for (Index = 0; Index < HW_MAX_EVENT_REGISTERS; Index++)
	{
	  //Get a Copy of the Interrupt Status Register
	  InterruptReg[Index] = pThisInterface->InterruptStatus[FunctSel][Index];

	  if (InterruptReg[Index])
	  {
	    // Clear Static Interrupt Value - (Only the bits copied)
	    InterruptFound = TRUE;
	    pThisInterface->InterruptStatus[FunctSel][Index] &= ~InterruptReg[Index];
	  }
	}

	RELEASE_LOCK(&pThisInterface->IsrLock,LockFlag);

	// * Exit Critical Section

	// * Process Interrupts. NOTE: InterruptReg is a copy of the actual
	// * interrupt Status register. Therefore, if an interrupt occurs while
	// * we are processing this group of interrupts, it is not a problem.
	if (InterruptFound) {
	  CAL_EVENT_LIST_T	* pEventHandler;
	  LIST_ENTRY		* pHeadList;
#ifdef DBG
	  static DWORD	UnhandledEvents = 0;
#endif

	  for (Index = 0; Index < HW_MAX_EVENT_REGISTERS; Index++)
	  {
	    DWORD		myInterruptReg = InterruptReg[Index];
	    DWORD		EventIndex;

	    // * Process All Bits in the Event Register
	    for (EventIndex = 0;
		 myInterruptReg && (EventIndex < BITS_PER_EVENT_REG);
		 EventIndex++)
	    {
	      DWORD Mask = 1 << EventIndex;

	      // * If Bit is Set, Check for ISR Event Handlers
	      if (Mask & myInterruptReg)
	      {
		BOOL	EventHandled = FALSE;
		DWORD CurrentEvent;

		// Encode Register Indexes into an Event
		CurrentEvent = BITS_PER_EVENT_REG * Index + EventIndex;

#ifdef DBG
		pThisInterface->IntStats[CurrentEvent].InterruptCount++;
#endif

		// * Execute Channel Handler if Defined
		pHeadList = &pThisInterface->EventHandler[FunctSel][HW_DPC_EVENT][CurrentEvent];

		// * Check for any attached DPC routines responsible for 
		// * processing an interrupt.

		for (pEventHandler = (CAL_EVENT_LIST_T *) pHeadList->pNext;
		     (LIST_ENTRY *) pEventHandler != pHeadList;
		     pEventHandler = (CAL_EVENT_LIST_T *) pEventHandler->HandlerList.pNext)
		{
		  EventHandled = TRUE;

		  if (pEventHandler->EventHandler)
		  {
#if MAX_RESPONSE_SIZE
		    // ChipAl supplies Response Param(s) to ISR / DPC Events
		    // *  DPC Routines Reads Params. They should be considered
		    // * Constant Values
		    (*pEventHandler->EventHandler)(pEventHandler->UserContext,
				&pThisInterface->ResponseParams[CurrentEvent]);
#else
		    (*pEventHandler->EventHandler)(pEventHandler->UserContext);
#endif
		  }
		}

		if (!EventHandled)
		{
		  // * This is an Enabled Interrupt that does not have a Channel
		  // * Event Handler.  Call the Management Handler if available.
		  // * Otherwise, Ignore - Take NO ACTION.
		  if (pThisInterface->MgmtEventHandler)
		  {
#if REPORT_UNHADLED_EVENTS
		    (*pThisInterface->MgmtEventHandler)
		      (pThisInterface->pNdisAdapter, CurrentEvent);
#endif

#ifdef DBG
		    ++UnhandledEvents;
#endif
		  }
		}

		// Clear current interrupt bit to allow early exit of loop
		myInterruptReg &= ~Mask;
	      } // If Mask & myInterruptReg
	    } // For EventIndex
	  } // For Index
	}
      } while (InterruptFound);
    }
  }
  TRACE_MESSAGE(DPC_EXIT);
}

/******************************************************************************
Name:
	ChipALDSLIsrDPCHandler

Function:
	Handler Interrupt Event

Description:
******************************************************************************/
VOID ChipALDSLIsrDPCHandler(
    IN IsrHandlerArg pAdapter)
{
  ChipALIsrDPCHandler(pAdapter, DSL_FUNCTION);
}

/******************************************************************************
Name:
	ChipALARMIsrDPCHandler

Function:
	Handler Interrupt Event

Description:
******************************************************************************/
VOID ChipALARMIsrDPCHandler(
    IN IsrHandlerArg pAdapter
)
{
  ChipALIsrDPCHandler(pAdapter, ARM_FUNCTION);
}

#if 0
/******************************************************************************
FUNCTION NAME:
	ChipALSetInformation

ABSTRACT:
	Requirment of the Project Architecture to process NDIS OIDs

RETURN:
	NDIS_STATUS

DETAILS:
******************************************************************************/
GLOBAL NDIS_STATUS ChipALSetInformation(
    IN CDSL_ADAPTER_T	* pThisAdapter,
    IN PVOID		InfoBuffer,
    IN ULONG		InfoBufferLength,
    OUT PULONG		BytesWritten,
    OUT PULONG		BytesNeeded)
{
  //	return NDIS_STATUS_INVALID_OID;
  return STATUS_FAILURE;
}
#endif

/******************************************************************************
FUNCTION NAME:
	ChipALRead

ABSTRACT:
	Perform a Device write

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
	The Read request is mapped to the responsible Register I/O control
	module. If a module does not require any special functions, the vector
	table entry will be mapped to a default memory Read routine.
******************************************************************************/
LOCAL CHIPAL_FAILCODE_T  lChipALRead(
    VOID		* pAdapter,
    HARDWARE_DEVICE_T	HwDevice,
    DWORD		LogicalDevice,
    DWORD		OffsetAddress,
    DWORD		RegisterLength,
    PVOID		pVal)
{
  CDSL_ADAPTER_T	* CdslAdapter = pAdapter;
  DWORD /*volatile*/	* pValue = (DWORD /*volatile*/ *)pVal;
  CHIPAL_T		* pThisInterface;
  CHIPAL_FAILCODE_T	ReturnValue;
  PVOID			BaseAddress;

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(LogicalDevice)
  {
  case ADDRESS_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR0;
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenBar0 ||
	!BaseAddress)
      return(-EINVAL);
#endif
    break;

  case ADDRESS_BAR_1:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR1;
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenBar1 ||
	!BaseAddress)
      return(-EINVAL);
#endif
    break;

  case ADDRESS_ARM_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR0;
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenARMBar0 ||
	!BaseAddress)
      return(-EINVAL);
#endif
    break;

  case REMOTE_ADDRESS_BAR_2:
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR0;
    break;

  default:
    return(CHIPAL_PARAM_OUT_OF_RANGE);
    break;
  }

  ReturnValue = (* pThisInterface->RegIoControl[HwDevice].RegIoRead)(
			pThisInterface->RegIoContext[HwDevice],
	                BaseAddress,
	                OffsetAddress,
	                RegisterLength,
	                pValue);

  return(ReturnValue);
}

/******************************************************************************
FUNCTION NAME:
	ChipALWrite

ABSTRACT:
	Perform a Device write

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
	The write request is mapped to the responsible Register I/O control 
	module. If a module does not require any special functions, the vector 
	table entry will be mapped to a default memory write routine.
******************************************************************************/
LOCAL CHIPAL_FAILCODE_T  lChipALWrite(
    VOID		* pAdapter,
    HARDWARE_DEVICE_T	HwDevice,
    DWORD		LogicalDevice,
    DWORD		OffsetAddress,
    DWORD		RegisterLength,
    DWORD		Value,
    DWORD		Mask)
{
  CDSL_ADAPTER_T	* CdslAdapter = pAdapter;
  CHIPAL_T		* pThisInterface;
  CHIPAL_FAILCODE_T	ReturnValue;
  PVOID			BaseAddress;
  enum WhichFunction	FunctSel;

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(LogicalDevice)
  {
  case ADDRESS_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR0;
    FunctSel = DSL_FUNCTION;
    //		DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //			         ChipALDebugFlag,
    //			         (DEV_LABEL "  command add bar 0 %lx",(DWORD)BaseAddress));
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenBar0 ||
	!BaseAddress)
      return(-EINVAL);
#endif
    break;

  case ADDRESS_BAR_1:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR1;
    //		DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //			         ChipALDebugFlag,
    //			         (DEV_LABEL "  command add bar 1 %lx",(DWORD)BaseAddress));
    FunctSel = DSL_FUNCTION;
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenBar1 ||
	!BaseAddress)
      return(-EINVAL);
#endif
    break;

  case ADDRESS_ARM_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR0;
    //		DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //			         ChipALDebugFlag,
    //			         (DEV_LABEL "  command add arm bar 0 %lx",(DWORD)BaseAddress));
    FunctSel = ARM_FUNCTION;
#ifdef DBG
    if (OffsetAddress >= pThisInterface->BusResources->LenARMBar0 ||
	!BaseAddress)
    {
      DBG_CDSL_DISPLAY(DBG_LVL_GENERIC,
		       DBG_FLAG_DONT_CARE,
		       (DEV_LABEL " invalid offset %lx  from %lx",
			(DWORD)OffsetAddress,
			(DWORD)BaseAddress));
      return(-EINVAL);
    }
#endif
    break;

  case REMOTE_ADDRESS_BAR_2:
    FunctSel = ARM_FUNCTION;
    //		DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //			         ChipALDebugFlag,
    //			         (DEV_LABEL "  command add remote bar 0 %lx",(DWORD)BaseAddress));
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR0;
    break;

  default:
    //		DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //			         ChipALDebugFlag,
    //			         (DEV_LABEL " device out of range",(DWORD)BaseAddress));
    return(CHIPAL_PARAM_OUT_OF_RANGE);
    break;
  }

  {
    LOCK_FLAG_Z(LockFlag); /* _Z To quieten compiler */

    // * If a Mask value is present, this is a Read+Modify+Write cycle.
    // * Therefore, it will be the beginning of a CRITICAL section.  All
    // * Interrupts must be Disabled.
    if (Mask)
      ACQUIRE_LOCK(&pThisInterface->IsrLock, LockFlag);

    //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //		         ChipALDebugFlag,
    //		         (DEV_LABEL " issue command",(DWORD)BaseAddress));
    ReturnValue = (* pThisInterface->RegIoControl[HwDevice].RegIoWrite)(
			pThisInterface->RegIoContext[HwDevice],
			BaseAddress,
			OffsetAddress,
			RegisterLength,
			Mask,
			Value);

    // * If Mask, End Critical Section.
    if (Mask)
      RELEASE_LOCK(&pThisInterface->IsrLock, LockFlag);
  }

#if BUS_BURST_NOT_ALLOWED
  {
    volatile DWORD	RegisterValue;	// Accesses to this variable get optimized
    // out without the keyword volatile

    // * This is a 'DUMMY' read.  The read operation prevents two back to back
    // * writes, which may could lead to a burst mode operation.
    RegisterValue = *((DWORD *)((BYTE *)BaseAddress + BASIC2_STATUS));
  }
#endif

  return(ReturnValue);
}

#if 0
/******************************************************************************
FUNCTION NAME:
	ChipALReadSync

ABSTRACT:
	Perform	synchronous Device read.  

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
	This routine will stay in a dead loop until the	device becomes ready.

******************************************************************************/
GLOBAL CHIPAL_FAILCODE_T  ChipALReadSync(
    CDSL_ADAPTER_T		* CdslAdapter,
    HARDWARE_DEVICE_T		HwDevice,
    DWORD			LogicalDevice,
    DWORD			OffSetAddress,
    DWORD			RegisterLength,
    DWORD 			* pValue)
{
  CHIPAL_T		* pThisInterface;
  CHIPAL_FAILCODE_T	ReturnValue;
  PVOID			BaseAddress;
  LOCK_FLAG(LockFlag);
#ifdef DBG
  DWORD			MaxLoop	= 0;
#endif

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(LogicalDevice)
  {
  case ADDRESS_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR0;
    break;
  case ADDRESS_BAR_1:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR1;
    break;
  case ADDRESS_ARM_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR0;
    break;
  default:
    return(CHIPAL_PARAM_OUT_OF_RANGE);
    break;
  }

  if (!BaseAddress)
    return(-EINVAL);

  ACQUIRE_LOCK(&pThisInterface->RegIoLock[HwDevice],LockFlag);

  // *	Wait for device	to become ready
  do
  {
    ReturnValue	= (* pThisInterface->RegIoControl[HwDevice].RegIoRead)(
				pThisInterface->RegIoContext[HwDevice],
				BaseAddress,
				OffSetAddress,
				RegisterLength,
				pValue);

#ifdef DBG
    if (MaxLoop++ > MAX_SYNC_WAIT)
    {
      ReturnValue = -EBUSY;
      break;
    }
#endif
  } while(ReturnValue == -ENOTREADY);

  RELEASE_LOCK(&pThisInterface->RegIoLock[HwDevice],LockFlag);

  return(ReturnValue);
}
#endif

#if 0
/******************************************************************************
FUNCTION NAME:
	ChipALWriteSync

ABSTRACT:
	Perform	synchronous Device write.  This	requires the completion of the
	write cycle with a STATUS_SUCCESS or a fail code other than -EBUSY,
	or STATUS_DEVICE_PENDING.

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
	This routine will stay in a dead loop until the device becomes ready.

******************************************************************************/
GLOBAL CHIPAL_FAILCODE_T  ChipALWriteSync(
    CDSL_ADAPTER_T	* CdslAdapter,
    HARDWARE_DEVICE_T	HwDevice,
    DWORD		LogicalDevice,
    DWORD		OffSetAddress,
    DWORD		RegisterLength,
    VOID		* UserValue,
    DWORD		MaskValue)
{
  CHIPAL_T		* pThisInterface;
  CHIPAL_FAILCODE_T	ReturnValue;
  VOID			* pRegister;
  DWORD			RegisterValue;
  PVOID			BaseAddress;
  LOCK_FLAG(LockFlag);
  CHIPAL_DEVICE_STATUS_T	DeviceStatus;
#ifdef DBG
  DWORD			MaxLoop	= 0;
#endif

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;

  switch(LogicalDevice)
  {
  case ADDRESS_BAR_0:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR0;
    break;
  case ADDRESS_BAR_1:
    BaseAddress = pThisInterface->BusResources->VirtualAddressBAR1;
    break;
  case ADDRESS_ARM_BAR_1:
    BaseAddress = pThisInterface->BusResources->VirtualAddressARMBAR1;
    break;
  default:
    return(CHIPAL_PARAM_OUT_OF_RANGE);
    break;
  }

  if (!BaseAddress)
    return(-EINVAL);

  ACQUIRE_LOCK(&pThisInterface->RegIoLock[HwDevice], LockFlag);

  // *	Wait for device	to become ready
  do
  {
    ReturnValue = (* pThisInterface->RegIoControl[HwDevice].RegIoWrite)(
			pThisInterface->RegIoContext[HwDevice],
			BaseAddress,
			OffSetAddress,
			RegisterLength,
			MaskValue,
			UserValue);
#ifdef DBG
    if (MaxLoop++ > MAX_SYNC_WAIT)
    {
      ReturnValue = -EBUSY;
      break;
    }
#endif
  } while (ReturnValue == -EBUSY);

  RELEASE_LOCK(&pThisInterface->RegIoLock[HwDevice], LockFlag);

  if (ReturnValue == STATUS_PENDING)
    ReturnValue = STATUS_SUCCESS;

  return(ReturnValue);
}
#endif

/******************************************************************************
FUNCTION NAME:
	ChipALService

ABSTRACT:
	Entry Point for Optional Chipal Services

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
	
******************************************************************************/
GLOBAL CHIPAL_FAILCODE_T ChipALService(
    CDSL_ADAPTER_T	* pThisAdapter,
    CHIPAL_SERVICE_T	Service,
    DWORD		Function,
    CAL_MODULE_PARAMS_T	* Parameter
)
{
  CHIPAL_T		* pThisInterface;
  CHIPAL_FAILCODE_T	Results;

  CHIPAL_ASSERT_TRACE("ChipALService() ",
		      (pThisAdapter &&
		       (Service >= MIN_CHIPAL_SERVICES &&
			Service < MAX_CHIPAL_SERVICES) &&
		       (Function >= MIN_CHIPAL_SERVICES_FN &&
			Function < MAX_CHIPAL_SERVICES_FN) &&
		       Parameter));

  // * Critical Entry Point.  Must also do Runtime check.
  if (!(pThisAdapter &&
	(Service >= MIN_CHIPAL_SERVICES &&
	 Service < MAX_CHIPAL_SERVICES) &&
	(Function >= MIN_CHIPAL_SERVICES_FN &&
	 Function < MAX_CHIPAL_SERVICES_FN) &&
	Parameter))
    return(CHIPAL_ERROR_FATAL(CHIPAL_INVALID_PARAM));

  pThisInterface = (CHIPAL_T *) pThisAdapter->ChipMod;
  Parameter->Context = pThisInterface->ServiceContext[Service];

  switch(Function)
  {
  case SVC_COMMAND:
    Results = (* pThisInterface->Service[Service].ServiceCommand)(Parameter);
    break;

  case SVC_CONFIGURE:
    Results = (* pThisInterface->Service[Service].ServiceConfigure)(Parameter);
    break;

  default:
    Results = CHIPAL_ERROR_FATAL(CHIPAL_INVALID_PARAM);
    break;
  }
  return(Results);
}

/******************************************************************************
FUNCTION NAME:
	ChipALAddEventHandler

ABSTRACT:
	Add Event Handler for any ChipAL Event (Interrupt).  This allows
	external entities to attach an Event Handler to specific Chipal Event.
	The Handler may be called at ISR or DPC time.

RETURN:
	ChipAL Fail Code:
		CHIPAL_SUCCESS - Event Registered Successfully
		CHIPAL_RESOURCE_ALLOCATION_FAILURE - Allocate Memory Failed
		CHIPAL_MODULE_NOT_INITIALIZED - ChipAl is not Initialized
DETAILS:
******************************************************************************/
LOCAL CHIPAL_FAILCODE_T lChipALAddEventHandler(
    IN PVOID			pAdapter,	// Driver Adapter Instance
    IN DWORD			FunctSelect,	// selects arm or dsl function
    IN CHIPAL_EVENT_T		Event,		// Enumerated Event to Set
    IN HW_EVENT_TYPE_T		IrqLevel,	// When Event should be called - DIRQl or Dispatch IRQL
    IN CHIPAL_EVENT_HANDLER_T	UserEventHandler,// Function to Call
    IN void			* UserContext	// Function Context - Unused by ChipAL
)
{
  IN CDSL_ADAPTER_T 	* CdslAdapter = pAdapter;
  IN enum WhichFunction	FunctSel = FunctSelect;
  CHIPAL_T 		* pThisInterface;
  NDIS_STATUS		Status;
  CAL_EVENT_LIST_T	* NewHandler;

  CHIPAL_ASSERT_TRACE("ChipALAddEventHandler() ",
		      (CdslAdapter &&
		       Event >= HW_EVENT_BEGIN &&
		       Event < HW_EVENT_END &&
		       HW_EVENT_END == HW_MAX_EVENTS &&
		       UserEventHandler &&
		       IrqLevel >= HW_EVENT_TYPE_BEGIN &&
		       IrqLevel < HW_EVENT_TYPE_END));

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;
  if (!pThisInterface)
    return(CHIPAL_ERROR_FATAL(CHIPAL_MODULE_NOT_INITIALIZED));

  // * Allocate Memory for New Event Handler
  Status = ALLOCATE_MEMORY(NewHandler, sizeof(*NewHandler));
  if (Status != STATUS_SUCCESS)
    return(CHIPAL_ERROR_FATAL(CHIPAL_RESOURCE_ALLOCATION_FAILURE));

  CLEAR_MEMORY(NewHandler, sizeof(*NewHandler));

  // * Load Handler Structure and Add to Event's Linked List of Handlers
  NewHandler->EventHandler = UserEventHandler;
  NewHandler->UserContext  = UserContext;

  InsertHeadList(&pThisInterface->EventHandler[FunctSel][IrqLevel][Event],
		 &NewHandler->HandlerList);

  return(CHIPAL_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	ChipALRemoveEventHandler

ABSTRACT:
	Remove Event Handler for Any ChipAL Event (Interrupt)

RETURN:
	ChipAL Fail Code:
	CHIPAL_SUCCESS - Handler Found and Released
	CHIPAL_MODULE_NOT_INITIALIZED - ChipAL is not Initialized
	CHIPAL_PARAM_OUT_OF_RANGE - - Could Not locate indicated handler.

DETAILS:
******************************************************************************/
LOCAL CHIPAL_FAILCODE_T lChipALRemoveEventHandler(
    IN PVOID			pAdapter,	// Driver Adapter Instance
    IN DWORD			FunctSelect,	// selects the ARM or DSL function
    IN CHIPAL_EVENT_T		Event,		// Event to Remove
    IN HW_EVENT_TYPE_T		IrqLevel,	// When Event was being called
    IN CHIPAL_EVENT_HANDLER_T	UserEventHandler// Handler to remove
)
{
  IN CDSL_ADAPTER_T 	* CdslAdapter = pAdapter;
  IN enum WhichFunction	FunctSel = FunctSelect;
  CHIPAL_T 		* pThisInterface;
  LIST_ENTRY		* HandlerEntry;
  BOOL			FoundHandler = FALSE;

  CHIPAL_ASSERT_TRACE("ChipALRemoveEventHandler() ",
		      (CdslAdapter &&
		       Event >= HW_EVENT_BEGIN &&
		       Event < HW_EVENT_END &&
		       UserEventHandler &&
		       IrqLevel >= HW_EVENT_TYPE_BEGIN  &&
		       IrqLevel < HW_EVENT_TYPE_END));

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;
  if (!pThisInterface)
    return(CHIPAL_ERROR_FATAL(CHIPAL_MODULE_NOT_INITIALIZED));

  // Get First Entry on Event's Linked List
  HandlerEntry = pThisInterface->EventHandler[FunctSel][IrqLevel][Event].pNext;

  // * Search All Handlers on List for the Indicated User Handler
  while (HandlerEntry !=
	 &pThisInterface->EventHandler[FunctSel][IrqLevel][Event])
  {
    if (((CAL_EVENT_LIST_T *)HandlerEntry)->EventHandler == UserEventHandler)
    {
      RemoveListEntry(HandlerEntry);

      FREE_MEMORY(HandlerEntry, sizeof(CAL_EVENT_LIST_T),
		  CHIPAL_CONTEXT_MEM_FLAGS);

      FoundHandler = TRUE;
      break;
    } else
      HandlerEntry = HandlerEntry->pNext;
  }

  if (FoundHandler)
    return(CHIPAL_SUCCESS);
  else
    return(CHIPAL_ERROR_WARNING(CHIPAL_PARAM_OUT_OF_RANGE));
}

/******************************************************************************
FUNCTION NAME:
	ChipALSetEvent

ABSTRACT:
	Convert an Enumerated event to a Register / Event index. Use these
	indexes to set a bit in the static interrupt event register that is 
	processed at DPC time.

RETURN:
	void

DETAILS:
	Optionally, if the event is one that was decoded from a bit in the
	real hardware register, a bit will be set in the ISR event register
	that is currently being processed.  Therefore, any decoded event set
	in this manner should be numerically greater than the event that was
	used in the event decode process.  For efficiency, there are no checks
	for this potential problem.

NOTE: 
	This code is expected to run in an ISR routine.  For optimization, it
	is an INLINE procedure. In a release version, Events will NOT be range
	checked. Passing invalid parameters will certainly cause a crash.
******************************************************************************/
LOCAL void lChipALSetEvent(
    PVOID			pAdapter,
    DWORD			FunctSelect,
    CHIPAL_EVENT_T		Event)
{
  CDSL_ADAPTER_T 	* pThisAdapter = pAdapter;
  enum WhichFunction	FunctSel = FunctSelect;
  HW_EVENT_INDEX_T	Index;
  CHIPAL_T		* pThisInterface;

  Index.EvntUnion.Value = (DWORD) Event;

#if VERIFY_EVENTS
  if (!(Event >= HW_EVENT_BEGIN &&
	Event < HW_EVENT_END &&
	pThisAdapter &&
	pThisAdapter->ChipMod))
    return;
#endif

  pThisInterface = (CHIPAL_T *)pThisAdapter->ChipMod;

  // Set Bit in DPC structure to allow deferred processing
  pThisInterface->InterruptStatus[FunctSel][Index.EvntUnion.Bits.Register] |=
    1 << Index.EvntUnion.Bits.Event;

  // * If bit is not reflected in the real hardware register, post an event
  // * to the IsrEvent structure.
  if (Index.EvntUnion.Bits.Register != HARDWARE_INTERRUPT_REGISTER)
    pThisInterface->IsrEvent[FunctSel][Index.EvntUnion.Bits.Register] |=
      1 << Index.EvntUnion.Bits.Event;
}

/******************************************************************************
FUNCTION NAME:
	ChipAlGetConfig

ABSTRACT:
	Get Current Configuration of ChipAl Module

RETURN:
	CHIPAL_FAILCODE_T - Fail Code as defined by Called Module;
	CHIPAL_MODULE_NOT_INITIALIZED - ChipAl is not Initialized
	CHIPAL_SERVICE_UNAVAILABLE - Called Module does not support GetConfig 
		Function

DETAILS:
	
******************************************************************************/
LOCAL CHIPAL_FAILCODE_T	lChipAlGetConfig(
    IN  VOID			* pAdapter,	// Adapter Instance
    I_O	VOID			* pConf)
{
  CDSL_ADAPTER_T 	* CdslAdapter = pAdapter;
  CAL_CONFIG_T		* Config = pConf;
  CHIPAL_T		* pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipAlGetConfig() ",
		      (CdslAdapter && Config != NULL_CAL_CONFIG));

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;
  if (!pThisInterface)
    return (CHIPAL_ERROR_FATAL(CHIPAL_MODULE_NOT_INITIALIZED));
  else
    switch(Config->ModuleType)
    {
    case MODULE_TYPE_BUS_CONTROLLER:
      if (pThisInterface->BusControl->GetConfig)
	return (* pThisInterface->BusControl->GetConfig)(CdslAdapter,&Config->Params);
      /* Fall thru */
    case MODULE_TYPE_IO_CONTROLLER:
    case MODULE_TYPE_DMA_CONTROLLER:
    case MODULE_SERVICE:
    default:
      return (CHIPAL_ERROR_WARNING(CHIPAL_SERVICE_UNAVAILABLE));
      break;
    }
}

/******************************************************************************
FUNCTION NAME:
	ChipAlSetConfig

ABSTRACT:
	Set Current Configuration of ChipAl Module

RETURN:
	CHIPAL_FAILCODE_T - Fail Code as defined by Called Module;
	CHIPAL_MODULE_NOT_INITIALIZED - ChipAl is not Initialized
	CHIPAL_SERVICE_UNAVAILABLE - Called Module does not support SetConfig 
		Function

DETAILS:
	
******************************************************************************/
LOCAL	CHIPAL_FAILCODE_T	lChipAlSetConfig(
    IN  VOID	 		* pAdapter,	// Adapter Instance
    I_O	VOID			* pConf)
{
  CDSL_ADAPTER_T 	* CdslAdapter = pAdapter;
  CAL_CONFIG_T		* Config = pConf;
  CHIPAL_T		* pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipAlGetConfig() ",
		      (CdslAdapter && Config != NULL_CAL_CONFIG));

  pThisInterface = (CHIPAL_T *) CdslAdapter->ChipMod;
  if (!pThisInterface)
    return(CHIPAL_ERROR_FATAL(CHIPAL_MODULE_NOT_INITIALIZED));
  else
    switch(Config->ModuleType)
    {
    case MODULE_TYPE_BUS_CONTROLLER:
      if (pThisInterface->BusControl->SetConfig)
	return (* pThisInterface->BusControl->GetConfig)(CdslAdapter,&Config->Params);
      /* Fall thru */
    case MODULE_TYPE_IO_CONTROLLER:
    case MODULE_TYPE_DMA_CONTROLLER:
    case MODULE_SERVICE:
    default:
      return(CHIPAL_ERROR_WARNING(CHIPAL_SERVICE_UNAVAILABLE));
      break;
    }
}

#if CAL_SOFT_TIMER
/******************************************************************************
FUNCTION NAME:
	chipAlSoftTimer

ABSTRACT:
	Provides Software timer when Hardware timer is not available. This
	timer will call all functions attached to HW_GENERAL_TIMER_A.  The
	period of the timer is determined by SOFT_TIMER_PERIOD.

RETURN:
	VOID

DETAILS:
******************************************************************************/
VOID chipAlSoftTimer(
    IN ULONG	pInterf
)
{
  CHIPAL_T		* pThisInterface = (CHIPAL_T*)pInterf;
  CAL_EVENT_LIST_T	* pEventHandler;
  DWORD 		CurrentEvent = 0;
  BOOL			EventHandled;
  LIST_ENTRY		* pHeadList;
  static BYTE		Temp = 0;/*CLN*/
  static DWORD		Count = 0;/*CLN*/

  if (pThisInterface) {
    if (pThisInterface->CalTimerState != TMR_TIMING) {
      // Tell Halt ok to proceed - we have exited
      SET_EVENT(&pThisInterface->CalTimerStoppedEvent);
    } else {
      switch(Temp) {
      case 1:
	ChipALLinkEnable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	Temp = 0;
	break;
      case 2:
	ChipALLinkDisable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	Temp = 0;
	break;
      case 3:
	Count++;
	if (Count == 1)
	  ChipALLinkDisable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	if (Count == 200) {
	  ChipALLinkEnable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	  Count = 0;
	}
	break;
      case 4:
	Count++;
	if (Count == 500)
	  ChipALLinkDisable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	else if (Count == 1000) {
	  ChipALLinkEnable(pThisInterface->pNdisAdapter,ADSL_CHANNEL);
	  Count = 0;
	}
	break;
      }

      EventHandled = FALSE;

      pHeadList = &pThisInterface->EventHandler[DSL_FUNCTION][HW_DPC_EVENT][HW_GENERAL_TIMER_A];

      for (pEventHandler = (CAL_EVENT_LIST_T *) pHeadList->pNext;
	   (LIST_ENTRY *) pEventHandler != pHeadList;
	   pEventHandler = (CAL_EVENT_LIST_T *) pEventHandler->HandlerList.pNext) {
	EventHandled = TRUE;

	if (pEventHandler->EventHandler) {
#if MAX_RESPONSE_SIZE	// ChipAl supplies Response Param(s) to ISR / DPC Events
			// *  DPC Routines Reads Params. They should be considered Constant Values
	  (*pEventHandler->EventHandler)(pEventHandler->UserContext, &pThisInterface->ResponseParams[CurrentEvent]);
#else
	  (*pEventHandler->EventHandler)(pEventHandler->UserContext);
#endif
	}
      }

      START_TIMER(&pThisInterface->CalTimer, SOFT_TIMER_PERIOD);
    }
  }
}
#endif

/******************************************************************************
FUNCTION NAME:
	chipALReadMemory

ABSTRACT:
	Perform Memory Mapped I/O Register Read

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
******************************************************************************/
MODULAR INLINE CHIPAL_FAILCODE_T chipALReadMemory(
    IN void				* RegIoContext,
    IN PVOID				BaseAddress,
    IN DWORD				OffSetAddress,
    IN DWORD				RegisterLength,
    OUT void 				* pReturnValue)
{
  if (!BaseAddress || !RegisterLength)
    return(CHIPAL_PARAM_OUT_OF_RANGE);
  else {
    void 	* pSourceRegister = (BYTE *)BaseAddress + OffSetAddress;

    // * Read IO Register
    switch(RegisterLength)
    {
    case sizeof(BYTE):
      *((BYTE *) pReturnValue) = *((BYTE *) pSourceRegister);
      break;
    case sizeof(WORD):
      *((WORD *) pReturnValue) = *((WORD *) pSourceRegister);
      break;
    case sizeof(DWORD):
      *((DWORD *) pReturnValue) = *((DWORD *) pSourceRegister);
      break;
    }
    return(CHIPAL_SUCCESS);
  }
}

/******************************************************************************
FUNCTION NAME:
	chipALWriteMemory

ABSTRACT:
	Perform Memory Mapped I/O Register Write

RETURN:
	CHIPAL_FAILCODE_T

DETAILS:
******************************************************************************/
MODULAR INLINE CHIPAL_FAILCODE_T chipALWriteMemory(
    IN void				* RegIoContext,
    IN PVOID				BaseAddress,
    IN DWORD				OffSetAddress,
    IN DWORD				RegisterLength,
    IN DWORD				MaskValue,
    IN DWORD				Value)
{
  void	* pRegister = (BYTE *)BaseAddress + OffSetAddress;

  // If Critical Section Applies, add Disable Interrupt code.

  // * Determine if Write is Bitmapped or if complete value should be written
  if (MaskValue)
  {
    DWORD	RegisterValue;

    // * Read IO Register
    switch(RegisterLength) {
    case sizeof(BYTE):
      RegisterValue = *((BYTE *) pRegister);
      break;
    case sizeof(WORD):
      RegisterValue = *((WORD *) pRegister);
      break;
    case sizeof(DWORD):
      RegisterValue = *((DWORD *) pRegister);
      break;
    default:
      RegisterValue = 0;
      break;
    }
    RegisterValue &= ~MaskValue;
    RegisterValue |= MaskValue & Value;
    Value = RegisterValue;
  }

  // * Write IO Register
  switch(RegisterLength) {
  case sizeof(BYTE):
    *((BYTE *) pRegister) = Value;
    break;
  case sizeof(WORD):
    *((WORD *) pRegister) = Value;
    break;
  case sizeof(DWORD):
    *((DWORD *) pRegister) = Value;
    break;
  }

  // End Critical Section

  return(CHIPAL_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	chipALGetChanConfig

ABSTRACT:
	Supply Default Dma Channel Configuration

RETURN:
	void

DETAILS:
******************************************************************************/
INLINE void chipALGetChanConfig(
    HW_CH_ENTRY_T * pChanConfig,
    CHIPAL_T      * pThisInterface)
{
#if PROJECTS_NAME == PROJECT_NAME_CHEETAH_1
  // 0 Host->Device
  pChanConfig->Description.Element.Direction = HOST_TO_DEVICE;
  // 0 Do Not Reverse order of Bits.  1 Reverse Bit Order of Byte
  pChanConfig->Description.Element.ReverseBitOrder = NO_REVERSE_BIT_ORDER;

  // 0 ChipAl Buffers,
  pChanConfig->Description.Element.LocalBuffers = LOCAL_BUFFERS;

  // 0 Non-Framed Mode
  pChanConfig->Description.Element.FrameMode = NON_FRAMED_MODE;

  // DMA Port
  pChanConfig->Description.Element.Port = CIRCULAR_DMA;

  // Bus Resources (Register Map)
  pChanConfig->BusResource = pThisInterface->BusResources;

  // Default Tuning Parameters
  pChanConfig->RegisteryDefault = pThisInterface->Tune;
#endif		// PROJECT_NAME_CHEETAH_1
}

/******************************************************************************
FUNCTION NAME:
	chipALFindLink

ABSTRACT:
	Locate a ChipAL Link based on a Cdsl Link

RETURN:
	Pointer to the ChipAL Link

DETAILS:
	A CDSL_LINK_T has a one to many relationship to the ChipAL Link
	structure.
	First, all possible ChipAL Links must be located.  Next, the 
	CHAN_DESCRIPT_T union is used to determine which link is required. The
	Mask value is used to determine which elements of the description
	should be examined.

	A static LastIndex value is kept to attempt to avoid a linear search.
	This should improve performance if the same link is being sought.

******************************************************************************/
LOCAL HW_CH_ENTRY_T * chipALFindLink(
    CHIPAL_T 			* pThisInterface,
    CDSL_LINK_T			* pCdslLink,
    CHAN_DESCRIPT_T		Description,
    DWORD			Mask)
{
  if (pThisInterface)
  {
    DWORD		LinkIndex;
    static DWORD	LastIndex = 0;/*CLN*/
    BOOL		FoundEntry = FALSE;

    // * Checked cached Index entry in an attempt to avoid a linear search
    if (pThisInterface->DmaChannel[LastIndex])
      if (((pThisInterface->DmaChannel[LastIndex]->Description.ElUnion.Value & Mask)
	   == Description.ElUnion.Value) &&
	  pThisInterface->DmaChannel[LastIndex]->UserLink == pCdslLink)
	FoundEntry = TRUE;

    // * Check All Possible ChipAL Link structures

    if (!FoundEntry)
      for (LinkIndex = 0; LinkIndex < MAX_DATA_CHANNEL; LinkIndex++)
      {
	if (pThisInterface->DmaChannel[LinkIndex]) {
	  // * If the specified description elements and the requested link match,
	  // * return the context for this link.
	  if (((pThisInterface->DmaChannel[LinkIndex]->Description.ElUnion.Value & Mask)
	       == Description.ElUnion.Value) &&
	      pThisInterface->DmaChannel[LinkIndex]->UserLink == pCdslLink)
	  {
	    LastIndex = LinkIndex;
	    FoundEntry = TRUE;
	    break;
	  }
	}
      }

    if (FoundEntry)
      return (pThisInterface->DmaChannel[LastIndex]);
  }
  return(NULL);
}

/******************************************************************************
FUNCTION NAME:
	chipALClearInterrupt

ABSTRACT:
	Clear Interrupt and any device specific causes of the interrupt. Any
	bit that is set will be saved in a static copy of the Interrupt Status
	array (pThisInterface->InterruptStatus[]). This is a cumulative value
	that is updated during each ISR call.  The value will be processed and
	cleared in the DPC routine.

	NOTE:  Isr routines called from this procedure are processed in order. 
	if ChipALSetEvent() is called

RETURN:
	void

DETAILS:
******************************************************************************/
LOCAL INLINE BOOLEAN chipALClearInterrupt(
    CHIPAL_T	* pThisInterface,
    enum WhichFunction	FunctSel,
    DWORD	HardWareInterrupt)
{
  //DWORD 		ProcessedInterrupts[HW_MAX_EVENT_REGISTERS];
  DWORD 		RegisterIndex;
  DWORD 		EventIndex;
  CAL_EVENT_LIST_T	* pEventHandler;
  LIST_ENTRY		* pHeadList;
  BOOLEAN               DPCRequired = FALSE;

  // * Update static copy of Interrupt Status register. Any bit set in
  // * InterruptStatus will result in a call to a channel specific
  // * handler in the DPC procedure.  Any special case mapping of the
  // * channel handler to interrupt status is handled in this routine.
  // *
  // * Note: InterruptStatus is cleared in the DPC
  (pThisInterface->BusControl->Disable)(pThisInterface->BusResources, FunctSel, DISABLE_INTERRUPT);

  // Copy the hardware interrupt register to the Interrupt Event structure (array).
  pThisInterface->IsrEvent[FunctSel][HARDWARE_INTERRUPT_REGISTER] = HardWareInterrupt;

  // * Call Bus Controller to clear Interrupt Status bit(s).
  (*pThisInterface->BusControl->ClearStatus)(pThisInterface->BusResources,FunctSel,HardWareInterrupt);

  // * Process all event array elements
  for (RegisterIndex = 0; RegisterIndex < HW_MAX_EVENT_REGISTERS; RegisterIndex++)
  {
    if (pThisInterface->IsrEvent[FunctSel][RegisterIndex]) {
      // * Process All Bits in an event array element
      for (EventIndex = 0;
	   (EventIndex < BITS_PER_EVENT_REG) &&
	     pThisInterface->IsrEvent[FunctSel][RegisterIndex];
	   EventIndex++)
      {
	DWORD Mask = 1 << EventIndex;

	// * If Bit is Set, Check for ISR Event Handlers
	if (Mask & pThisInterface->IsrEvent[FunctSel][RegisterIndex])
        {
	  DWORD CurrentEvent;

	  // Encode Register Indexes into an Event
	  CurrentEvent = BITS_PER_EVENT_REG * RegisterIndex + EventIndex;

	  // * Check for any attached ISR routines responsible for clearing
	  // * an interrupt.  If one or more are found, they must decode the
	  // * event and decide which DPC level or secondary ISR event(s)
	  // * should be posted.
	  pHeadList = &pThisInterface->EventHandler[FunctSel][HW_ISR_EVENT][CurrentEvent];

	  for (pEventHandler = (CAL_EVENT_LIST_T *) pHeadList->pNext;
	       (LIST_ENTRY *) pEventHandler != pHeadList;
	       pEventHandler = (CAL_EVENT_LIST_T *) pEventHandler->HandlerList.pNext)
	  {
	    if (pEventHandler->EventHandler)
	    {
	      // Call Event Handler
#if MAX_RESPONSE_SIZE   // ChipAl supplies Response Param(s) to ISR / DPC Events

	      // * ISR routine Decodes Event and Writes Response Params to the appropriate
	      // * ResponseRegister.  DPC Reads Params as Constant Values
	      (*pEventHandler->EventHandler)(pEventHandler->UserContext, &pThisInterface->ResponseParams[0]);
#else
	      (*pEventHandler->EventHandler)(pEventHandler->UserContext);
#endif
	    }
	  }

	  // * Save Bit to allow DPC processing
	  pHeadList = &pThisInterface->EventHandler[FunctSel][HW_DPC_EVENT][CurrentEvent];
	  if (pHeadList->pNext != pHeadList)
	  {
	    ChipALSetEvent(pThisInterface->pNdisAdapter, FunctSel, CurrentEvent);
	    DPCRequired = TRUE;
	  }

	  // * Clear Processed bit to allow early exit of loop
	  pThisInterface->IsrEvent[FunctSel][RegisterIndex] &= ~Mask;
	  if (!pThisInterface->IsrEvent[FunctSel][RegisterIndex])
	  {
	    // All Interrupts processed in current register. Go to next register.
	    break;
	  }
	} // If Mask
      }	// For EventIndex
    }
  } // For RegisterIndex

  (pThisInterface->BusControl->Enable)(pThisInterface->BusResources, FunctSel, 0);

  return DPCRequired;
}

/******************************************************************************
Name:
	chipALAllocateInterface

Function:
	Add New Interface to Linked List

Description:
******************************************************************************/
INLINE CHIPAL_T * chipALAllocateInterface(
    void)
{
  NDIS_STATUS Status;
  CHIPAL_T * pThisInterface;

  CHIPAL_ASSERT_TRACE("chipALAllocateInterface() ", TRUE);

  // * First Time Setup
  if (!pChipALHead)
    pChipALHead = &ChipALHead;

  // * Locate New Hardware InterfaceContext
  pThisInterface = pChipALHead;
  while (pThisInterface->Next)
    pThisInterface = pThisInterface->Next;

  Status = ALLOCATE_MEMORY(pThisInterface->Next, sizeof(*(pThisInterface->Next)));
  if (Status != STATUS_SUCCESS)
    return(NULL);

  CLEAR_MEMORY(pThisInterface->Next, sizeof(*(pThisInterface->Next)));

  pThisInterface = pThisInterface->Next;
  pThisInterface->Next = NULL;

  return(pThisInterface);
}

/******************************************************************************
Name:
	chipALReleaseInterface

Function:
	Release Memory for Adapter Instance and relink list if necessary

Description:
******************************************************************************/
INLINE void chipALReleaseInterface(
    CHIPAL_T * AdapterToRelease)
{
  CHIPAL_T * pThisInterface;

  CHIPAL_ASSERT_TRACE("chipALReleaseInterface() ", TRUE);

  pThisInterface = pChipALHead;

  while (pThisInterface->Next && pThisInterface->Next != AdapterToRelease)
    pThisInterface = pThisInterface->Next;

  if (pThisInterface->Next == AdapterToRelease)
  {
    CHIPAL_T * pTempInterface = pThisInterface->Next->Next;

    FREE_MEMORY(pThisInterface->Next, sizeof(*(pThisInterface->Next)),
		CHIPAL_CONTEXT_MEM_FLAGS);

    pThisInterface->Next = pTempInterface;
  }
}

/******************************************************************************
FUNCTION NAME:
	ChipALDeviceSpecific

ABSTRACT:
	Process Device Specific OIDs

RETURN:

DETAILS:
******************************************************************************/
NDIS_STATUS ChipALDeviceSpecific(
    CDSL_ADAPTER_T	* pThisAdapter,
    BACK_DOOR_T		* pBackDoorBuf,
    PULONG		pNumBytesNeeded,
    PULONG		pNumBytesWritten)
{
  NDIS_STATUS     	status = STATUS_SUCCESS;
  CHIPAL_T 		* pThisInterface;
  BUS_RESOURCE_T	* pBusResource;
  SK_BUFF_T		ReturnPacket;

  if (pBackDoorBuf->TotalSize < SIZEOF_DWORD_ALIGN(BACK_DOOR_T) &&
      pBackDoorBuf->TotalSize != SIZEOF_DWORD_ALIGN(MINI_BACK_DOOR_T))
  {
    pBackDoorBuf->ResultCode = ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
    pBackDoorBuf->NeededSize = SIZEOF_DWORD_ALIGN(BACK_DOOR_T);

    *pNumBytesNeeded = SIZEOF_DWORD_ALIGN(BACK_DOOR_T);

    return ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
  }

  pThisInterface = (CHIPAL_T *) pThisAdapter->ChipMod;
  if (!pThisInterface)
  {
    pBackDoorBuf->ResultCode = ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);
    pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;

    return ERROR_FATAL(RESULT_ADAPTER_NOT_FOUND);
  }

  pBusResource = pThisInterface->BusResources;

  switch(pBackDoorBuf->ReqCode)
  {
  case BD_CHIPAL_GET_STATS:
    chipALGetStats(pThisAdapter, pBackDoorBuf);

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten =	(sizeof(pBackDoorBuf->Params.BdChipALStats) +
			 sizeof(pBackDoorBuf->ResultCode) +
			 sizeof(pBackDoorBuf->NeededSize));
    break;

  case BD_CHIPAL_CLEAR_STATS:
    chipALClearStats(pThisAdapter, pBackDoorBuf);

    pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten =	(sizeof(pBackDoorBuf->Params.BdChipALStats) +
			 sizeof(pBackDoorBuf->ResultCode) +
			 sizeof(pBackDoorBuf->NeededSize));
    break;

  case BD_CHIPAL_ATM_GET_TX_CELL:
    CmgmtIndicateLinkUp(pThisAdapter);
    if (!pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX])
    {
      // The Receive path is not configured.
      pBackDoorBuf->Params.BdChipALAtmCell.Length = 0;

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    else if (!pThisInterface->DmaControl->GetLastBuffer)
    {
      // The Receive path is not configured.
      pBackDoorBuf->Params.BdChipALAtmCell.Length = 0;

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    else
    {
      (* pThisInterface->DmaControl->GetLastBuffer)
	(pThisInterface->DmaChannel[ADSL_DMA_CHAN_TX], &ReturnPacket);

      pBackDoorBuf->Params.BdChipALAtmCell.Length = MIN_VALUE(
			      ReturnPacket.len, RX_BLOCK_SIZE);

      COPY_MEMORY(pBackDoorBuf->Params.BdChipALAtmCell.DataBytes,
		  ReturnPacket.data,
		  pBackDoorBuf->Params.BdChipALAtmCell.Length);

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    break;

  case BD_CHIPAL_ATM_GET_RX_CELL:
    if (!pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX])
    {
      // The Receive path is not configured.
      pBackDoorBuf->Params.BdChipALAtmCell.Length = 0;

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    else if (!pThisInterface->DmaControl->GetLastBuffer)
    {
      // The Receive path is not configured.
      pBackDoorBuf->Params.BdChipALAtmCell.Length = 0;

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    else
    {
      (* pThisInterface->DmaControl->GetLastBuffer)
	(pThisInterface->DmaChannel[ADSL_DMA_CHAN_RX], &ReturnPacket);

      pBackDoorBuf->Params.BdChipALAtmCell.Length = MIN_VALUE(
				ReturnPacket.len, RX_BLOCK_SIZE);

      COPY_MEMORY(pBackDoorBuf->Params.BdChipALAtmCell.DataBytes,
		  ReturnPacket.data,
		  pBackDoorBuf->Params.BdChipALAtmCell.Length);

      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      pBackDoorBuf->NeededSize = pBackDoorBuf->TotalSize;

      *pNumBytesNeeded = pBackDoorBuf->TotalSize;
      *pNumBytesWritten = (sizeof(pBackDoorBuf->Params.BdChipALAtmCell) +
			   sizeof(pBackDoorBuf->ResultCode) +
			   sizeof(pBackDoorBuf->NeededSize));
    }
    break;

#ifdef DBG
  case BD_CARDMGMT_SET_DEBUG_FLAG:
    ChipALDebugFlag &= DEBUG_MODULE_ID_MASK;
    ChipALDebugFlag |= pBackDoorBuf->Params.BdCardMgmtDebugFlag;
    break;
#endif //DBG

  default:
    //	These BackDoor ReqCodes have not been implemented yet.
    //		case BD_CHIPAL_START_LOOPBACK:
    //		case BD_CHIPAL_START_1_LOOPBACK:
    //		case BD_CHIPAL_STOP_LOOPBACK:
    //		case BD_CHIPAL_GET_TEST_RESULTS:
    //		case BD_CHIPAL_ENQ_TEST_COMPLETE:
    //		case BD_CHIPAL_GET_TUNING_PARAMS:
    //		case BD_CHIPAL_SET_TUNING_PARAMS:
    status = STATUS_UNSUPPORTED;
    break;
  }

  return status;
}

/******************************************************************************
FUNCTION NAME:
	chipALGetStats

ABSTRACT:
	Get All Hal Statistics

RETURN:
	void

DETAILS:
******************************************************************************/
LOCAL INLINE void chipALGetStats(
    CDSL_ADAPTER_T * pThisAdapter,
    BACK_DOOR_T	* pBackDoorBuf)
{
  CHIPAL_T		* pThisInterface;

  CLEAR_MEMORY(&pBackDoorBuf->Params, sizeof(BD_CHIPAL_STATS_T));

  pThisInterface = (CHIPAL_T *) pThisAdapter->ChipMod;
  if (pThisInterface) {
#if DMA_STATS
    {
      BD_CHIPAL_STATS_T	TempStats;
      DWORD	Index;

      // * Sum DMA Transmit and Receive Numbers
      for (Index = 0; Index < MAX_DATA_CHANNEL; Index++)
	if (pThisInterface->DmaChannel[Index])
	{
	  (* pThisInterface->DmaControl->ChanGetStats)(
			pThisInterface->DmaChannel[Index],
			&TempStats);
	  pBackDoorBuf->Params.BdChipALStats.BytesTransmitted += TempStats.BytesTransmitted;
	  pBackDoorBuf->Params.BdChipALStats.BytesReceived += TempStats.BytesReceived;
	}
    }

    // * Data was summed in the standard BD_CHIPAL_STATS_T structure.  It must
    // * now be copied to the BD_CHIPAL_ATM_STATS_T struct. AS xxx.Params is a
    // * union, the TempStats location is used to avoid overwriting data.  Both
    // * structs report data in Kbytes.
    TempStats = pBackDoorBuf->Params.BdChipALStats;
    pBackDoorBuf->Params.BdChipALAtmStats.NumTxCells = TempStats.BytesTransmitted;
    pBackDoorBuf->Params.BdChipALAtmStats.NumRxCells = TempStats.BytesReceived;
#endif	// DMA_STATS
  }
}

/******************************************************************************
FUNCTION NAME:
	chipALClearStats

ABSTRACT:
	Get All Hal Statistics

RETURN:
	void

DETAILS:
******************************************************************************/
LOCAL INLINE void chipALClearStats(
    CDSL_ADAPTER_T * pThisAdapter,
    BACK_DOOR_T	  * pBackDoorBuf)
{
  CHIPAL_T	* pThisInterface = (CHIPAL_T *) pThisAdapter->ChipMod;

  if (pThisInterface)
  {
#if DMA_STATS
    {
      DWORD Index;

      // * Clear DMA Stats
      for (Index = 0; Index < MAX_DATA_CHANNEL; Index++)
	if (pThisInterface->DmaChannel[Index])
	  (* pThisInterface->DmaControl->ChanClearStats)(pThisInterface->DmaChannel[Index]);
    }
#endif	// DMA_STATS
  }
}

#ifdef DBG	// Development Test Routines
/******************************************************************************
Name:
	ChipALTestReadBuff

Function:
	Read Buffer from ChipAL

Description:
	ChipALTestxxx functions are used for development testing with the 
	ControlPanel Applet. They are not intended for the final 
	product.

******************************************************************************/
DWORD chipALTestReadBuff(
    char * Destination,
    DWORD BuffSize,
    DWORD * SizeWritten)
{
  DWORD WriteLen;
#define TEST_STR "The Quick Brown Ugly Fox"

  CHIPAL_ASSERT_TRACE("ChipALTestReadBuff ",
		      (Destination &&
		       BuffSize &&
		       SizeWritten));

  WriteLen = sizeof(TEST_STR);

  if (BuffSize < WriteLen)
    return(STATUS_FAILURE);

  COPY_MEMORY(Destination, TEST_STR, WriteLen);
  *SizeWritten = WriteLen;

  return(STATUS_SUCCESS);
}

/******************************************************************************
Name:
	ChipALTestWriteBuff

Function:
	Write Buffer to ChipAl

Description:
	ChipALTestxxx functions are used for development testing with the 
	ControlPanel Applet. They are not intended for the final product.

******************************************************************************/
DWORD chipALTestWriteBuff(
    CDSL_ADAPTER_T * ThisAdapter,
    char * Destination,
    DWORD SizeInBuff)
{
#define TEST_BUFF_SIZE 80

  CHIPAL_ASSERT_TRACE("ChipALTestWriteBuff() ",
		      (Destination &&
		       SizeInBuff &&
		       ThisAdapter));

  if (!ThisAdapter ||
      !ThisAdapter->ChipMod)
    return(STATUS_FAILURE);
  else
    return(STATUS_SUCCESS);
}

/******************************************************************************
FUNCTION NAME:
	ChipALTestReadConfig

ABSTRACT:
	Get Adapter Status and Configuration

RETURN:
	ChipAL Fail Code

DETAILS:
******************************************************************************/
DWORD chipALTestReadConfig(
    CDSL_ADAPTER_T	* ThisAdapter,
    BD_CHIPAL_CONFIG_T	* pConfigStatus
)
{
  CHIPAL_T * pThisInterface;

  CHIPAL_ASSERT_TRACE("ChipALTestReadConfig() ", ThisAdapter);

  if (!ThisAdapter)
    return(STATUS_FAILURE);

  if (!ThisAdapter->ChipMod)
  {
    pConfigStatus->Status	 = CHIPAL_STATUS_DOWN;
    pConfigStatus->Irq		 = 0;
    pConfigStatus->BaseAddress	 = 0;
    pConfigStatus->SharedAddress = 0;

    return(STATUS_SUCCESS);
  }

  pThisInterface = (CHIPAL_T *) ThisAdapter->ChipMod;

  // * Check Bus Resources
  if (pThisInterface->BusResources)
  {
    pConfigStatus->Status	= CHIPAL_STATUS_CONFIGURED;
    pConfigStatus->Irq		= pThisInterface->BusResources->PriIntVector;
    pConfigStatus->BaseAddress	= pThisInterface->BusResources->VirtualAddressBAR0;
    pConfigStatus->SharedAddress= (DWORD) pThisInterface->BusResources->VirtualAddressSharedMem;
  }
  else
  {
    pConfigStatus->Status	 = CHIPAL_STATUS_INITIALIZED;
    pConfigStatus->Irq		 = 0;
    pConfigStatus->BaseAddress	 = 0;
    pConfigStatus->SharedAddress = 0;
  }

  return(STATUS_SUCCESS);
}

#ifdef TRACING_ENABLED
/******************************************************************************
FUNCTION NAME:
	chipALDisplayTraceQueue

ABSTRACT:
	Display Memory Trace.  DEBUG ONLY

RETURN:
	void

DETAILS:
******************************************************************************/
LOCAL void chipALDisplayTraceQueue(
    void)
{
  static const trace_name[MAX_TRACE] = {
    "UDEFINED",
    "ISR_ENTRY",
    "ISR_EXIT",
    "SYNC_ENTRY",
    "SYNC_EXIT",
    "TX_ENTRY",
    "TX_EXIT",
    "RX_ENTRY",
    "RX_EXIT",
    "DPC_ENTRY",
    "DPC_EXIT"
  };
  WORD	TempIdx;

  for (TempIdx = 0; TempIdx < TRACE_QUEUE_LEN; TempIdx++)
    DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
		     ChipALDebugFlag,
		     ("Trace: %4d %s", TraceQueue[TempIdx],
		      (TraceQueue[TempIdx] >= MAX_TRACE) ? MAX_TRACE : TraceQueue[TempIdx]));
}
#endif // TRACING_ENABLED
#endif // DBG	 Development Test Routines
