/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	????
**
**FILE NAME:
**	CnxTTY.c
**
**ABSTRACT:
**	This file implements tty interface
**
**DETAILS:
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Tigris/Linux_2.4.3_ATM_Driver/KernelModule/CnxTTY.c $
**	$Revision: 5 $
**	$Date: 8/09/01 1:47p $
*******************************************************************************
******************************************************************************/
#ifndef MODULE
#error "MODULE should be defined"
#endif

#include "linuxmodver.h"

#include <linux/tty.h>
#include <linux/mm.h>
#include "LnxTools.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	#define  __NO_VERSION__
#endif
#include <linux/module.h>

#include <asm/semaphore.h>

#include "CardMgmt.h"
#include "CardALTigrisDp.h"
#include "datapump.h"
#include "sysinfo.h"
#include "serial.h"
#include "terminal.h"

#if HAS_TTY
/*  #define CNXTTY_MAJOR 47 */
#define CNXTTY_MAJOR 0
#define CNXTTY_NORMAL_TYPE 1
#define CNXTTY_PORTS 1 /* if this is not 1, a lot of code is broken */

struct tty_struct *cnxtty_table[CNXTTY_PORTS];
static struct termios *cnxtty_termios[CNXTTY_PORTS];
static struct termios *cnxtty_termios_locked[CNXTTY_PORTS];
static int cnxtty_refcount;
static unsigned char *temp_buffer;
static struct semaphore temp_buffer_sem;

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#ifdef DEBUG_CNXTTY2
#define DBP(a,b...)	printk("\n\r %s: " a, __FUNCTION__ , ## b) 
#else
#define DBP(a,b...)
#endif 

#if 0
static void DBP_TTY_DRIVER(const struct tty_driver *d) {
  DBP("TTY DRIVER");
  DBP("magic            %d", (d)->magic);
  DBP("driver_name      %s", (d)->driver_name);
  DBP("name             %s", (d)->name);
  DBP("name_base        %d", (d)->name_base);
  DBP("major            %d", (d)->major);
  DBP("minor_start      %d", (d)->minor_start);
  DBP("num              %d", (d)->num);
  DBP("type             %d", (d)->type);
  DBP("subtype          %d", (d)->subtype);
  DBP("init_termios --not done--");
  DBP("flags            %d\n", (d)->flags);
  DBP("*refcount        %x", (unsigned int) (d)->refcount);
  DBP("*proc_entry      %x", (unsigned int) (d)->proc_entry);
  DBP("*other           %x", (unsigned int) (d)->other);
  DBP("**table          %x", (unsigned int) (d)->table);
  DBP("**termios        %x", (unsigned int) (d)->termios);
  DBP("**termios_locked %x", (unsigned int) (d)->termios_locked);
  DBP("*driver_state    %x", (unsigned int) (d)->driver_state);
  DBP("*open            %x", (unsigned int) (d)->open);
  DBP("*close           %x", (unsigned int) (d)->close);
  DBP("*write           %x", (unsigned int) (d)->write);
  DBP("*put_char        %x", (unsigned int) (d)->put_char);
  DBP("*flush_chars     %x", (unsigned int) (d)->flush_chars);
  DBP("*write_room      %x", (unsigned int) (d)->write_room);
  DBP("*chars_in_buffer %x", (unsigned int) (d)->chars_in_buffer);
  DBP("*ioctl           %x", (unsigned int) (d)->ioctl);
  DBP("*set_termios     %x", (unsigned int) (d)->set_termios);
  DBP("*throttle        %x", (unsigned int) (d)->throttle);
  DBP("*unthrottle      %x", (unsigned int) (d)->unthrottle);
  DBP("*stop            %x", (unsigned int) (d)->unthrottle);
  DBP("*start           %x", (unsigned int) (d)->start);
  DBP("*hangup          %x", (unsigned int) (d)->hangup);
  DBP("*break_ctl       %x", (unsigned int) (d)->break_ctl);
  DBP("*flush_buffer    %x", (unsigned int) (d)->flush_buffer);
  DBP("*set_ldisc       %x", (unsigned int) (d)->set_ldisc);
  DBP("*wait_until_sent %x", (unsigned int) (d)->wait_until_sent);
  DBP("*send_xchar      %x", (unsigned int) (d)->send_xchar);
  DBP("*read_proc       %x", (unsigned int) (d)->read_proc);
  DBP("*write_proc      %x", (unsigned int) (d)->write_proc);
  DBP("*next            %x", (unsigned int) (d)->next);
  DBP("*prev            %x", (unsigned int) (d)->prev);
}

#define DBP_TTY(t)                       \
DBP("\n\rTTY");                          \
DBP("magic            %d", (t)->magic);  \
DBP_TTY_DRIVER((t)->driver);             \
DBP("pgrp             %d", (t)->pgrp);   \
DBP("session          %d", (t)->session);\
DBP("flags            %d", (t)->flags);  \
DBP("count            %d", (t)->count);
#endif

#if 0
static CDSL_ADAPTER_T *pThisAdapter;/*CLN*/
#endif

static int _alloc_temp_buffer(
    void)
{
  if (!temp_buffer) {
    temp_buffer = (unsigned char *) __get_free_page(GFP_KERNEL);
    if (!temp_buffer)
      return -1;
  }

  return 0;
}

static int _free_temp_buffer(
    void)
{
  if (temp_buffer) {
    free_page((unsigned long)temp_buffer);
    temp_buffer = 0;
  }

  return 0;
}

static int cnxtty_open(
    struct tty_struct *tty,
    struct file * filp)
{
  if(_alloc_temp_buffer()) {
    DBP("ERROR allocating memory");
    return -ENOMEM;
  }

  //	printk ("cnxtty_open driver_data=%p\n", tty->driver_data);
  ChipAlTigrisDpSER_DetectTerminalConnected(tty->driver_data);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
  MOD_INC_USE_COUNT;
#endif

  return 0;
}

static void cnxtty_close(
    struct tty_struct * tty,
    struct file * filp)
{
  //	printk ("cnxtty_close driver_data=%p\n", tty->driver_data);
  ChipAlTigrisDpTRM_AutoUpdateDisable (tty->driver_data, LOCAL_LOGON, TRUE);
  ChipAlTigrisDpSER_ClearRemoteLogon (tty->driver_data);

  _free_temp_buffer();

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
  MOD_DEC_USE_COUNT;
#endif
}

#if 0
static void _cnxtty_write_wait(
    struct tty_struct * tty)
{
  long timeout = MAX_SCHEDULE_TIMEOUT;

  DECLARE_WAITQUEUE(wait, current);
  add_wait_queue(&tty->write_wait, &wait);
  do {
    set_current_state(TASK_INTERRUPTIBLE);
    if (signal_pending(current))
      goto stop_waiting;
    timeout = schedule_timeout(timeout);
  } while (timeout);

stop_waiting:
  current->state = TASK_RUNNING;
  remove_wait_queue(&tty->write_wait, &wait);
}

static int cnxtty_putchar(
    tty_struct * gTTY,
    int c)
{
  if (gTTY) {
    if (0 < gTTY->ldisc.receive_room(gTTY)) {
      char ch[1];

      ch[0] = (char)c;
      gTTY->ldisc.receive_buf(gTTY, &ch[0], 0, 1);
    }
  }
	
  return 1;
}
#endif

static int cnxtty_write(
    struct tty_struct *tty,
    int from_user,
    const unsigned char *buf,
    int count)
{
  int returnValue = 0;
  int numberToWrite;
  int charLoop;
  unsigned char *pBuffer;

  /* check space so we don't copy needlessly */
  numberToWrite = MIN(count, tty->ldisc.receive_room(tty));
  if (!numberToWrite) {
    DBP("ERROR no room in the receive buffer");
    goto cnxtty_write_end;
  }

  if (from_user) {
    down(&temp_buffer_sem);

    numberToWrite = MIN(numberToWrite, PAGE_SIZE);

    /* verify the access */
    if (! access_ok(VERIFY_READ, buf, numberToWrite)) {
      DBP("ERROR user does not have access to this buffer");
      returnValue = -EFAULT;
      goto cnxtty_write_end;
    }

    if (copy_from_user(temp_buffer, buf, numberToWrite)) {
      DBP("ERROR copying from user ");
      returnValue = -EFAULT;
      goto cnxtty_write_end;
    }

    /*  		DBP("Sending %d bytes to line", numberToWrite); */
    pBuffer = temp_buffer;

    up(&temp_buffer_sem);
  } else
    pBuffer = (unsigned char *)buf; /* cast for cc */

  //	printk ("cnxtty_write driver_data=%p\n", tty->driver_data);
  for (charLoop = 0; charLoop < numberToWrite; charLoop++) {
    /*  	  tty->ldisc.receive_buf(tty, &pBuffer[charLoop], 0, 1); */
    /*  	  gTTY = tty; */
    ChipAlTigrisDpSER_RcvIntrHandler (tty->driver_data, pBuffer[charLoop]);
  }

  returnValue = numberToWrite;

 cnxtty_write_end:

  /*    	DBP("returns %d\n", returnValue);  */

  /*  	_cnxtty_write_wait(tty); */

  return returnValue;
}

static int cnxtty_write_room(
    struct tty_struct *tty
)
{
  return 0;
}

static void cnxtty_flush_buffer(
    struct tty_struct *tty
)
{
}

static int cnxtty_chars_in_buffer(
    struct tty_struct *tty
)
{
  int returnValue = -1;

  returnValue = 1; /* just a lie to get minicom to go a read */

  DBP("returns %d\n", returnValue);

  return returnValue;
}

static void cnxtty_unthrottle(
    struct tty_struct * tty
)
{
}

static void cnxtty_set_termios(
    struct tty_struct *tty,
    struct termios *old_termios
)
{
  tty->termios->c_cflag &= ~(CSIZE | PARENB);
  tty->termios->c_cflag |= (CS8 | CREAD);
}

static void cnxtty_hangup(
    struct tty_struct *tty
)
{
}

static struct tty_driver cnxtty_driver = {
  magic:		TTY_DRIVER_MAGIC,
  driver_name:		"test_tty",
  name:			"ttyCX",
  major:		CNXTTY_MAJOR,
  minor_start:		0,
  num:			CNXTTY_PORTS,
  type:			TTY_DRIVER_TYPE_SERIAL,
  subtype:		CNXTTY_NORMAL_TYPE,
  /*
  init_termios:		tty_std_termios,
  init_termios.c_cflag: B38400 | CS8 | CREAD,
  */
  flags:		TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW,
  refcount:		&cnxtty_refcount,
  table:		cnxtty_table,
  termios:		cnxtty_termios,
  termios_locked:	cnxtty_termios_locked,

  open:			cnxtty_open,
  close:		cnxtty_close,
  write:		cnxtty_write,
  write_room:		cnxtty_write_room,
  flush_buffer:		cnxtty_flush_buffer,
  chars_in_buffer:	cnxtty_chars_in_buffer,
  unthrottle:		cnxtty_unthrottle,
  set_termios:		cnxtty_set_termios,
  hangup:		cnxtty_hangup,
};

/*
 *  Setup
 */
int cnxtty_init (
    CDSL_ADAPTER_T *pAdapter
)
{
  int status;
  int returnValue = 1;

  /*   	DBP("Hello"); */

  /* init the globals */
  cnxtty_refcount = 0;
  temp_buffer = 0;
  sema_init(&temp_buffer_sem, 1);

  cnxtty_driver.init_termios = tty_std_termios;
  cnxtty_driver.init_termios.c_cflag = B38400 | CS8 | CREAD;

  /*  	DBP_TTY_DRIVER(&cnxtty_driver); */

  status = tty_register_driver(&cnxtty_driver);

  /*  	DBP_TTY_DRIVER(&cnxtty_driver); */

  /*  	DBP("tty_register_driver returns %d", status); */

  if (!status) {
    //	printk ("cnxtty_init pAdapter=%p\n", pAdapter);
    (*cnxtty_driver.table)->driver_data = pAdapter;

    returnValue = 0;
  }

  /*   	DBP("returns %d", returnValue); */

  //	printk ("cnxtty_init pAdapter=%p\n", pAdapter);
  (*cnxtty_driver.table)->driver_data = pAdapter;

  return returnValue;
}

void cnxtty_fini (
    CDSL_ADAPTER_T *pAdapter
) {
  tty_unregister_driver(&cnxtty_driver);
}

void cnxtty_SER_XmitMsg (
    PDATAPUMP_CONTEXT pDatapump,
    const char *buf,
    UINT16 count)
{
  struct tty_struct *tty = cnxtty_table[0];
  int numberToWrite;

  /* check space so we don't copy needlessly */
  numberToWrite = MIN(count, tty->ldisc.receive_room(tty));
  if (!numberToWrite) {
    DBP("ERROR no room in the receive buffer");
  } else
    tty->ldisc.receive_buf(tty, buf, 0, numberToWrite);
}
#else
void cnxtty_SER_XmitMsg (
    PDATAPUMP_CONTEXT pDatapump,
    const char *buf,
    UINT16 count)
{
  /* Called from SER_XmitMsg in dpcontroller.o */
}
#endif
