/******************************************************************************
*******************************************************************************
****	Copyright (c) 1997, 1998, 1999, 2000, 2001
****	Conexant Systems Inc. (formerly Rockwell Semiconductor Systems)
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	FrameAL
**
**FILE NAME:
**	FrameAL.c
**
**ABSTRACT:
**	Framer shim between NDIS miniport driver and HDLCOCT.
**
**DETAILS:
**
*******************************************************************************
** KEYWORDS:
**	$Archive: /Projects/Tigris and Derivatives Linux/Reference Designs/Tigris/Linux 2.4 ATM Driver/KernelModule/FrameAL.c $
**	$Revision: 2 $
**	$Date: 2/28/01 11:05a $
*******************************************************************************
******************************************************************************/

#include "linuxmodver.h"

#include "CardMgmt.h"
#include "FrameAL.h"
#include "CardMgmtV.h"

#define NO_MGMT	0

#if CAL_BASIC2_INSTALLED	
// this is a Cheetah project
// max line rate at which traffic shaping can occur. Above this line rate Shaping is disabled
	#define CLIENTLIMIT	832000
#else
	#define CLIENTLIMIT	928000
#endif

// check for line rate change at this interval (based on Hobbs general timer)
#define LINE_CHECK_INTERVAL	0x1000

// Timer interval of 10ms
#define TIMER_INTERVAL_UPPER 	1

#define TIMER_INTERVAL_LOWER	100

// default bandwidth limit, used when Registy value is out of limits
#define DEFAULT_BURST_SIZE	500

#define ERROR_FATAL(error)	(LEVEL_FATAL   | MODULE_FRAMEAL | INSTANCE_ATM | error)
#define ERROR_WARNING(error)	(LEVEL_WARNING | MODULE_FRAMEAL | INSTANCE_ATM | error)
#define ERROR_NORMAL(error)	(LEVEL_NORMAL  | MODULE_FRAMEAL | INSTANCE_ATM | error)

#ifndef MEMORY_TAG
	#define	MEMORY_TAG	'LARF' 
#endif

// ***************************************
//		FILE STRUCTURES
// ***************************************

typedef struct {
  // ******************************************
  // Adapter FrameAL Module Context structure
  // ******************************************

  // All the following are filled with values
  // obtained from reading the registry

#if 0
  BD_FRAMEAL_ENCAPSULATION_T	RfcEncapsulationMode;
#endif

  DWORD				CellBurstSize;
} CDSL_ADAPTER_FRAMEAL_T;

/******************************************************************************
	Private Prototypes
******************************************************************************/
LOCAL DWORD FrameALUpdateATMStats(
    IN 	CDSL_VC_T * pVCHdl,
    IN 	BD_FRAMEAL_ATM_STATS_T * pATMStats
);

LOCAL DWORD FrameALUpdateAALStats(
    IN 	CDSL_VC_T * pVCHdl,
    IN 	BD_FRAMEAL_AAL_STATS_T * pAALStats
);

#define VC_STATS_OK 	0
#define VC_STATS_FAILED 1

LOCAL void FrameALTimerHandler(
    IN ULONG 	LinkHdl
);

/******************************************************************************
FUNCTION NAME:
	FrameALAdapterInit

ABSTRACT:
	Initializes a new Adapter instance. 
	
RETURN:
	NDIS Status
	 
DETAILS:
	FrameAL does not perform any initialization at adapter init time.
******************************************************************************/

GLOBAL NDIS_STATUS FrameALAdapterInit(
    IN	CDSL_ADAPTER_T	*AdprHdl) {
  // Nothing to do here
  // All initialization occurs at link init time.

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALAdapterShutdown

ABSTRACT:
	To provide a point from which FrameAL module can deinitialize
	any adapter context structure specific data it may have intialized.

RETURN:
	NDIS status

DETAILS:
	Check to see if we had allocated an adapter context structure and
	free up the memory if we had. 

******************************************************************************/

GLOBAL NDIS_STATUS FrameALAdapterShutdown(
    IN	CDSL_ADAPTER_T	*AdprHdl) {
  // Get the FrameAL adapter context
  CDSL_ADAPTER_FRAMEAL_T *FrameALAdprCtx = AdprHdl->FrameMod;

  // If we had previously allocated an adapter context structure free it now.
  if (FrameALAdprCtx) {
    FREE_MEMORY(FrameALAdprCtx, sizeof(*FrameALAdprCtx), 0);
    AdprHdl->FrameMod = NULL;
  }

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkStartTX

ABSTRACT:
	BM calls this function when new data arrives on a VC.  

RETURN:
	NDIS status

DETAILS:
	FrameAL passes this on to Mux to call ChipAl at the right time.

******************************************************************************/

GLOBAL NDIS_STATUS FrameALLinkStartTX(
    IN CDSL_LINK_T 	*LinkHdle) {
  if (LinkHdle)
    return FrameALATMStartTx(LinkHdle);
  else
    return STATUS_FAILURE;
}

/******************************************************************************
FUNCTION NAME:
	FrameALVCStartTX

ABSTRACT:
	BM calls this function to when new data arrives on a VC.  

RETURN:
	NDIS status

DETAILS:
	FrameAL passes this on to Mux to call ChipAl at the right time.

******************************************************************************/

GLOBAL NDIS_STATUS FrameALVCStartTX(
    IN CDSL_VC_T 	*VCHdl) {
  if (VCHdl)
    return FrameALLinkStartTX(VCHdl->LinkHdl);
  else
    return STATUS_FAILURE;
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkInit

ABSTRACT:
	Initializes a new link instance.  

RETURN:
	NDIS status

DETAILS:
	This function allocates memory for the frameAL link context.
******************************************************************************/

GLOBAL NDIS_STATUS FrameALLinkInit(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN  FRAMEAL_LINK_PARAMETERS * LinkParameters) {
  FRAMEAL_LINK_T  *pFrameALLinkT;
  NDIS_STATUS	Status;
  CDSL_ADAPTER_T *Adapter = LinkHdl->pThisAdapter;
  CDSL_ADAPTER_FRAMEAL_T *FrameALAdprCtx = Adapter->FrameMod;

  Status = ALLOCATE_MEMORY(pFrameALLinkT, sizeof(*pFrameALLinkT));
  if (Status != STATUS_SUCCESS)
    return Status;

  // clear FrameAL Link structure
  CLEAR_MEMORY(pFrameALLinkT, sizeof(*pFrameALLinkT));

  // save Frame link structure pointer in link structure
  LinkHdl->FrameMod = pFrameALLinkT;

  // save parameter block in FrameAL link Stucture
  COPY_MEMORY(&pFrameALLinkT->Parameters, LinkParameters,
	      sizeof(*LinkParameters));

  //
  // Copy into the FrameAL Link Structure any registry settings
  pFrameALLinkT->Parameters.CellBurstSize = FrameALAdprCtx->CellBurstSize;

  // set parameters for Traffic shaper.
  pFrameALLinkT->FrameALShaperInfo.ClientLimit = CLIENTLIMIT;
  pFrameALLinkT->FrameALShaperInfo.LineCheckInterval = LINE_CHECK_INTERVAL;
  pFrameALLinkT->FrameALShaperInfo.DefaultBurstSize = DEFAULT_BURST_SIZE;

  pFrameALLinkT->FrameALShaperInfo.TimerIntervalUpper = TIMER_INTERVAL_UPPER;
  pFrameALLinkT->FrameALShaperInfo.TimerIntervalLower = TIMER_INTERVAL_LOWER;

  Status = FrameALATMLinkInit(LinkHdl);

  if (Status != STATUS_SUCCESS)
    return Status;

  INIT_EVENT(&pFrameALLinkT->FrameALTimerStoppedEvent);

  INIT_TIMER(&pFrameALLinkT->FrameALTimer, FrameALTimerHandler,
	     (ULONG)LinkHdl);

  pFrameALLinkT->FrameALTimerState = TMR_TIMING;

  //	start back ground loop
  START_TIMER(&pFrameALLinkT->FrameALTimer, FRAMEAL_TIMER_PERIOD);

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALVCInit

ABSTRACT:
	Initializes a new VC instance.  

RETURN:
	NDIS status

DETAILS:
	This function allocates memory for the frameAL VC context.
******************************************************************************/

GLOBAL NDIS_STATUS FrameALVCInit(
    IN	CDSL_VC_T		*VCHdl,		// Link Handle to data stream
    IN  FRAMEAL_VC_PARAMETERS * VCParameters) {
#if 0
  CDSL_ADAPTER_T		*Adapter = VCHdl->LinkHdl->pThisAdapter;
  CDSL_ADAPTER_FRAMEAL_T	*FrameALAdprCtx = (CDSL_ADAPTER_FRAMEAL_T *) Adapter->FrameMod;
#endif
  FRAMEAL_VC_T			*pFrameALVC;
  NDIS_STATUS			Status;

  // Allocate NDIS block of memory

  Status = ALLOCATE_MEMORY(pFrameALVC, sizeof(*pFrameALVC));
  if (Status != STATUS_SUCCESS)
    return Status;

  // clear FrameAL Link structure
  CLEAR_MEMORY(pFrameALVC, sizeof(*pFrameALVC));

  // save Frame link structure pointer in link structure
  VCHdl->FrameALVCMod = pFrameALVC;

  // save parameter block in FrameAL link Stucture
  COPY_MEMORY(&pFrameALVC->Parameters, VCParameters, sizeof(*VCParameters));

#if 0
  pFrameALVC->Parameters.RfcEncapsulationMode = FrameALAdprCtx->RfcEncapsulationMode;
#endif

  Status = FrameALATMVCInit(VCHdl);
  if (Status != STATUS_SUCCESS)
    return Status;

  Status = FrameALAALVCInit(VCHdl);
  if (Status != STATUS_SUCCESS)
    return Status;

  // mark this VC as in use
  pFrameALVC->State = FRAMEAL_VC_ACTIVE;

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkShutdown

ABSTRACT:
	Shutdowns a link instance.  

RETURN:
	NDIS Status

DETAILS:
	This function releases any buffers (TX or RX) allocated to FrameAL by
	buffer management. After the buffers are released, the memory for 
	the FrameAL link context is released..
******************************************************************************/

GLOBAL NDIS_STATUS FrameALLinkShutdown(
    IN	CDSL_LINK_T	*LinkHdl) {
  FRAMEAL_LINK_T  *pFrameALLinkT = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

  if (pFrameALLinkT)
  {
    BOOLEAN 	TimerCanceled;

    // Let the ATM layer release its block of memory
    FrameALATMLinkShutdown(LinkHdl);
    pFrameALLinkT->FrameALTimerState =TMR_CANCELLED;
    CANCEL_TIMER(&pFrameALLinkT->FrameALTimer, &TimerCanceled);

    if (!TimerCanceled)
      TimerCanceled = WAIT_EVENT_TIMEOUT(&pFrameALLinkT->FrameALTimerStoppedEvent,
					 FRAMEAL_TIMER_PERIOD);

    // release NDIS block of memory
    FREE_MEMORY(LinkHdl->FrameMod, sizeof(FRAMEAL_LINK_T), 0);

    //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
    //		         LinkHdl->pThisAdapter,
    //		         ("FrameAL: Exit link shutdown"));

    LinkHdl->FrameMod = NULL;
  }

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALVCShutdown

ABSTRACT:
	Shutdowns a VC instance.  

RETURN:
	NDIS Status

DETAILS:
   	First the sublayers are notified of the VC shutdown. Then FrameAL 
   	releases the memory for	the FrameAL link context is released.
******************************************************************************/

GLOBAL NDIS_STATUS FrameALVCShutdown(
    IN	CDSL_VC_T	*VCHdl) {
  FRAMEAL_VC_T  *pFrameALVC = (FRAMEAL_VC_T*)VCHdl->FrameALVCMod;
  NDIS_STATUS Status;

  if (!pFrameALVC) // must already be shutdown
    return STATUS_SUCCESS;

  // Let the ATM layer release its block of memory
  Status = FrameALATMVCShutdown(VCHdl);
  if (Status == STATUS_SUCCESS) // Let the ATM layer release its block of memory
    Status = FrameALAALVCShutdown(VCHdl);

  if (Status == COMMAND_PENDING) {
    pFrameALVC->State = FRAMEAL_VC_DEACTIVE_PEND;
    return Status;
  }

  VCHdl->FrameALVCMod = NULL;

  // release NDIS block of memory
  FREE_MEMORY(pFrameALVC, sizeof(*pFrameALVC), 0);

  //	DBG_CDSL_DISPLAY(DBG_LVL_MISC_INFO,
  //		         LinkHdl->pThisAdapter,
  //		         ("FrameAL: Exit link shutdown"));

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALEvent

ABSTRACT:
	Processes an event. 
	
RETURNS:
	None
		  
DETAILS:
	No events are currently defined
******************************************************************************/

GLOBAL void FrameALEvent(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	WORD		 Event,		// Event enumeration
    IN	void		*Data) {	// Pointer to further data dependent on event
  // Nothing to do here
}

/******************************************************************************
FUNCTION NAME:
	FrameALCfgInit

ABSTRACT:
	Processes an configuration information.   

RETURN:
	NDIS Status

DETAILS:
	
******************************************************************************/

GLOBAL NDIS_STATUS FrameALCfgInit(
    IN	CDSL_ADAPTER_T	*AdprCtx,	// Driver supplied adapter context
    // (&ThisAdapter)  Required for events that are not associated with a link
    IN PTIG_USER_PARAMS  pUserParams) {
  int Status;
  CDSL_ADAPTER_FRAMEAL_T	*FrameALAdprCtx;

  // Allocate memory for the FrameAL adapter context structure.
  Status = ALLOCATE_MEMORY(FrameALAdprCtx, sizeof(*FrameALAdprCtx));
  if (Status == STATUS_SUCCESS) {
    // Initialize the FrameAL context structure values to NULL
    // and fill the Adapter Context structure with the pointer value
    // to the newly allocated FrameAL adpater context structure.
    CLEAR_MEMORY(FrameALAdprCtx, sizeof(*FrameALAdprCtx));
    AdprCtx->FrameMod = FrameALAdprCtx;
  } else {
    // FrameAL could not allocate an adapter context
    // structure so bail out with an error status.
    return Status;
  }

#if 0
  FrameALAdprCtx->RfcEncapsulationMode = pUserParams->RfcEncapsulationMode;
#endif
  FrameALAdprCtx->CellBurstSize = pUserParams->CellBurstSize;

  return STATUS_SUCCESS;
}

/******************************************************************************
FUNCTION NAME:
	FrameALStatsRequest

ABSTRACT:
	Return the statistics block to the caller..  

RETURNS:
	Pointer to Link Statistics

DETAILS:
    Return the a pointer to link statistics of 
    a given link.		    
******************************************************************************/

GLOBAL FRAMEAL_LINK_STATS_T * FrameALStatsRequest(
    IN	CDSL_LINK_T		*LinkHdl) {
  // Get pointer to statistics structure for this link
  FRAMEAL_LINK_T  *pFrameALLinkT = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

  return (&pFrameALLinkT->Stats);
}

/******************************************************************************
FUNCTION NAME:
	FrameALDeviceSpecific

ABSTRACT:
	Process Device Specific OIDs

RETURN:
	Ndis Status

DETAILS:
	Only OID is for frame AL statistics. 
******************************************************************************/
GLOBAL NDIS_STATUS FrameALDeviceSpecific(
    CDSL_ADAPTER_T		* pThisAdapter,
    BACK_DOOR_T			* pBackDoorBuf,
    PULONG			pNumBytesNeeded,
    PULONG			pNumBytesWritten) {
  NDIS_STATUS		Status=STATUS_SUCCESS;
  DWORD 		VCIndex;
  FRAMEAL_LINK_T	* pFrameALLink;
  CDSL_LINK_T		* LinkHdl;
  FRAMEAL_LINK_T	* pFrameALLinkT;
  CDSL_ADAPTER_FRAMEAL_T	* FrameALAdprCtx;

  //	Assume BACKDOOR ReqCode will be processed,
  //	set the BACKDOOR Status to Code Not Supported.
  //	NOTE:	The BackDoor Status should ONLY BE SET TO
  //	CODE_NOT_SUPPORTED at the beginning of BackDoor
  //	parsing by the first OID parser (CardMgmt)!!!
  //	ALL OTHER successor OID parsing of BackDoor should
  //	only set the BackDoor Status when it is intended
  //	to be set to a specific Status (i.e., BAD_SIGNATURE,
  //	FAILURE, or SUCCESS)!!  The BackDoor Status should
  //	not be set at the beginning of BackDoor parsing
  //	intending to change it if needed!!!

  ASSERT(pThisAdapter);

  if (pBackDoorBuf->TotalSize < SIZEOF_DWORD_ALIGN(BACK_DOOR_T)) {
    if (pBackDoorBuf->TotalSize != SIZEOF_DWORD_ALIGN(MINI_BACK_DOOR_T)) {
      pBackDoorBuf->ResultCode = ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
      pBackDoorBuf->NeededSize = SIZEOF_DWORD_ALIGN(BACK_DOOR_T);
      *pNumBytesNeeded = pBackDoorBuf->NeededSize;
      return ERROR_WARNING(RESULT_REQ_BUFF_TOO_SMALL);
    }
  }

  // only process request codes for FRAMEAL
  switch (pBackDoorBuf->ReqCode) {
  case  BD_FRAMEAL_GET_STATS:
    // Get statistics for display on Control Panel

    //	get stats for link.
    LinkHdl = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (LinkHdl->Allocated && LinkHdl->FrameMod) {
      pFrameALLinkT = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;
      pBackDoorBuf->Params.BdFrameALStats.NumTxBytes		= pFrameALLinkT->Stats.TXBytes.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumTxFrames		= pFrameALLinkT->Stats.TXFrames.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumRxBytes		= pFrameALLinkT->Stats.RXBytes.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumRxFrames 		= pFrameALLinkT->Stats.RXFrames.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumCrcErrs		= pFrameALLinkT->Stats.RXErrors.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumAbortErrs		= pFrameALLinkT->Stats.RXAborts.Cnt;
      pBackDoorBuf->Params.BdFrameALStats.NumTxUnderrunErrs	= pFrameALLinkT->Stats.TXErrors.Cnt;

      // mark OID as successful
      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
    } else {
      pBackDoorBuf->Params.BdFrameALStats.NumTxBytes		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumTxFrames		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumRxBytes		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumRxFrames 		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumCrcErrs		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumAbortErrs 		= 0;
      pBackDoorBuf->Params.BdFrameALStats.NumTxUnderrunErrs	= 0;

      pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_DATA_UNAVAILABLE);
    }

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten = sizeof(pBackDoorBuf->Params.BdFrameALStats) +
                        sizeof(pBackDoorBuf->ResultCode) +
                        sizeof(pBackDoorBuf->NeededSize);
    break;

  case BD_FRAMEAL_ATM_GET_STATS:
    pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_DATA_UNAVAILABLE);

    //
    //	Return ATM layer stats to the control panel
    //  Control panel sends down an array of VC Indexs of VCs the control
    //  panel wants stats for.
    //  The driver checks each index to see if there is a match to a valid VC.
    // 	If the Index is valid, the stats for that VC are placed in next 
    //  available entry of the VCArray
    //  Link Stats are marked using a special value for the VC index (0xffff).
    //  If the driver finds that value in the list of requested VCs. The links
    //  stats are placed in the next available VCArray entry.

    //  Will only work for one link
    LinkHdl = GET_LINK_FROM_ADAPTER(pThisAdapter);
    if (LinkHdl->Allocated && LinkHdl->FrameMod) {
      // check for a request of the links stats.
      if (pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex ==
	  BD_ATM_STATS_INDEX_FOR_LINK) {
	// get a ptr to structure where the links stats are stored
	pFrameALLink = LinkHdl->FrameMod;

	if (pFrameALLink) {
	  // move the stats to the VcArray entry
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumTxBytes =
	    pFrameALLink->Stats.TXBytes.Cnt;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumTxCells =
	    pFrameALLink->Stats.TXATMCellsCLP[0].Cnt +
	    pFrameALLink->Stats.TXATMCellsCLP[1].Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumTxMgmtCells =
	    pFrameALLink->Stats.TXATMCellsMgmCell.Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumTxClpEqual0Cells =
	    pFrameALLink->Stats.TXATMCellsCLP[0].Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumTxClpEqual1Cells =
	    pFrameALLink->Stats.TXATMCellsCLP[1].Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxBytes =
	    pFrameALLink->Stats.RXBytes.Cnt;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxCells =
	    pFrameALLink->Stats.RXATMCellsGoodCLP[0].Cnt +
	    pFrameALLink->Stats.RXATMCellsGoodCLP[1].Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxMgmtCells =
	    pFrameALLink->Stats.RXATMCellsMgmCell.Cnt;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxClpEqual0Cells =
	    pFrameALLink->Stats.RXATMCellsGoodCLP[0].Cnt;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxClpEqual1Cells =
	    pFrameALLink->Stats.RXATMCellsGoodCLP[1].Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxHecErrs =
	    pFrameALLink->Stats.RXATMCellsBadHEC.Cnt;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxCellAlignErrs = 0;
	  pBackDoorBuf->Params.BdFrameALAtmStats.NumRxUnroutCellErrs =
	    pFrameALLink->Stats.RXATMCellsMisrouted.Cnt;

	  pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = BD_ATM_STATS_INDEX_FOR_LINK;

	  pBackDoorBuf->ResultCode = RESULT_SUCCESS;
	} else // A link structure is not available
	  break;
      } else {
	VCIndex = pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex;
	if (VCIndex < MAX_VC_PER_LINK) {
	  // get the stats for the requested VC
	  Status = FrameALUpdateATMStats(&LinkHdl->VC_Info [VCIndex],
					 &pBackDoorBuf->Params.BdFrameALAtmStats);
	  if (Status == VC_STATS_OK) {
	    // update number of VCarray entries in use
	    // pBackDoorBuf->Params.BdFrameALAtmStats.VcIndex = VCIndex;
	    pBackDoorBuf->ResultCode = RESULT_SUCCESS;
	  }
	}
      }
    }

    if (pBackDoorBuf->ResultCode != RESULT_SUCCESS) {
      pBackDoorBuf->Params.BdFrameALAtmStats.NumTxBytes = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumTxCells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumTxMgmtCells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumTxClpEqual0Cells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumTxClpEqual1Cells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxBytes = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxCells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxMgmtCells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxClpEqual0Cells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxClpEqual1Cells = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxHecErrs = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxCellAlignErrs = 0;
      pBackDoorBuf->Params.BdFrameALAtmStats.NumRxUnroutCellErrs = 0;
    }

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten = sizeof(pBackDoorBuf->Params.BdFrameALAtmStats) +
                        sizeof(pBackDoorBuf->ResultCode) +
                        sizeof(pBackDoorBuf->NeededSize);
    break;

  case BD_FRAMEAL_AAL_GET_STATS:
   pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_DATA_UNAVAILABLE);

   pBackDoorBuf->Params.BdFrameALAalStats.AalType = BD_TESTMOD_AAL_TYPE_START;

   //
   //  Return AAL layer stats to the control panel
   //  Control panel sends down an array of VC Indexs of VCs the control panel
   //  wants stats for.
   //  The driver checks each index to see if there is a match to a valid VC.
   //  If the Index is valid, the stats for that VC are placed in next
   //  available entry of the VCArray.
   //  Link Stats are marked using a special value for the VC index (0xffff).
   //  If the driver finds that value in the list of requested VCs. The links 
   //  stats are placed in the next available VCArray entry.

   // Will only work for one link
   LinkHdl = GET_LINK_FROM_ADAPTER(pThisAdapter);
   if (LinkHdl->Allocated && LinkHdl->FrameMod) {
     // check for a request of the links stats.
     if (pBackDoorBuf->Params.BdFrameALAalStats.VcIndex == BD_AAL_STATS_INDEX_FOR_LINK) {
       // get a ptr to structure where the links stats are stored/

       pFrameALLink = LinkHdl->FrameMod;
       if (pFrameALLink) {
	 BD_FRAMEAL_AAL_STATS_T *pStats = &pBackDoorBuf->Params.BdFrameALAalStats;
	 const FRAMEAL_AAL_LINK_T *pAALLinkBlock = &pFrameALLink->AALLinkBlock;

	 pStats->NumTxGoodBytes = pAALLinkBlock->AALTxGoodByte.Cnt;
	 pStats->NumTxGoodFrames = pAALLinkBlock->AALTxGood.Cnt;
	 pStats->NumTxDiscardedBytes = pAALLinkBlock->AALTxDiscarded.Cnt;
	 pStats->NumTxDiscardedFrames = pAALLinkBlock->AALTxDiscardedByte.Cnt;
	 pStats->NumRxGoodBytes = pAALLinkBlock->AALRxGoodByte.Cnt;
	 pStats->NumRxGoodFrames = pAALLinkBlock->AALRxGood.Cnt;
	 pStats->NumRxDiscardedBytes = pAALLinkBlock->AALRxDiscardedByte.Cnt;
	 pStats->NumRxDiscardedFrames = pAALLinkBlock->AALRxDiscarded.Cnt;
	 pStats->NumRxCrcErrs = pAALLinkBlock->AALRxInvalidCRC.Cnt;
	 pStats->NumRxInvalidLenErrs = pAALLinkBlock->AALRxInvalidLen.Cnt;
	 pStats->NumRxTimeoutErrs = pAALLinkBlock->AALRxReassemblyTimeout.Cnt;

	 // pStats->VcIndex = BD_AAL_STATS_INDEX_FOR_LINK;

	 pBackDoorBuf->ResultCode = RESULT_SUCCESS;
       } else // A link structure is not available
	 break;
     } else {
       VCIndex = pBackDoorBuf->Params.BdFrameALAalStats.VcIndex;
       if (VCIndex < MAX_VC_PER_LINK) {
	 // get the stats for the requested VC
	 Status = FrameALUpdateAALStats(&LinkHdl->VC_Info[VCIndex],
					&pBackDoorBuf->Params.BdFrameALAalStats);

	 if (Status == VC_STATS_OK) {
	   // pBackDoorBuf->Params.BdFrameALAalStats.VcIndex = VCIndex;
	   pBackDoorBuf->ResultCode = RESULT_SUCCESS;
	 }
       }
     }
   }

   if (pBackDoorBuf->ResultCode != RESULT_SUCCESS) {
     pBackDoorBuf->Params.BdFrameALAalStats.NumTxGoodBytes = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumTxGoodFrames = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumTxDiscardedBytes = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumTxDiscardedFrames = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxGoodBytes = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxGoodFrames = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxDiscardedBytes = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxDiscardedFrames = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxCrcErrs = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxInvalidLenErrs = 0;
     pBackDoorBuf->Params.BdFrameALAalStats.NumRxTimeoutErrs = 0;
   }

   *pNumBytesNeeded = pBackDoorBuf->TotalSize;
   *pNumBytesWritten = sizeof(pBackDoorBuf->Params.BdFrameALAalStats) +
                       sizeof(pBackDoorBuf->ResultCode) +
                       sizeof(pBackDoorBuf->NeededSize);
   break;

  case BD_FRAMEAL_GET_CONFIG:
    // Get FrameAL configuration for display on Control Panel
    FrameALAdprCtx = (CDSL_ADAPTER_FRAMEAL_T *) pThisAdapter->FrameMod;

#if 0
    switch (FrameALAdprCtx->RfcEncapsulationMode) {
    case BD_ENCAPSULATION_PPPOA_VCMUX:	// PPP over ATM VXMUX
    case BD_ENCAPSULATION_PPPOA_LLC:	// PPP over ATM LLC
    case BD_ENCAPSULATION_BIPOA_LLC:	// Bridged IP over ATM LLCSNAP
    case BD_ENCAPSULATION_RIPOA_LLC:	// Routed IP over ATM LLCSNAP
    case BD_ENCAPSULATION_BIPOA_VCMUX:	// Bridged IP over ATM VCMUX
    case BD_ENCAPSULATION_RIPOA_VCMUX:	// Routed IP over ATM VCMUX
      // The next three enumerations are added for CNXT-defined encapsulations,
      // they are not found in the ADSL Forum documentation.
    case BD_ENCAPSULATION_PPPOA_NONE:	// PPP over ATM no encapsulation
    case BD_ENCAPSULATION_BIPOA_NONE:	// Bridged IP over ATM no encapsulation
    case BD_ENCAPSULATION_RIPOA_NONE:	// Routed IP over ATM no encapsulation
      pBackDoorBuf->Params.BdFrameALConfig.RfcEncapsulationMode =
	FrameALAdprCtx->RfcEncapsulationMode;
      // mark OID as successful
      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      break;

    case BD_ENCAPSULATION_IPOA:		// Classical IP over ATM
    case BD_ENCAPSULATION_NATIVE_ATM:	// Native ATM
    case BD_ENCAPSULATION_PROPRIETARY:	// Proprietary
    default:
      pBackDoorBuf->Params.BdFrameALConfig.RfcEncapsulationMode =
	FrameALAdprCtx->RfcEncapsulationMode;
      pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_PARAM_OUT_OF_RANGE);
      break;
    }
#else
    pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_PARAM_OUT_OF_RANGE);
#endif

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten = sizeof(pBackDoorBuf->Params.BdFrameALConfig) +
                        sizeof(pBackDoorBuf->ResultCode) +
                        sizeof(pBackDoorBuf->NeededSize);
    break;

  case BD_FRAMEAL_SET_CONFIG:
    // Get FrameAL configuration for display on Control Panel
    FrameALAdprCtx = (CDSL_ADAPTER_FRAMEAL_T *) pThisAdapter->FrameMod;

#if 0
    switch (pBackDoorBuf->Params.BdFrameALConfig.RfcEncapsulationMode) {
    case BD_ENCAPSULATION_PPPOA_VCMUX:	// PPP over ATM VXMUX
    case BD_ENCAPSULATION_PPPOA_LLC:	// PPP over ATM LLC
    case BD_ENCAPSULATION_BIPOA_LLC:	// Bridged IP over ATM LLCSNAP
    case BD_ENCAPSULATION_RIPOA_LLC:	// Routed IP over ATM LLCSNAP
    case BD_ENCAPSULATION_BIPOA_VCMUX:	// Bridged IP over ATM VCMUX
    case BD_ENCAPSULATION_RIPOA_VCMUX:	// Routed IP over ATM VCMUX
      // The next three enumerations are added for CNXT-defined encapsulations,
      // they are not found in the ADSL Forum documentation.
    case BD_ENCAPSULATION_PPPOA_NONE:	// PPP over ATM no encapsulation
    case BD_ENCAPSULATION_BIPOA_NONE:	// Bridged IP over ATM no encapsulation
    case BD_ENCAPSULATION_RIPOA_NONE:	// Routed IP over ATM no encapsulation
      FrameALAdprCtx->RfcEncapsulationMode =
	pBackDoorBuf->Params.BdFrameALConfig.RfcEncapsulationMode;
      // mark OID as successful
      pBackDoorBuf->ResultCode = RESULT_SUCCESS;
      break;

    case BD_ENCAPSULATION_IPOA:		// Classical IP over ATM
    case BD_ENCAPSULATION_NATIVE_ATM:	// Native ATM
    case BD_ENCAPSULATION_PROPRIETARY:	// Proprietary
    default:
      pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_PARAM_OUT_OF_RANGE);
      break;
    }
#else
    pBackDoorBuf->ResultCode = ERROR_NORMAL(RESULT_PARAM_OUT_OF_RANGE);
#endif

    *pNumBytesNeeded = pBackDoorBuf->TotalSize;
    *pNumBytesWritten = sizeof(pBackDoorBuf->Params.BdFrameALConfig) +
                        sizeof(pBackDoorBuf->ResultCode) +
                        sizeof(pBackDoorBuf->NeededSize);
    break;

  default:
    Status = STATUS_UNSUPPORTED;
    break;
  }

  return Status;
}

/******************************************************************************
FUNCTION NAME:
	FrameALUpdateATMStats
	
ABSTRACT:

RETURN:
	ATM Stats Status

DETAILS:
	Get the stats for VC.
******************************************************************************/
LOCAL DWORD FrameALUpdateATMStats(
    IN 	CDSL_VC_T * pVCHdl,
    IN 	BD_FRAMEAL_ATM_STATS_T * pATMStats) {
  if (pVCHdl) {
    FRAMEAL_VC_T *pFrameALVC;

    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;
    if (pFrameALVC) {
      pATMStats->Vpi = pFrameALVC->Parameters.Vpi;
      pATMStats->Vci = pFrameALVC->Parameters.Vci;
      pATMStats->NumTxBytes = pFrameALVC->Stats.TXBytes.Cnt;
      pATMStats->NumTxCells = pFrameALVC->Stats.TXATMCellsCLP[0].Cnt +
	                      pFrameALVC->Stats.TXATMCellsCLP[1].Cnt;

      pATMStats->NumTxMgmtCells	= pFrameALVC->Stats.TXATMCellsMgmCell.Cnt;

      pATMStats->NumTxClpEqual0Cells = pFrameALVC->Stats.TXATMCellsCLP[0].Cnt;
      pATMStats->NumTxClpEqual1Cells = pFrameALVC->Stats.TXATMCellsCLP[1].Cnt;
      pATMStats->NumRxBytes = pFrameALVC->Stats.RXBytes.Cnt;
      pATMStats->NumRxCells = pFrameALVC->Stats.RXATMCellsGoodCLP[0].Cnt +
			      pFrameALVC->Stats.RXATMCellsGoodCLP[1].Cnt;

      pATMStats->NumRxMgmtCells	= pFrameALVC->Stats.RXATMCellsMgmCell.Cnt;
      pATMStats->NumRxClpEqual0Cells = pFrameALVC->Stats.RXATMCellsGoodCLP[0].Cnt;
      pATMStats->NumRxClpEqual1Cells = pFrameALVC->Stats.RXATMCellsGoodCLP[1].Cnt;

      return VC_STATS_OK;
    }
  }

  return VC_STATS_FAILED;
}

/******************************************************************************
FUNCTION NAME:
	FrameALUpdateAALStats
	
ABSTRACT:

RETURN:
	AAL Stats Status

DETAILS:  
	Get the stats for VC.
******************************************************************************/
LOCAL DWORD FrameALUpdateAALStats(
    IN 	CDSL_VC_T * pVCHdl,
    IN 	BD_FRAMEAL_AAL_STATS_T * pAALStats) {
  if (pVCHdl) {
    FRAMEAL_VC_T *pFrameALVC;

    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;
    if (pFrameALVC) {
      pAALStats->AalType		= pFrameALVC->Parameters.AALType;
      pAALStats->Vpi			= pFrameALVC->Parameters.Vpi;
      pAALStats->Vci			= pFrameALVC->Parameters.Vci;
      pAALStats->NumTxGoodBytes		= pFrameALVC->AALVcCtx.AALTxGoodByte.Cnt;
      pAALStats->NumTxGoodFrames	= pFrameALVC->AALVcCtx.AALTxGood.Cnt;
      pAALStats->NumTxDiscardedBytes	= pFrameALVC->AALVcCtx.AALTxDiscarded.Cnt;
      pAALStats->NumTxDiscardedFrames	= pFrameALVC->AALVcCtx.AALTxDiscardedByte.Cnt;
      pAALStats->NumRxGoodBytes		= pFrameALVC->AALVcCtx.AALRxGoodByte.Cnt;
      pAALStats->NumRxGoodFrames	= pFrameALVC->AALVcCtx.AALRxGood.Cnt;
      pAALStats->NumRxDiscardedBytes	= pFrameALVC->AALVcCtx.AALRxDiscardedByte.Cnt;
      pAALStats->NumRxDiscardedFrames	= pFrameALVC->AALVcCtx.AALRxDiscarded.Cnt;
      pAALStats->NumRxCrcErrs		= pFrameALVC->AALVcCtx.AALRxInvalidCRC.Cnt;
      pAALStats->NumRxInvalidLenErrs	= pFrameALVC->AALVcCtx.AALRxInvalidLen.Cnt;
      pAALStats->NumRxTimeoutErrs	= pFrameALVC->AALVcCtx.AALRxReassemblyTimeout.Cnt;

      return VC_STATS_OK;
    }
  }

  return VC_STATS_FAILED;
}

/******************************************************************************
FUNCTION NAME:
	FrameALSetInformationHdlr

ABSTRACT:
	This function is called from the CDSLSetInformationHdlr function
	(the function called by NDIS to handle this drivers
	SetInformationHandler calls) to handle the Information Handler
	Sets that come through NDIS via WAN.

RETURN:
	NDIS Status

DETAILS:
	Frame AL does not support any Set OIDs at this time.
******************************************************************************/
/* not supported by Linux */
#if 0
NDIS_STATUS FrameALSetInformationHdlr(
    IN CDSL_ADAPTER_T	* pThisAdapter,
    IN NDIS_OID		Oid,
    IN PVOID		InfoBuffer,
    IN ULONG		InfoBufferLength,
    OUT PULONG		pBytesRead,
    OUT PULONG		pBytesNeeded) {
  NDIS_STATUS		Status;
  ULONG			NumBytesNeeded;
  ULONG			NumBytesRead;

  //	DBG_CDSL_DISPLAY(DBG_LVL_WARNING,
  //		         pThisAdapter,
  //		         ("    Function Entry 'FrameALSetInformationHdlr':   OID=0x%08lX.\n", Oid));

  NumBytesNeeded = 0;
  NumBytesRead = 0;

  //	Determine which OID is being requested and do the right thing.
  switch (Oid) {
  default:
    //	Unknown OID
    //	DBG_CDSL_DISPLAY(DBG_LVL_WARNING,
    //		         pThisAdapter,
    //		         ("    Function 'FrameALSetInformationHdlr':   Unknown OID  -  0x%08lX.\n", Oid));
    Status = NDIS_STATUS_INVALID_OID;
    break;
  }

  //	Fill in the size fields before we leave.
  *pBytesNeeded += NumBytesNeeded;
  *pBytesRead += NumBytesRead;

  return (Status);
}
#endif

#if PROJECTS_NDIS_IF == PROJECT_NDIS_IF_CO
/******************************************************************************
FUNCTION NAME:
	FrameALStatsRequestHdlr	
ABSTRACT:
	This function is called from the CoNDIS request handler to 
	get stats.

RETURN:
	NDIS Status

DETAILS:
******************************************************************************/

GLOBAL NDIS_STATUS FrameALRequestHdlr(
    IN		CDSL_ADAPTER_T 	*pAdapter,
    IN		CDSL_VC_T 	*pVCHdl,
    IN OUT	PNDIS_REQUEST	pRequest) {
  NDIS_STATUS	Status = STATUS_SUCCESS;
  ULONG		Data = 0;
  PCHAR	   	DataToMove = (PUCHAR)&Data;
  ULONG	   	DataLength = 0;
  ULONG		Statistics;
  CDSL_LINK_T	*LinkHdl;
  FRAMEAL_VC_T  *pFrameALVC;
  FRAMEAL_LINK_T	*pFrameALLink;

  if (!pVCHdl) {
    if (pAdapter) {
      LinkHdl = GET_LINK_FROM_ADAPTER(pAdapter->AdapterLink);

      pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;

      switch (pRequest->DATA.QUERY_INFORMATION.Oid) {
      case OID_GEN_CO_BYTES_XMIT:
	DataToMove = (PUCHAR)&pFrameALLink->Stats.COTXBytes.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.COTXBytes.Cnt);
	break;

      case OID_GEN_CO_BYTES_RCV:
	DataToMove = (PUCHAR)&pFrameALLink->Stats.CORXBytes.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.CORXBytes.Cnt);
	break;

      case OID_GEN_CO_RCV_CRC_ERROR:
	Data = pFrameALLink->AALLinkBlock.AALRxInvalidCRC.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.RXBytes.Cnt);
	break;

      case OID_GEN_CO_XMIT_PDUS_OK:
	Data = pFrameALLink->AALLinkBlock.AALTxGood.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_RCV_PDUS_OK:
	Data = pFrameALLink->AALLinkBlock.AALRxGood.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_XMIT_PDUS_ERROR:
	Data = pFrameALLink->AALLinkBlock.AALTxDiscarded.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_RCV_PDUS_ERROR:
	Data = pFrameALLink->AALLinkBlock.AALRxInvalidCRC.Cnt +
	  pFrameALLink->AALLinkBlock.AALRxInvalidLen.Cnt +
	  pFrameALLink->AALLinkBlock.AALRxReassemblyTimeout.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_RCV_PDUS_NO_BUFFER:
	break;

      case OID_ATM_RCV_CELLS_OK:
	DataToMove = (PUCHAR)&pFrameALLink->Stats.CORXATMCells.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.CORXATMCells.Cnt);
	break;

      case OID_ATM_XMIT_CELLS_OK:
	DataToMove = (PUCHAR)&pFrameALLink->Stats.COTXATMCells.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_ATM_RCV_CELLS_DROPPED:
	break;

      case OID_ATM_CELLS_HEC_ERROR:
	DataToMove = (PUCHAR)&pFrameALLink->Stats.CORXATMCellsDrops.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.CORXATMCellsDrops.Cnt);
	break;

      case OID_ATM_RCV_REASSEMBLY_ERROR:
	break;

      case OID_ATM_RCV_INVALID_VPI_VCI:
	Data = pFrameALLink->Stats.RXATMCellsMisrouted.Cnt;
	DataLength = sizeof(pFrameALLink->Stats.RXATMCellsMisrouted.Cnt);
	break;

      default:
	Status = NDIS_STATUS_NOT_SUPPORTED;
	break;
      }
    }
  } else {
    pFrameALVC = (FRAMEAL_VC_T *) pVCHdl->FrameALVCMod;

    if (pFrameALVC) {
      switch (pRequest->DATA.QUERY_INFORMATION.Oid) {
      case OID_GEN_CO_BYTES_XMIT:
	DataToMove = (PUCHAR)&pFrameALVC->Stats.COTXBytes.Cnt;
	DataLength = sizeof(pFrameALVC->Stats.COTXBytes.Cnt);
	break;

      case OID_GEN_CO_BYTES_RCV:
	DataToMove = (PUCHAR)&pFrameALVC->Stats.CORXBytes.Cnt;
	DataLength = sizeof(pFrameALVC->Stats.CORXBytes.Cnt);
	break;

      case OID_GEN_CO_XMIT_PDUS_OK:
	Data = pFrameALVC->AALVcCtx.AALTxGood.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_RCV_PDUS_OK:
	Data = pFrameALVC->AALVcCtx.AALRxGood.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_XMIT_PDUS_ERROR:
	Data = pFrameALVC->AALVcCtx.AALTxDiscarded.Cnt;
	DataLength = sizeof(ULONG);
	break;

      case OID_GEN_CO_RCV_PDUS_ERROR:
      case OID_GEN_CO_RCV_PDUS_NO_BUFFER:
      case OID_ATM_RCV_CELLS_OK:
	DataToMove = (PUCHAR)&pFrameALVC->Stats.CORXATMCells.Cnt;
	DataLength = sizeof(pFrameALVC->Stats.CORXATMCells.Cnt);
	break;

      case OID_ATM_XMIT_CELLS_OK:
	DataToMove = (PUCHAR)&pFrameALVC->Stats.CORXATMCells.Cnt;
	DataLength = sizeof(pFrameALVC->Stats.CORXATMCells.Cnt);
	break;

      case OID_ATM_RCV_CELLS_DROPPED:
      case OID_ATM_CELLS_HEC_ERROR:
      case OID_ATM_RCV_REASSEMBLY_ERROR:
	break;

      default:
	Status = NDIS_STATUS_NOT_SUPPORTED;
	break;
      }
    }
  }

  if (STATUS_SUCCESS == Status) {
    if (pRequest->DATA.QUERY_INFORMATION.InformationBufferLength < DataLength) {
      pRequest->DATA.QUERY_INFORMATION.BytesNeeded = DataLength;
      Status = NDIS_STATUS_INVALID_LENGTH;
    } else {
      if (DataLength > 0) {
	NdisMoveMemory(pRequest->DATA.QUERY_INFORMATION.InformationBuffer,
		       DataToMove,
		       DataLength);

	pRequest->DATA.QUERY_INFORMATION.BytesWritten = DataLength;
      }
    }
  }

  return(Status);
}
#endif //PROJECTS_NDIS_IF == PROJECT_NDIS_IF_CO

/******************************************************************************
FUNCTION NAME:
	FrameALGetTxFullBuffer

ABSTRACT:
	Provides a buffer of data that is to be transmitted.  

RETURNS: 
	NDISWAN packet pointer. 

DETAILS:
	The caller may optionally provide a buffer descriptor and associated 
	buffer that FrameAL must use. If FrameAL is transforming the data, 
	it has to do a copy anyway. In this mode, FrameAL must fill the 
	buffer exactly. Otherwise FrameAL will return a different buffer 
	descriptor, possibly avoiding any copying. In either case, 
	FrameAL returns a pointer to the buffer descriptor used
******************************************************************************/

GLOBAL SK_BUFF_T	*FrameALGetTxFullBuffer(
    IN	void		*pLink,		// Link Handle to data stream
    IN	SK_BUFF_T	*InBufDescPtr) {// Pointer to a buffer descriptor for FrameAL to use
  if (InBufDescPtr)
    return FrameALATMGetTxFullBuffer((CDSL_LINK_T*)pLink,InBufDescPtr);
  else
    return InBufDescPtr;
}

/******************************************************************************
FUNCTION NAME:
	FrameALPutTxEmptyBuffer

ABSTRACT:
	This function returns a buffer descriptor to FrameAL..  

RETURNS:
	None

DETAILS:
	This returns a buffer descriptor and its associated buffer back to 
	FrameAL after ChipAL no longer needs the data. If FrameAL is 
	using buffers that ChipAL provides when calling
	FrameALGetTxFullBuffer, then this function is not used.
******************************************************************************/

GLOBAL void FrameALPutTxEmptyBuffer(
    IN	void	       	*pLink,			//Link Handle to data stream
    IN	SK_BUFF_T      	*BufferDescPtr) {	//Pointer to a buffer descriptor
  CDSL_LINK_T		*LinkHdl = pLink;

  // let AAL sublayer release buffer
  FrameALATMTxPutEmptyBuf(LinkHdl,BufferDescPtr);
}

#if 0
/******************************************************************************
FUNCTION NAME:
	FrameALGetRxEmptyBuffer

ABSTRACT:
	This function returns a recevie buffer descriptor to ChipAL.  

RETURNS:
	None

DETAILS:
	This returns a buffer descriptor and its associated buffer for ChipAL 
	to fill up with received data. If ChipAL is providing its own buffers, 
	this function is not used.
******************************************************************************/

GLOBAL void FrameALGetRxEmptyBuffer(
    IN	void	       	*pLink,	       	// Link Handle to data stream
    OUT	SK_BUFF_T      	*BufferDescPtr) {//Pointer to a buffer descriptor
  // Not implemented for Linux
  //	CDSL_LINK_T		*LinkHdl = pLink;
  //NOT USED
}
#endif

/******************************************************************************
FUNCTION NAME:
	 FrameALPutRxFullBuffer

ABSTRACT:
	This function delivers received raw data to HDLC,	  

RETURNS:
	NDISWAN packet pointer

DETAILS:
	This function gives a received buffer descriptor and its associated 
	buffer to FrameAL. This will be a buffer previously provided 
	by FrameALGetRxEmptybuffer or a ChipAL provided buffer.

******************************************************************************/

GLOBAL SK_BUFF_T *FrameALPutRxFullBuffer(
    IN	void	       	*pLink,		// Link Handle to data stream
    IN	SK_BUFF_T      	*BufferDescPtr) {// Pointer to a buffer descriptor
  if (BufferDescPtr)
    return FrameALATMCellAlign((CDSL_LINK_T *)pLink, BufferDescPtr);
  else
    return (BufferDescPtr);
}

#if 0
/******************************************************************************
FUNCTION NAME:
	FrameALTxQNotEmpty

ABSTRACT:
	This function wakes up Frame AL TX when new data is ready.  

RETURNS:
	None

DETAILS:
	This function is not used in this implementation
******************************************************************************/

GLOBAL void FrameALTxQNotEmpty(
    IN	CDSL_LINK_T    	*LinkHdl) {		// Link Handle to data stream
#if 0
  NDIS_WAN_PACKET	BufferDescPtr;
  WORD Count;

  for (Count = 0; Count < 50; Count++) {
    FrameALGetTxFullBuffer(LinkHdl, &BufferDescPtr);
    if (BufferDescPtr.CurrentLength > 0)
      FrameALPutTxEmptyBuffer(LinkHdl, &BufferDescPtr);
  }
#endif 

  FrameALLinkStartTX(LinkHdl);
}
#endif

/******************************************************************************
FUNCTION NAME:
	FrameALTimerHandler

ABSTRACT:
	This function is called by ChipAL when the Hobbs timer expires

RETURNS:
	None

DETAILS:
******************************************************************************/
LOCAL void FrameALTimerHandler(
    IN ULONG 	pLinkH) {
  IN CDSL_LINK_T	* LinkHdl = (CDSL_LINK_T*)pLinkH;
  FRAMEAL_LINK_T	*pFrameALLinkT = (FRAMEAL_LINK_T *) LinkHdl->FrameMod;

  if (pFrameALLinkT) {
    if (pFrameALLinkT->FrameALTimerState == TMR_TIMING) {
      FrameALATMTimerHandler(LinkHdl);
      START_TIMER(&pFrameALLinkT->FrameALTimer, FRAMEAL_TIMER_PERIOD);
    } else // Timer is being shutdown or stopped
      SET_EVENT(&pFrameALLinkT->FrameALTimerStoppedEvent);
  }
}

/******************************************************************************
FUNCTION NAME:
	FrameALClearStats

ABSTRACT:
	This function is called to clear the link stats

RETURNS:
	None

DETAILS:
******************************************************************************/

GLOBAL void FrameALClearStats(
    IN	CDSL_LINK_T    	*LinkHdl) {		// Link Handle to data stream
  if (LinkHdl) {
    FRAMEAL_LINK_T  *pFrameALLink = (FRAMEAL_LINK_T*)LinkHdl->FrameMod;
    int cnt;

    if (pFrameALLink) {
      CLEAR_MEMORY(&pFrameALLink->Stats, sizeof(FRAMEAL_LINK_STATS_T));

      CLEAR_MEMORY(&pFrameALLink->AALLinkBlock, sizeof(FRAMEAL_AAL_LINK_T));
    }

    // clear the per vc stats
    for (cnt = 0; cnt < MAX_VC_PER_LINK; cnt++) {
      FRAMEAL_VC_T *pFrameALVC = (FRAMEAL_VC_T *) LinkHdl->VC_Info[cnt].FrameALVCMod;

      if (pFrameALVC) {
	// clear the AAL stats
	ZERO_CTR(pFrameALVC->AALVcCtx.AALTxGoodByte);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALTxGood);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALTxDiscarded);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALTxDiscardedByte);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxGoodByte);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxGood);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxDiscardedByte);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxDiscarded);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxInvalidCRC);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxInvalidLen);
	ZERO_CTR(pFrameALVC->AALVcCtx.AALRxReassemblyTimeout);

	// clear the atm stats
	ZERO_CTR(pFrameALVC->Stats.TXBytes);
	ZERO_CTR(pFrameALVC->Stats.TXATMCellsCLP[0]);
	ZERO_CTR(pFrameALVC->Stats.TXATMCellsCLP[1]);
	ZERO_CTR(pFrameALVC->Stats.TXATMCellsMgmCell);
	ZERO_CTR(pFrameALVC->Stats.TXATMCellsCLP[0]);
	ZERO_CTR(pFrameALVC->Stats.TXATMCellsCLP[1]);
	ZERO_CTR(pFrameALVC->Stats.RXBytes);
	ZERO_CTR(pFrameALVC->Stats.RXATMCellsGoodCLP[0]);
	ZERO_CTR(pFrameALVC->Stats.RXATMCellsGoodCLP[1]);
	ZERO_CTR(pFrameALVC->Stats.RXATMCellsMgmCell);
	ZERO_CTR(pFrameALVC->Stats.RXATMCellsGoodCLP[0]);
	ZERO_CTR(pFrameALVC->Stats.RXATMCellsGoodCLP[1]);
      }
    }
  }
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkUp

ABSTRACT:
	This function is called when the line enters showtime

RETURNS:
	None

DETAILS:
******************************************************************************/

GLOBAL void FrameALLinkUp(
    IN	CDSL_LINK_T	*LinkHdl) {		// Link Handle to data stream
  FrameALATMLinkUp(LinkHdl);
}

/******************************************************************************
FUNCTION NAME:
	FrameALLinkDown

ABSTRACT:
	This function is called when the line enters showtime

RETURNS:
	None

DETAILS:
******************************************************************************/

GLOBAL void FrameALLinkDown(
    IN  CDSL_LINK_T		*LinkHdl) { // Link Handle to data stream
  FrameALATMLinkDown(LinkHdl);
}

GLOBAL void FrameALBldHdr(
    IN  FRAMEAL_VC_T		* pFrameALVC,
    IN  SK_BUFF_T		* pBufDesc,
    IN  OOB_DATA_T		oob_data
) {
#if PRECOMPUTED_HEADERS
  CELLHEAD_DATA(pBufDesc) = &pFrameALVC->ATMVCBlock.TxPreCompHeader[oob_data.v & (PRECOMPUTED_HEADER_NUM-1)];
#else
  CELLHEAD_DATA(pBufDesc) = &pFrameALVC->ATMVCBlock.TxPreCompHeader;

#if PRECOMPUTED_HEADER_NUM == 32
  CELLHEAD_DATA(pBufDesc)->GFC = oob_data.b.GFC;
#if 1
  if (OOB_DATA(pBufDesc).b.GFC)
    printk("CnxADSL" " %s: GFC=%01x\n",
	   __FUNCTION__, oob_data.b.GFC);
#endif
#else
  CELLHEAD_DATA(pBufDesc)->GFC = 0;
#endif
  CELLHEAD_DATA(pBufDesc)->PTI_USER = oob_data.b.PTI_USER;
  CELLHEAD_DATA(pBufDesc)->PTI_CI = oob_data.b.PTI_CI;
  CELLHEAD_DATA(pBufDesc)->PTI_MGM = oob_data.b.PTI_MGM;
  CELLHEAD_DATA(pBufDesc)->CLP = oob_data.b.CLP;
  CELLHEAD_DATA(pBufDesc)->HEC = pFrameALVC->ATMVCBlock.TxPreCompHEC[oob_data.v & (PRECOMPUTED_HEADER_NUM-1)];
#endif
  CELLHEAD_LEN(pBufDesc) = CELL_HEADER_SIZE;
}
