/******************************************************************************
*******************************************************************************
****	Copyright (c) 1998
****	Rockwell Semiconductor Systems
****	Personal Computing Division
****	All Rights Reserved
****
****	CONFIDENTIAL and PROPRIETARY --
****		No Dissemination or use without prior written permission.
****
*******************************************************************************
**
**MODULE NAME:
**	ATM Layer of FrameAL
**
**FILE NAME:
**	FrameALATM.h
**
**ABSTRACT:
**	This file contains interface definitions for FrameAL ATM layer function
**	Those functions Traffic Shaping, Mux/DeMux and CellAlign functions. 
**
**DETAILS:
**
*******************************************************************************
** $Archive: /Projects/Tigris and Derivatives/Reference Designs/Tigris/Linux/Kernel Driver/FrameALATM.h $
** $Revision: 1 $
** $Date: 1/09/01 10:54a $
*******************************************************************************
******************************************************************************/

#ifndef _FRAMEALATM_H_
#define _FRAMEALATM_H_

#if PROJECTS_NDIS_IF == PROJECT_NDIS_IF_CO
	#include "FrameALATMCoShaper.h"
#else
	#include "FrameALATMShaper.h"
#endif

#include "ATM.h"

#define PRECOMPUTED_HEADERS 0
#if PRECOMPUTED_HEADERS
#define PRECOMPUTED_HEADER_NUM 16
#else
#define PRECOMPUTED_HEADER_NUM 32
#endif

#define IDLE_PAYLOAD_CHAR	0x6a

enum framealvc_cnxt {
  FRAMEALVC_CNXT_PUTTX = 0,
  FRAMEALVC_CNXT_GETTX,
  FRAMEALVC_CNXT_PUTRX,
  FRAMEALVC_CNXT_RXCOMP,
/**/
  FRAMEALVC_CNXT_NO
};

typedef struct	_FRAMEAL_ATM_VC_T
{
#define HAS_CELLPERBUCKET 0
#if HAS_CELLPERBUCKET
  DWORD			CellPerBucket;		// cells sent in this time interval

  DWORD			CellPerBucketThreshold; // Maxs cells sent in an interval
#endif

#if PRECOMPUTED_HEADERS
  ATM_HEADER_T		TxPreCompHeader[PRECOMPUTED_HEADER_NUM];
#else
  ATM_HEADER_T		TxPreCompHeader;
  BYTE			TxPreCompHEC[PRECOMPUTED_HEADER_NUM];
#endif

#define HAS_CELLSTOSAP 0
#if HAS_CELLSTOSAP
  WORD			CellsToSap;
#endif

  BOOLEAN		InVCALCNXT[FRAMEALVC_CNXT_NO];

  FRAMEAL_ATM_VC_SHAPER_T	VCShaper;

  FRAMEAL_ATM_OAM_VC_T	OAMVc;

  // function pointers for SAP service routines
  SK_BUFF_T*		(*TxGetFullBuf) (
					 IN  CDSL_VC_T *VcCtx,
					 I_O SK_BUFF_T *BufDescPtr
					 );
  VOID			(*TxPutEmptyBuf)(
					 IN CDSL_VC_T *LinkHdl,
					 IN SK_BUFF_T *InBufDescPtr
					 );
  VOID			(*RxPutFullBuf) (
					 IN  CDSL_VC_T  *VcCtx,
					 I_O SK_BUFF_T  *BufDescPtr
					 );
  VOID			(*RxPutBufComplete)(
					    IN CDSL_VC_T *VcCtx,
					    I_O SK_BUFF_T  *BufDescPtr
					    );
} FRAMEAL_ATM_VC_T, *pFRAMEAL_ATM_VC_T;

typedef struct	_FRAMEAL_ATM_LINK_T
{
  BOOLEAN	BlockedForTrafficShaping;

  BYTE		IdleCharacter;

  WORD		CellAlignState;

  // link receive data
  ATM_HEADER_T	RxCellHeader;
  DWORD		RxHeaderCount;

  DWORD		RxPayloadCount;

  WORD		CurrentVC;

  WORD		LinkState;	// does the link need to send a start tx or not

  FRAMEAL_SHAPER_STATUS	TrafficShapingActive;

  FRAMEAL_ATM_LINK_SHAPER_T LinkShaper;

#define HAS_IDLEHEADER 0
#if HAS_IDLEHEADER
  ATM_HEADER_T	IdleHeader;
#endif
#define HAS_IDLEPAYLOAD 0
#if HAS_IDLEPAYLOAD
  BYTE		IdlePayload[CELL_PAYLOAD_SIZE];
#endif

#define HAS_LINKOAM 1
#if HAS_LINKOAM
  ATM_HEADER_T	LinkOAMHeader;
  BYTE		LinkOAMPayload[CELL_PAYLOAD_SIZE];
#endif

  FRAMEAL_ATM_OAM_LINK_T  OAMLink;
} FRAMEAL_ATM_LINK_T;

GLOBAL SK_BUFF_T *FrameALATMGetTxFullBuffer
(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*InBufDescPtr	// Pointer to a buffer descriptor for FrameAL to use
);

GLOBAL SK_BUFF_T *FrameALATMTxPutEmptyBuf
(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*InBufDescPtr	// Pointer to a buffer descriptor for FrameAL to use
);

GLOBAL SK_BUFF_T *FrameALATMCellAlign
(
    IN	CDSL_LINK_T	*LinkHdl,	// Link Handle to data stream
    IN	SK_BUFF_T	*BufferDescPtr	// Pointer to a buffer descriptor
);

GLOBAL NDIS_STATUS FrameALATMLinkInit
(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
);

GLOBAL NDIS_STATUS FrameALATMLinkShutdown
(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
);

GLOBAL NDIS_STATUS FrameALATMVCInit
(
    IN	CDSL_VC_T	*VCHdl			// VC Handle to data stream
);

GLOBAL NDIS_STATUS FrameALATMVCShutdown
(
    IN	CDSL_VC_T	*VCHdl			// VC Handle to data stream
);

GLOBAL NDIS_STATUS FrameALATMStartTx
(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
);

GLOBAL void FrameALATMGetTXCell
(
    IN	CDSL_VC_T	*VCHdl,			// VC Handle to data stream
    IN	SK_BUFF_T	*BufferDescPtr		// Pointer to a buffer descriptor
);

GLOBAL void FrameALATMTimerHandler
(
    IN	CDSL_LINK_T	*LinkHdl		// Link Handle to data stream
);

GLOBAL void FrameALATMLinkUp
(
    IN	CDSL_LINK_T	*LinkHdl
);

GLOBAL void FrameALATMLinkDown
(
    IN	CDSL_LINK_T	*LinkHdl
);
#endif //  _FRAMEALATM_H_
