#include <LnxTools.h>

void INIT_EVENT(
    EVENT_HNDL* pEvent) {
  pEvent->SetFlag = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  pEvent->WaitCnt = 0;
#endif
  init_waitqueue_head(&pEvent->WaitQue);
  INIT_SPIN_LOCK(&pEvent->FlagLock);
}

int WAIT_EVENT_TIMEOUT(
    EVENT_HNDL* pEvent,
    ULONG msecDelay) {
  if (msecDelay)
  {
    int TimeRemain = (msecDelay * HZ) / 1000;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    //CLN//LOCK_FLAG(LockFlag);

    //CLN//ACQUIRE_LOCK(&pEvent->FlagLock, LockFlag);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    // indicate that I am waiting
    atomic_inc((atomic_t *)&pEvent->WaitCnt);
#endif

    TimeRemain = wait_event_interruptible_timeout(pEvent->WaitQue,
						  pEvent->SetFlag,TimeRemain);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    // remove wait indication
    atomic_dec((atomic_t *)&pEvent->WaitCnt);
#endif

    //CLN//RELEASE_LOCK(&pEvent->FlagLock, LockFlag);
#else
    unsigned long flags;

    // atomically sleep if event is not set - we must protect the window between
    // finding out if the flag is already set and sleeping as if an interrupt
    // occurs in the window and calls SET_EVENT, SET_EVENT will signal the wait
    // queue before we get placed on it and we will miss this signal.

    save_and_cli(flags);

    // indicate that I am waiting
    pEvent->WaitCnt++;

    if (!pEvent->SetFlag) {
      while (!pEvent->SetFlag && TimeRemain) {
	// safe to call sleep with interrupts off as kernel will re-enable
	// interrupts after placing us on the queue and before calling schedule
	TimeRemain = interruptible_sleep_on_timeout(&pEvent->WaitQue, TimeRemain);
      }
    }
    // remove wait indication
    pEvent->WaitCnt--;

    restore_flags(flags);
#endif
  }
  return pEvent->SetFlag;  /* 0=>as if timed out, 1=>now set */
}

void SET_EVENT(
    EVENT_HNDL* pEvent) {
  LOCK_FLAG(LockFlag);

  ACQUIRE_LOCK(&pEvent->FlagLock, LockFlag);

  // if the event is cleared then
  // set the flag, wake up all waiting tasks.
  if (!pEvent->SetFlag) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    int cnt;
#endif

    pEvent->SetFlag = TRUE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    for (cnt=0; cnt<pEvent->WaitCnt; cnt++)
#endif
      wake_up_interruptible(&pEvent->WaitQue);
  }

  RELEASE_LOCK(&pEvent->FlagLock, LockFlag);
}

void SET_EVENT_FROM_ISR(
    EVENT_HNDL* pEvent) {
  ACQUIRE_LOCK_AT_ISR(&pEvent->FlagLock);

  // if the event is cleared then set the flag, wake up all waiting tasks.
  if (!pEvent->SetFlag) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    int cnt;
#endif

    pEvent->SetFlag = TRUE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    for (cnt=0; cnt<pEvent->WaitCnt; cnt++)
#endif
      wake_up_interruptible(&pEvent->WaitQue);
  }

  RELEASE_LOCK_AT_ISR(&pEvent->FlagLock);
}

void RESET_EVENT(
    EVENT_HNDL* pEvent) {
  LOCK_FLAG(LockFlag);

  ACQUIRE_LOCK(&pEvent->FlagLock, LockFlag);

  // clear the event flag
  pEvent->SetFlag = FALSE;

  // empty out the queue
  while (waitqueue_active(&pEvent->WaitQue))
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    wait_event_interruptible(pEvent->WaitQue, 0);
#else
    interruptible_sleep_on(&pEvent->WaitQue);
#endif

  RELEASE_LOCK(&pEvent->FlagLock, LockFlag);
}
